/*
** 2000-05-29
**
** The author disclaims copyright to this source code.  In place of
** a legal notice, here is a blessing:
**
**    May you do good and not evil.
**    May you find forgiveness for yourself and forgive others.
**    May you share freely, never taking more than you give.
**
*************************************************************************
** Driver template for the LEMON parser generator.
**
** The "lemon" program processes an LALR(1) input grammar file, then uses
** this template to construct a parser.  The "lemon" program inserts text
** at each "%%" line.  Also, any "P-a-r-s-e" identifer prefix (without the
** interstitial "-" characters) contained in this template is changed into
** the value of the %name directive from the grammar.  Otherwise, the content
** of this template is copied straight through into the generate parser
** source file.
**
** The following is the concatenation of all %include directives from the
** input grammar file:
*/
#include <stdio.h>
#include <assert.h>
/************ Begin %include sections from the grammar ************************/
#line 52 "parse.y"

#include "sqliteInt.h"

/*
** Disable all error recovery processing in the parser push-down
** automaton.
*/
#define YYNOERRORRECOVERY 1

/*
** Make yytestcase() the same as testcase()
*/
#define yytestcase(X) testcase(X)

/*
** Indicate that sqlite3ParserFree() will never be called with a null
** pointer.
*/
#define YYPARSEFREENEVERNULL 1

/*
** In the amalgamation, the parse.c file generated by lemon and the
** tokenize.c file are concatenated.  In that case, sqlite3RunParser()
** has access to the the size of the yyParser object and so the parser
** engine can be allocated from stack.  In that case, only the
** sqlite3ParserInit() and sqlite3ParserFinalize() routines are invoked
** and the sqlite3ParserAlloc() and sqlite3ParserFree() routines can be
** omitted.
*/
#ifdef SQLITE_AMALGAMATION
# define sqlite3Parser_ENGINEALWAYSONSTACK 1
#endif

/*
** Alternative datatype for the argument to the malloc() routine passed
** into sqlite3ParserAlloc().  The default is size_t.
*/
#define YYMALLOCARGTYPE  u64

/*
** An instance of the following structure describes the event of a
** TRIGGER.  "a" is the event type, one of TK_UPDATE, TK_INSERT,
** TK_DELETE, or TK_INSTEAD.  If the event is of the form
**
**      UPDATE ON (a,b,c)
**
** Then the "b" IdList records the list "a,b,c".
*/
struct TrigEvent { int a; IdList * b; };

struct FrameBound     { int eType; Expr *pExpr; };

/*
** Disable lookaside memory allocation for objects that might be
** shared across database connections.
*/
static void disableLookaside(Parse *pParse){
  pParse->disableLookaside++;
  pParse->db->lookaside.bDisable++;
}

#line 455 "parse.y"

  /*
  ** For a compound SELECT statement, make sure p->pPrior->pNext==p for
  ** all elements in the list.  And make sure list length does not exceed
  ** SQLITE_LIMIT_COMPOUND_SELECT.
  */
  static void parserDoubleLinkSelect(Parse *pParse, Select *p){
    assert( p!=0 );
    if( p->pPrior ){
      Select *pNext = 0, *pLoop;
      int mxSelect, cnt = 0;
      for(pLoop=p; pLoop; pNext=pLoop, pLoop=pLoop->pPrior, cnt++){
        pLoop->pNext = pNext;
        pLoop->selFlags |= SF_Compound;
      }
      if( (p->selFlags & SF_MultiValue)==0 && 
        (mxSelect = pParse->db->aLimit[SQLITE_LIMIT_COMPOUND_SELECT])>0 &&
        cnt>mxSelect
      ){
        sqlite3ErrorMsg(pParse, "too many terms in compound SELECT");
      }
    }
  }
#line 947 "parse.y"


  /* Construct a new Expr object from a single identifier.  Use the
  ** new Expr to populate pOut.  Set the span of pOut to be the identifier
  ** that created the expression.
  */
  static Expr *tokenExpr(Parse *pParse, int op, Token t){
    Expr *p = sqlite3DbMallocRawNN(pParse->db, sizeof(Expr)+t.n+1);
    if( p ){
      /* memset(p, 0, sizeof(Expr)); */
      p->op = (u8)op;
      p->affExpr = 0;
      p->flags = EP_Leaf;
      p->iAgg = -1;
      p->pLeft = p->pRight = 0;
      p->x.pList = 0;
      p->pAggInfo = 0;
      p->y.pTab = 0;
      p->op2 = 0;
      p->iTable = 0;
      p->iColumn = 0;
      p->u.zToken = (char*)&p[1];
      memcpy(p->u.zToken, t.z, t.n);
      p->u.zToken[t.n] = 0;
      if( sqlite3Isquote(p->u.zToken[0]) ){
        sqlite3DequoteExpr(p);
      }
#if SQLITE_MAX_EXPR_DEPTH>0
      p->nHeight = 1;
#endif  
      if( IN_RENAME_OBJECT ){
        return (Expr*)sqlite3RenameTokenMap(pParse, (void*)p, &t);
      }
    }
    return p;
  }

#line 1118 "parse.y"

  /* A routine to convert a binary TK_IS or TK_ISNOT expression into a
  ** unary TK_ISNULL or TK_NOTNULL expression. */
  static void binaryToUnaryIfNull(Parse *pParse, Expr *pY, Expr *pA, int op){
    sqlite3 *db = pParse->db;
    if( pA && pY && pY->op==TK_NULL && !IN_RENAME_OBJECT ){
      pA->op = (u8)op;
      sqlite3ExprDelete(db, pA->pRight);
      pA->pRight = 0;
    }
  }
#line 1309 "parse.y"

  /* Add a single new term to an ExprList that is used to store a
  ** list of identifiers.  Report an error if the ID list contains
  ** a COLLATE clause or an ASC or DESC keyword, except ignore the
  ** error while parsing a legacy schema.
  */
  static ExprList *parserAddExprIdListTerm(
    Parse *pParse,
    ExprList *pPrior,
    Token *pIdToken,
    int hasCollate,
    int sortOrder
  ){
    ExprList *p = sqlite3ExprListAppend(pParse, pPrior, 0);
    if( (hasCollate || sortOrder!=SQLITE_SO_UNDEFINED)
        && pParse->db->init.busy==0
    ){
      sqlite3ErrorMsg(pParse, "syntax error after column name \"%.*s\"",
                         pIdToken->n, pIdToken->z);
    }
    sqlite3ExprListSetName(pParse, p, pIdToken, 1);
    return p;
  }
#line 1770 "parse.y"

#if TK_SPAN>255
# error too many tokens in the grammar
#endif
#line 194 "parse.c"
/**************** End of %include directives **********************************/
/* These constants specify the various numeric values for terminal symbols
** in a format understandable to "makeheaders".  This section is blank unless
** "lemon" is run with the "-m" command-line option.
***************** Begin makeheaders token definitions *************************/
/**************** End makeheaders token definitions ***************************/

/* The next sections is a series of control #defines.
** various aspects of the generated parser.
**    YYCODETYPE         is the data type used to store the integer codes
**                       that represent terminal and non-terminal symbols.
**                       "unsigned char" is used if there are fewer than
**                       256 symbols.  Larger types otherwise.
**    YYNOCODE           is a number of type YYCODETYPE that is not used for
**                       any terminal or nonterminal symbol.
**    YYFALLBACK         If defined, this indicates that one or more tokens
**                       (also known as: "terminal symbols") have fall-back
**                       values which should be used if the original symbol
**                       would not parse.  This permits keywords to sometimes
**                       be used as identifiers, for example.
**    YYACTIONTYPE       is the data type used for "action codes" - numbers
**                       that indicate what to do in response to the next
**                       token.
**    sqlite3ParserTOKENTYPE     is the data type used for minor type for terminal
**                       symbols.  Background: A "minor type" is a semantic
**                       value associated with a terminal or non-terminal
**                       symbols.  For example, for an "ID" terminal symbol,
**                       the minor type might be the name of the identifier.
**                       Each non-terminal can have a different minor type.
**                       Terminal symbols all have the same minor type, though.
**                       This macros defines the minor type for terminal 
**                       symbols.
**    YYMINORTYPE        is the data type used for all minor types.
**                       This is typically a union of many types, one of
**                       which is sqlite3ParserTOKENTYPE.  The entry in the union
**                       for terminal symbols is called "yy0".
**    YYSTACKDEPTH       is the maximum depth of the parser's stack.  If
**                       zero the stack is dynamically sized using realloc()
**    sqlite3ParserARG_SDECL     A static variable declaration for the %extra_argument
**    sqlite3ParserARG_PDECL     A parameter declaration for the %extra_argument
**    sqlite3ParserARG_PARAM     Code to pass %extra_argument as a subroutine parameter
**    sqlite3ParserARG_STORE     Code to store %extra_argument into yypParser
**    sqlite3ParserARG_FETCH     Code to extract %extra_argument from yypParser
**    sqlite3ParserCTX_*         As sqlite3ParserARG_ except for %extra_context
**    YYERRORSYMBOL      is the code number of the error symbol.  If not
**                       defined, then do no error processing.
**    YYNSTATE           the combined number of states.
**    YYNRULE            the number of rules in the grammar
**    YYNTOKEN           Number of terminal symbols
**    YY_MAX_SHIFT       Maximum value for shift actions
**    YY_MIN_SHIFTREDUCE Minimum value for shift-reduce actions
**    YY_MAX_SHIFTREDUCE Maximum value for shift-reduce actions
**    YY_ERROR_ACTION    The yy_action[] code for syntax error
**    YY_ACCEPT_ACTION   The yy_action[] code for accept
**    YY_NO_ACTION       The yy_action[] code for no-op
**    YY_MIN_REDUCE      Minimum value for reduce actions
**    YY_MAX_REDUCE      Maximum value for reduce actions
*/
#ifndef INTERFACE
# define INTERFACE 1
#endif
/************* Begin control #defines *****************************************/
#define YYCODETYPE unsigned short int
#define YYNOCODE 307
#define YYACTIONTYPE unsigned short int
#define YYWILDCARD 98
#define sqlite3ParserTOKENTYPE Token
typedef union {
  int yyinit;
  sqlite3ParserTOKENTYPE yy0;
  const char* yy8;
  Select* yy25;
  int yy32;
  Expr* yy46;
  struct FrameBound yy57;
  u8 yy118;
  ExprList* yy138;
  Upsert* yy288;
  With* yy297;
  IdList* yy406;
  Window* yy455;
  struct {int value; int mask;} yy495;
  TriggerStep* yy527;
  struct TrigEvent yy572;
  SrcList* yy609;
} YYMINORTYPE;
#ifndef YYSTACKDEPTH
#define YYSTACKDEPTH 100
#endif
#define sqlite3ParserARG_SDECL
#define sqlite3ParserARG_PDECL
#define sqlite3ParserARG_PARAM
#define sqlite3ParserARG_FETCH
#define sqlite3ParserARG_STORE
#define sqlite3ParserCTX_SDECL Parse *pParse;
#define sqlite3ParserCTX_PDECL ,Parse *pParse
#define sqlite3ParserCTX_PARAM ,pParse
#define sqlite3ParserCTX_FETCH Parse *pParse=yypParser->pParse;
#define sqlite3ParserCTX_STORE yypParser->pParse=pParse;
#define YYFALLBACK 1
#define YYNSTATE             543
#define YYNRULE              381
#define YYNTOKEN             179
#define YY_MAX_SHIFT         542
#define YY_MIN_SHIFTREDUCE   790
#define YY_MAX_SHIFTREDUCE   1170
#define YY_ERROR_ACTION      1171
#define YY_ACCEPT_ACTION     1172
#define YY_NO_ACTION         1173
#define YY_MIN_REDUCE        1174
#define YY_MAX_REDUCE        1554
/************* End control #defines *******************************************/
#define YY_NLOOKAHEAD ((int)(sizeof(yy_lookahead)/sizeof(yy_lookahead[0])))

/* Define the yytestcase() macro to be a no-op if is not already defined
** otherwise.
**
** Applications can choose to define yytestcase() in the %include section
** to a macro that can assist in verifying code coverage.  For production
** code the yytestcase() macro should be turned off.  But it is useful
** for testing.
*/
#ifndef yytestcase
# define yytestcase(X)
#endif


/* Next are the tables used to determine what action to take based on the
** current state and lookahead token.  These tables are used to implement
** functions that take a state number and lookahead value and return an
** action integer.  
**
** Suppose the action integer is N.  Then the action is determined as
** follows
**
**   0 <= N <= YY_MAX_SHIFT             Shift N.  That is, push the lookahead
**                                      token onto the stack and goto state N.
**
**   N between YY_MIN_SHIFTREDUCE       Shift to an arbitrary state then
**     and YY_MAX_SHIFTREDUCE           reduce by rule N-YY_MIN_SHIFTREDUCE.
**
**   N == YY_ERROR_ACTION               A syntax error has occurred.
**
**   N == YY_ACCEPT_ACTION              The parser accepts its input.
**
**   N == YY_NO_ACTION                  No such action.  Denotes unused
**                                      slots in the yy_action[] table.
**
**   N between YY_MIN_REDUCE            Reduce by rule N-YY_MIN_REDUCE
**     and YY_MAX_REDUCE
**
** The action table is constructed as a single large table named yy_action[].
** Given state S and lookahead X, the action is computed as either:
**
**    (A)   N = yy_action[ yy_shift_ofst[S] + X ]
**    (B)   N = yy_default[S]
**
** The (A) formula is preferred.  The B formula is used instead if
** yy_lookahead[yy_shift_ofst[S]+X] is not equal to X.
**
** The formulas above are for computing the action when the lookahead is
** a terminal symbol.  If the lookahead is a non-terminal (as occurs after
** a reduce action) then the yy_reduce_ofst[] array is used in place of
** the yy_shift_ofst[] array.
**
** The following are the tables generated in this section:
**
**  yy_action[]        A single table containing all actions.
**  yy_lookahead[]     A table containing the lookahead for each entry in
**                     yy_action.  Used to detect hash collisions.
**  yy_shift_ofst[]    For each state, the offset into yy_action for
**                     shifting terminals.
**  yy_reduce_ofst[]   For each state, the offset into yy_action for
**                     shifting non-terminals after a reduce.
**  yy_default[]       Default action for each state.
**
*********** Begin parsing tables **********************************************/
#define YY_ACTTAB_COUNT (1913)
static const YYACTIONTYPE yy_action[] = {
 /*     0 */   537,  339,  537, 1241, 1220,  537,   12,  537,  112,  109,
 /*    10 */   209,  537, 1241,  537, 1205,  462,  112,  109,  209,  386,
 /*    20 */   338,  462,   42,   42,   42,   42,  445,   42,   42,   70,
 /*    30 */    70,  922, 1208,   70,   70,   70,   70, 1443,  403,  923,
 /*    40 */   531,  531,  531,  119,  120,  110, 1148, 1148,  991,  994,
 /*    50 */   984,  984,  117,  117,  118,  118,  118,  118,  425,  386,
 /*    60 */  1498,  542,    2, 1176, 1442,  519,  141, 1518,  289,  519,
 /*    70 */   134,  519,   95,  259,  495, 1215,  189, 1254,  518,  494,
 /*    80 */   484,  437,  296,  119,  120,  110, 1148, 1148,  991,  994,
 /*    90 */   984,  984,  117,  117,  118,  118,  118,  118,  270,  116,
 /*   100 */   116,  116,  116,  115,  115,  114,  114,  114,  113,  418,
 /*   110 */   264,  264,  264,  264,  423, 1479,  352, 1481,  123,  351,
 /*   120 */  1479,  508, 1094,  534, 1034,  534, 1099,  386, 1099,  239,
 /*   130 */   206,  112,  109,  209,   96, 1094,  376,  219, 1094,  116,
 /*   140 */   116,  116,  116,  115,  115,  114,  114,  114,  113,  418,
 /*   150 */   480,  119,  120,  110, 1148, 1148,  991,  994,  984,  984,
 /*   160 */   117,  117,  118,  118,  118,  118,  353,  422, 1407,  264,
 /*   170 */   264,  114,  114,  114,  113,  418,  883,  121,  416,  416,
 /*   180 */   416,  882,  534,  116,  116,  116,  116,  115,  115,  114,
 /*   190 */   114,  114,  113,  418,  212,  415,  414,  386,  443,  383,
 /*   200 */   382,  118,  118,  118,  118,  111,  177,  116,  116,  116,
 /*   210 */   116,  115,  115,  114,  114,  114,  113,  418,  112,  109,
 /*   220 */   209,  119,  120,  110, 1148, 1148,  991,  994,  984,  984,
 /*   230 */   117,  117,  118,  118,  118,  118,  386,  438,  312, 1163,
 /*   240 */  1155,   80, 1155, 1127,  514,   79,  116,  116,  116,  116,
 /*   250 */   115,  115,  114,  114,  114,  113,  418,  514,  428,  418,
 /*   260 */   119,  120,  110, 1148, 1148,  991,  994,  984,  984,  117,
 /*   270 */   117,  118,  118,  118,  118,  428,  427,  116,  116,  116,
 /*   280 */   116,  115,  115,  114,  114,  114,  113,  418,  115,  115,
 /*   290 */   114,  114,  114,  113,  418, 1127, 1127, 1128, 1129, 1094,
 /*   300 */   258,  258,  192,  386,  408,  371, 1168,  326,  118,  118,
 /*   310 */   118,  118, 1094,  534,  374, 1094,  116,  116,  116,  116,
 /*   320 */   115,  115,  114,  114,  114,  113,  418,  119,  120,  110,
 /*   330 */  1148, 1148,  991,  994,  984,  984,  117,  117,  118,  118,
 /*   340 */   118,  118,  386,  354,  445,  428,  829,  238, 1127, 1128,
 /*   350 */  1129,  515, 1466,  116,  116,  116,  116,  115,  115,  114,
 /*   360 */   114,  114,  113,  418, 1127, 1467,  119,  120,  110, 1148,
 /*   370 */  1148,  991,  994,  984,  984,  117,  117,  118,  118,  118,
 /*   380 */   118, 1169,   82,  116,  116,  116,  116,  115,  115,  114,
 /*   390 */   114,  114,  113,  418,  405,  112,  109,  209,  161,  445,
 /*   400 */   250,  267,  336,  478,  331,  477,  236,  951, 1127,  386,
 /*   410 */   888, 1521,  329,  822,  852,  162,  274, 1127, 1128, 1129,
 /*   420 */   338,  169,  116,  116,  116,  116,  115,  115,  114,  114,
 /*   430 */   114,  113,  418,  119,  120,  110, 1148, 1148,  991,  994,
 /*   440 */   984,  984,  117,  117,  118,  118,  118,  118,  386,  438,
 /*   450 */   312, 1502, 1112, 1176,  161,  288,  528,  311,  289,  883,
 /*   460 */   134, 1127, 1128, 1129,  882,  537,  143, 1254,  288,  528,
 /*   470 */   297,  275,  119,  120,  110, 1148, 1148,  991,  994,  984,
 /*   480 */   984,  117,  117,  118,  118,  118,  118,   70,   70,  116,
 /*   490 */   116,  116,  116,  115,  115,  114,  114,  114,  113,  418,
 /*   500 */   264,  264,   12,  264,  264,  395, 1127,  483, 1473, 1094,
 /*   510 */   204,  482,    6,  534, 1258,  386,  534, 1474,  825,  972,
 /*   520 */   504,    6, 1094,  500,   95, 1094,  534,  219,  116,  116,
 /*   530 */   116,  116,  115,  115,  114,  114,  114,  113,  418,  119,
 /*   540 */   120,  110, 1148, 1148,  991,  994,  984,  984,  117,  117,
 /*   550 */   118,  118,  118,  118,  386, 1339,  971,  422,  956, 1127,
 /*   560 */  1128, 1129,  231,  512, 1473,  475,  472,  471,    6,  113,
 /*   570 */   418,  825,  962,  298,  503,  470,  961,  452,  119,  120,
 /*   580 */   110, 1148, 1148,  991,  994,  984,  984,  117,  117,  118,
 /*   590 */   118,  118,  118,  395,  537,  116,  116,  116,  116,  115,
 /*   600 */   115,  114,  114,  114,  113,  418,  202,  961,  961,  963,
 /*   610 */   231,  971, 1127,  475,  472,  471,   13,   13,  951, 1127,
 /*   620 */   834,  386, 1207,  470,  399,  183,  447,  962,  462,  162,
 /*   630 */   397,  961, 1246, 1246,  116,  116,  116,  116,  115,  115,
 /*   640 */   114,  114,  114,  113,  418,  119,  120,  110, 1148, 1148,
 /*   650 */   991,  994,  984,  984,  117,  117,  118,  118,  118,  118,
 /*   660 */   386,  271,  961,  961,  963, 1127, 1128, 1129,  311,  433,
 /*   670 */   299, 1406, 1127, 1128, 1129,  178, 1471,  138,  162,   32,
 /*   680 */     6, 1127,  288,  528,  119,  120,  110, 1148, 1148,  991,
 /*   690 */   994,  984,  984,  117,  117,  118,  118,  118,  118,  909,
 /*   700 */   390,  116,  116,  116,  116,  115,  115,  114,  114,  114,
 /*   710 */   113,  418, 1127,  429,  817,  537, 1127,  265,  265,  981,
 /*   720 */   981,  992,  995,  324, 1055,   93,  520,    5,  338,  537,
 /*   730 */   534,  288,  528, 1522, 1127, 1128, 1129,   70,   70, 1056,
 /*   740 */   116,  116,  116,  116,  115,  115,  114,  114,  114,  113,
 /*   750 */   418,   70,   70, 1495, 1057,  537,   98, 1244, 1244,  264,
 /*   760 */   264,  908,  371, 1076, 1127, 1127, 1128, 1129,  817, 1127,
 /*   770 */  1128, 1129,  534,  519,  140,  863,  386,   13,   13,  456,
 /*   780 */   192,  193,  521,  453,  319,  864,  322,  284,  365,  430,
 /*   790 */   985,  402,  379, 1077, 1548,  101,  386, 1548,    3,  395,
 /*   800 */   119,  120,  110, 1148, 1148,  991,  994,  984,  984,  117,
 /*   810 */   117,  118,  118,  118,  118,  386,  451, 1127, 1128, 1129,
 /*   820 */   119,  120,  110, 1148, 1148,  991,  994,  984,  984,  117,
 /*   830 */   117,  118,  118,  118,  118, 1127, 1354, 1412, 1169,  119,
 /*   840 */   108,  110, 1148, 1148,  991,  994,  984,  984,  117,  117,
 /*   850 */   118,  118,  118,  118, 1412, 1414,  116,  116,  116,  116,
 /*   860 */   115,  115,  114,  114,  114,  113,  418,  272,  535, 1075,
 /*   870 */   877,  877,  337, 1492,  309,  462,  116,  116,  116,  116,
 /*   880 */   115,  115,  114,  114,  114,  113,  418,  537, 1127, 1128,
 /*   890 */  1129,  537,  360,  537,  356,  116,  116,  116,  116,  115,
 /*   900 */   115,  114,  114,  114,  113,  418,  386,  264,  264,   13,
 /*   910 */    13,  273, 1127,   13,   13,   13,   13,  304, 1253,  386,
 /*   920 */   534, 1077, 1549,  404, 1412, 1549,  496,  277,  451,  186,
 /*   930 */  1252,  120,  110, 1148, 1148,  991,  994,  984,  984,  117,
 /*   940 */   117,  118,  118,  118,  118,  110, 1148, 1148,  991,  994,
 /*   950 */   984,  984,  117,  117,  118,  118,  118,  118,  105,  529,
 /*   960 */   537,    4, 1339,  264,  264, 1127, 1128, 1129, 1039, 1039,
 /*   970 */   459,  795,  796,  797,  536,  532,  534,  242,  301,  807,
 /*   980 */   303,  462,   69,   69,  451, 1353,  116,  116,  116,  116,
 /*   990 */   115,  115,  114,  114,  114,  113,  418, 1075,  419,  116,
 /*  1000 */   116,  116,  116,  115,  115,  114,  114,  114,  113,  418,
 /*  1010 */   526,  537, 1146,  192,  350,  105,  529,  537,    4,  497,
 /*  1020 */   162,  337, 1492,  310, 1249,  385, 1550,  372,    9,  462,
 /*  1030 */   242,  400,  532,   13,   13,  499,  971,  843,  436,   70,
 /*  1040 */    70,  359,  103,  103,    8,  339,  278,  187,  278,  104,
 /*  1050 */  1127,  419,  539,  538, 1339,  419,  961,  302, 1339, 1172,
 /*  1060 */     1,    1,  542,    2, 1176, 1146, 1146,  526,  476,  289,
 /*  1070 */    30,  134,  317,  288,  528,  285,  844, 1014, 1254,  276,
 /*  1080 */  1472,  506,  410, 1194,    6,  207,  505,  961,  961,  963,
 /*  1090 */   964,   27,  449,  971,  415,  414,  234,  233,  232,  103,
 /*  1100 */   103,   31, 1152, 1127, 1128, 1129,  104, 1154,  419,  539,
 /*  1110 */   538,  264,  264,  961, 1399, 1153,  264,  264, 1470, 1146,
 /*  1120 */   537,  216,    6,  401,  534, 1197,  392,  458,  406,  534,
 /*  1130 */   537,  485,  358,  537,  261,  537, 1339,  907,  219, 1155,
 /*  1140 */   467, 1155,   50,   50,  961,  961,  963,  964,   27, 1497,
 /*  1150 */  1116,  421,   70,   70,  268,   70,   70,   13,   13,  369,
 /*  1160 */   369,  368,  253,  366,  264,  264,  804,  235,  422,  105,
 /*  1170 */   529,  516,    4,  287,  487,  510,  493,  534,  486,  213,
 /*  1180 */  1055,  294,  490,  384, 1127,  450,  532,  338,  413,  293,
 /*  1190 */   522,  417,  335, 1036,  509, 1056,  107, 1036,   16,   16,
 /*  1200 */  1469, 1094,  334, 1105,    6,  411, 1145,  264,  264,  419,
 /*  1210 */  1057,  102,  511,  100, 1094,  264,  264, 1094,  922,  215,
 /*  1220 */   534,  526,  907,  264,  264,  208,  923,  154,  534,  457,
 /*  1230 */   156,  525,  391,  142,  218,  506,  534, 1127, 1128, 1129,
 /*  1240 */   507,  139, 1131,   38,  214,  530,  392,  971,  329, 1454,
 /*  1250 */   907, 1105,  537,  103,  103,  105,  529,  537,    4,  537,
 /*  1260 */   104,  424,  419,  539,  538,  537,  502,  961,  517,  537,
 /*  1270 */  1072,  537,  532,  373,   54,   54,  288,  528,  387,   55,
 /*  1280 */    55,   15,   15,  288,  528,   17,  136,   44,   44, 1451,
 /*  1290 */   537,   56,   56,   57,   57,  419, 1131,  291,  961,  961,
 /*  1300 */   963,  964,   27,  393,  163,  537,  426,  526,  263,  206,
 /*  1310 */   208,  517,   58,   58,  235,  440,  842,  841,  197,  105,
 /*  1320 */   529,  506,    4, 1033,  439, 1033,  505,   59,   59,  308,
 /*  1330 */   849,  850,   95,  971,  537,  907,  532,  948,  832,  103,
 /*  1340 */   103,  105,  529,  537,    4, 1021,  104,  537,  419,  539,
 /*  1350 */   538, 1116,  421,  961,  537,  268,   60,   60,  532,  419,
 /*  1360 */   369,  369,  368,  253,  366,   61,   61,  804,  965,   45,
 /*  1370 */    45,  526,  537, 1032, 1277, 1032,   46,   46,  537,  391,
 /*  1380 */   213,  419,  294,  266,  961,  961,  963,  964,   27,  292,
 /*  1390 */   293,  295,  832,  526,   48,   48, 1290,  971, 1289, 1021,
 /*  1400 */    49,   49,  432,  103,  103,  887,  953,  537, 1457,  241,
 /*  1410 */   104,  305,  419,  539,  538,  925,  926,  961,  444,  971,
 /*  1420 */   215,  241,  965, 1224,  537,  103,  103, 1431,  154,   62,
 /*  1430 */    62,  156,  104, 1430,  419,  539,  538,   97,  529,  961,
 /*  1440 */     4,  537,  454,  537,  314,  214,   63,   63,  961,  961,
 /*  1450 */   963,  964,   27,  537,  532,  446, 1286,  318,  241,  537,
 /*  1460 */   321,  323,  325,   64,   64,   14,   14, 1237,  537, 1223,
 /*  1470 */   961,  961,  963,  964,   27,   65,   65,  419,  537,  387,
 /*  1480 */   537,  125,  125,  537,  288,  528,  537, 1486,  537,  526,
 /*  1490 */    66,   66,  313,  524,  537,   95,  468, 1221, 1511,  237,
 /*  1500 */    51,   51,   67,   67,  330,   68,   68,  426,   52,   52,
 /*  1510 */   149,  149, 1222,  340,  341,  971,  150,  150, 1298,  463,
 /*  1520 */   327,  103,  103,   95,  537, 1338, 1273,  537,  104,  537,
 /*  1530 */   419,  539,  538, 1284,  537,  961,  268,  283,  523, 1344,
 /*  1540 */  1204,  369,  369,  368,  253,  366,   75,   75,  804,   53,
 /*  1550 */    53,   71,   71,  537, 1196,  537,  126,  126,  537, 1017,
 /*  1560 */   537,  213,  237,  294,  537, 1185,  961,  961,  963,  964,
 /*  1570 */    27,  293,  537, 1184,  537,   72,   72,  127,  127, 1186,
 /*  1580 */   128,  128,  124,  124, 1505,  537,  148,  148,  537,  256,
 /*  1590 */   195,  537, 1270,  537,  147,  147,  132,  132,  537,   11,
 /*  1600 */   537,  215,  537,  199,  343,  345,  347,  131,  131,  154,
 /*  1610 */   129,  129,  156,  130,  130,   74,   74,  537,  370, 1323,
 /*  1620 */    76,   76,   73,   73,   43,   43,  214,  431,  211, 1331,
 /*  1630 */   300,  916,  880,  815,  241,  107,  137,  307,  881,   47,
 /*  1640 */    47,  107,  473,  378,  203,  448,  333, 1403, 1220, 1402,
 /*  1650 */   349,  190,  527,  191,  363,  198, 1508, 1163,  245,  165,
 /*  1660 */   387, 1450, 1448, 1160,   78,  288,  528, 1408,   81,  394,
 /*  1670 */    82,  442,  175,  159,  167,   93, 1328,   35, 1320,  434,
 /*  1680 */   170,  171,  172,  173,  435,  466,  221,  375,  426,  377,
 /*  1690 */  1334,  179,  455,  441, 1397,  225,   87,   36,  461, 1419,
 /*  1700 */   316,  257,  227,  184,  320,  464,  228,  479, 1187,  229,
 /*  1710 */   380, 1240, 1239,  407, 1238, 1212,  834,  332, 1231,  381,
 /*  1720 */   409, 1211,  204, 1210, 1491,  498, 1520, 1281,   92,  281,
 /*  1730 */  1230,  489,  282,  492,  342,  243, 1282,  344,  244, 1280,
 /*  1740 */   346,  412, 1279, 1477,  348,  122, 1476,  517,   10,  357,
 /*  1750 */   286, 1305, 1304,   99, 1383,   94,  501,  251, 1193,   34,
 /*  1760 */  1263,  355,  540,  194, 1262,  361,  362, 1122,  252,  254,
 /*  1770 */   255,  388,  541, 1182, 1177,  151, 1435,  389, 1436, 1434,
 /*  1780 */  1433,  791,  152,  135,  279,  200,  201,  420,  196,   77,
 /*  1790 */   153,  290,  269,  210, 1031,  133, 1029,  945,  166,  155,
 /*  1800 */   217,  168,  866,  306,  220, 1045,  174,  949,  157,  396,
 /*  1810 */    83,  398,  176,   84,   85,  164,   86,  158, 1048,  222,
 /*  1820 */   223, 1044,  144,   18,  224,  315, 1037,  180,  241,  460,
 /*  1830 */  1157,  226,  181,   37,  806,  465,  334,  230,  328,  469,
 /*  1840 */   182,   88,  474,   19,   20,  160,   89,  280,  145,   90,
 /*  1850 */   481,  845, 1110,  146,  997,  205, 1080,   39,   91,   40,
 /*  1860 */   488, 1081,  915,  491,  260,  262,  185,  910,  240,  107,
 /*  1870 */  1100, 1096, 1098, 1104,   21, 1084,   33,  513,  247,   22,
 /*  1880 */    23,   24, 1103,   25,  188,   95, 1012,  998,  996,   26,
 /*  1890 */  1000, 1054,    7, 1053, 1001,  246,   28,   41,  533,  966,
 /*  1900 */   816,  106,   29,  367,  248,  249, 1513, 1512,  364, 1117,
 /*  1910 */  1173, 1173,  876,
};
static const YYCODETYPE yy_lookahead[] = {
 /*     0 */   187,  187,  187,  216,  217,  187,  206,  187,  264,  265,
 /*    10 */   266,  187,  225,  187,  209,  187,  264,  265,  266,   19,
 /*    20 */   187,  187,  209,  210,  209,  210,  187,  209,  210,  209,
 /*    30 */   210,   31,  209,  209,  210,  209,  210,  285,  224,   39,
 /*    40 */   203,  204,  205,   43,   44,   45,   46,   47,   48,   49,
 /*    50 */    50,   51,   52,   53,   54,   55,   56,   57,  230,   19,
 /*    60 */   181,  182,  183,  184,  230,  245,  233,  208,  189,  245,
 /*    70 */   191,  245,   26,  206,  254,  216,  276,  198,  254,  198,
 /*    80 */   254,  281,  187,   43,   44,   45,   46,   47,   48,   49,
 /*    90 */    50,   51,   52,   53,   54,   55,   56,   57,  259,   99,
 /*   100 */   100,  101,  102,  103,  104,  105,  106,  107,  108,  109,
 /*   110 */   231,  232,  231,  232,  286,  302,  303,  302,   22,  304,
 /*   120 */   302,  303,   76,  244,   11,  244,   86,   19,   88,  248,
 /*   130 */   249,  264,  265,  266,   26,   89,  198,  258,   92,   99,
 /*   140 */   100,  101,  102,  103,  104,  105,  106,  107,  108,  109,
 /*   150 */   105,   43,   44,   45,   46,   47,   48,   49,   50,   51,
 /*   160 */    52,   53,   54,   55,   56,   57,  212,  288,  273,  231,
 /*   170 */   232,  105,  106,  107,  108,  109,  131,   69,  203,  204,
 /*   180 */   205,  136,  244,   99,  100,  101,  102,  103,  104,  105,
 /*   190 */   106,  107,  108,  109,   15,  103,  104,   19,  260,  103,
 /*   200 */   104,   54,   55,   56,   57,   58,   22,   99,  100,  101,
 /*   210 */   102,  103,  104,  105,  106,  107,  108,  109,  264,  265,
 /*   220 */   266,   43,   44,   45,   46,   47,   48,   49,   50,   51,
 /*   230 */    52,   53,   54,   55,   56,   57,   19,  124,  125,   60,
 /*   240 */   148,   24,  150,   59,  187,   67,   99,  100,  101,  102,
 /*   250 */   103,  104,  105,  106,  107,  108,  109,  187,  187,  109,
 /*   260 */    43,   44,   45,   46,   47,   48,   49,   50,   51,   52,
 /*   270 */    53,   54,   55,   56,   57,  204,  205,   99,  100,  101,
 /*   280 */   102,  103,  104,  105,  106,  107,  108,  109,  103,  104,
 /*   290 */   105,  106,  107,  108,  109,   59,  112,  113,  114,   76,
 /*   300 */   231,  232,  187,   19,   19,   22,   23,   23,   54,   55,
 /*   310 */    56,   57,   89,  244,  199,   92,   99,  100,  101,  102,
 /*   320 */   103,  104,  105,  106,  107,  108,  109,   43,   44,   45,
 /*   330 */    46,   47,   48,   49,   50,   51,   52,   53,   54,   55,
 /*   340 */    56,   57,   19,  212,  187,  274,   23,   26,  112,  113,
 /*   350 */   114,  294,  295,   99,  100,  101,  102,  103,  104,  105,
 /*   360 */   106,  107,  108,  109,   59,  295,   43,   44,   45,   46,
 /*   370 */    47,   48,   49,   50,   51,   52,   53,   54,   55,   56,
 /*   380 */    57,   98,  146,   99,  100,  101,  102,  103,  104,  105,
 /*   390 */   106,  107,  108,  109,  109,  264,  265,  266,  187,  187,
 /*   400 */   115,  116,  117,  118,  119,  120,  121,   73,   59,   19,
 /*   410 */   105,   23,  127,   23,   26,   81,  259,  112,  113,  114,
 /*   420 */   187,   72,   99,  100,  101,  102,  103,  104,  105,  106,
 /*   430 */   107,  108,  109,   43,   44,   45,   46,   47,   48,   49,
 /*   440 */    50,   51,   52,   53,   54,   55,   56,   57,   19,  124,
 /*   450 */   125,  182,   23,  184,  187,  134,  135,  123,  189,  131,
 /*   460 */   191,  112,  113,  114,  136,  187,  233,  198,  134,  135,
 /*   470 */   198,  259,   43,   44,   45,   46,   47,   48,   49,   50,
 /*   480 */    51,   52,   53,   54,   55,   56,   57,  209,  210,   99,
 /*   490 */   100,  101,  102,  103,  104,  105,  106,  107,  108,  109,
 /*   500 */   231,  232,  206,  231,  232,  187,   59,  296,  297,   76,
 /*   510 */   160,  161,  301,  244,  232,   19,  244,  297,   59,   23,
 /*   520 */    87,  301,   89,  245,   26,   92,  244,  258,   99,  100,
 /*   530 */   101,  102,  103,  104,  105,  106,  107,  108,  109,   43,
 /*   540 */    44,   45,   46,   47,   48,   49,   50,   51,   52,   53,
 /*   550 */    54,   55,   56,   57,   19,  187,   97,  288,   23,  112,
 /*   560 */   113,  114,  115,  296,  297,  118,  119,  120,  301,  108,
 /*   570 */   109,  112,  113,  255,  141,  128,  117,  281,   43,   44,
 /*   580 */    45,   46,   47,   48,   49,   50,   51,   52,   53,   54,
 /*   590 */    55,   56,   57,  187,  187,   99,  100,  101,  102,  103,
 /*   600 */   104,  105,  106,  107,  108,  109,   26,  148,  149,  150,
 /*   610 */   115,   97,   59,  118,  119,  120,  209,  210,   73,   59,
 /*   620 */   122,   19,  209,  128,  256,   72,  187,  113,  187,   81,
 /*   630 */   223,  117,  227,  228,   99,  100,  101,  102,  103,  104,
 /*   640 */   105,  106,  107,  108,  109,   43,   44,   45,   46,   47,
 /*   650 */    48,   49,   50,   51,   52,   53,   54,   55,   56,   57,
 /*   660 */    19,  255,  148,  149,  150,  112,  113,  114,  123,  124,
 /*   670 */   125,  230,  112,  113,  114,   22,  297,   22,   81,   22,
 /*   680 */   301,   59,  134,  135,   43,   44,   45,   46,   47,   48,
 /*   690 */    49,   50,   51,   52,   53,   54,   55,   56,   57,  139,
 /*   700 */   192,   99,  100,  101,  102,  103,  104,  105,  106,  107,
 /*   710 */   108,  109,   59,  116,   59,  187,   59,  231,  232,   46,
 /*   720 */    47,   48,   49,   16,   12,  145,  198,   22,  187,  187,
 /*   730 */   244,  134,  135,  222,  112,  113,  114,  209,  210,   27,
 /*   740 */    99,  100,  101,  102,  103,  104,  105,  106,  107,  108,
 /*   750 */   109,  209,  210,  187,   42,  187,  154,  227,  228,  231,
 /*   760 */   232,  139,   22,   23,   59,  112,  113,  114,  113,  112,
 /*   770 */   113,  114,  244,  245,  233,   63,   19,  209,  210,  271,
 /*   780 */   187,   24,  254,  275,   77,   73,   79,  245,  195,  260,
 /*   790 */   117,  223,  199,   22,   23,  154,   19,   26,   22,  187,
 /*   800 */    43,   44,   45,   46,   47,   48,   49,   50,   51,   52,
 /*   810 */    53,   54,   55,   56,   57,   19,  187,  112,  113,  114,
 /*   820 */    43,   44,   45,   46,   47,   48,   49,   50,   51,   52,
 /*   830 */    53,   54,   55,   56,   57,   59,  263,  187,   98,   43,
 /*   840 */    44,   45,   46,   47,   48,   49,   50,   51,   52,   53,
 /*   850 */    54,   55,   56,   57,  204,  205,   99,  100,  101,  102,
 /*   860 */   103,  104,  105,  106,  107,  108,  109,  255,  130,   98,
 /*   870 */   132,  133,  299,  300,  198,  187,   99,  100,  101,  102,
 /*   880 */   103,  104,  105,  106,  107,  108,  109,  187,  112,  113,
 /*   890 */   114,  187,  241,  187,  243,   99,  100,  101,  102,  103,
 /*   900 */   104,  105,  106,  107,  108,  109,   19,  231,  232,  209,
 /*   910 */   210,  282,   59,  209,  210,  209,  210,   16,  230,   19,
 /*   920 */   244,   22,   23,  223,  274,   26,   19,  223,  187,  223,
 /*   930 */   198,   44,   45,   46,   47,   48,   49,   50,   51,   52,
 /*   940 */    53,   54,   55,   56,   57,   45,   46,   47,   48,   49,
 /*   950 */    50,   51,   52,   53,   54,   55,   56,   57,   19,   20,
 /*   960 */   187,   22,  187,  231,  232,  112,  113,  114,  123,  124,
 /*   970 */   125,    7,    8,    9,  187,   36,  244,   24,   77,   21,
 /*   980 */    79,  187,  209,  210,  187,  263,   99,  100,  101,  102,
 /*   990 */   103,  104,  105,  106,  107,  108,  109,   98,   59,   99,
 /*  1000 */   100,  101,  102,  103,  104,  105,  106,  107,  108,  109,
 /*  1010 */    71,  187,   59,  187,  187,   19,   20,  187,   22,  112,
 /*  1020 */    81,  299,  300,  282,  230,  199,  291,  292,   22,  187,
 /*  1030 */    24,  256,   36,  209,  210,  187,   97,   35,   80,  209,
 /*  1040 */   210,  268,  103,  104,   48,  187,  220,  223,  222,  110,
 /*  1050 */    59,  112,  113,  114,  187,   59,  117,  156,  187,  179,
 /*  1060 */   180,  181,  182,  183,  184,   59,  113,   71,   66,  189,
 /*  1070 */    22,  191,  230,  134,  135,  245,   74,  119,  198,  282,
 /*  1080 */   297,   85,  224,  198,  301,  187,   90,  148,  149,  150,
 /*  1090 */   151,  152,   19,   97,  103,  104,  123,  124,  125,  103,
 /*  1100 */   104,   53,  111,  112,  113,  114,  110,  116,  112,  113,
 /*  1110 */   114,  231,  232,  117,  156,  124,  231,  232,  297,  113,
 /*  1120 */   187,   24,  301,  256,  244,  201,  202,  256,  126,  244,
 /*  1130 */   187,  198,  187,  187,   23,  187,  187,   26,  258,  148,
 /*  1140 */    19,  150,  209,  210,  148,  149,  150,  151,  152,    0,
 /*  1150 */     1,    2,  209,  210,    5,  209,  210,  209,  210,   10,
 /*  1160 */    11,   12,   13,   14,  231,  232,   17,   46,  288,   19,
 /*  1170 */    20,  223,   22,  236,  198,   66,  187,  244,  245,   30,
 /*  1180 */    12,   32,  198,  246,   59,  112,   36,  187,  245,   40,
 /*  1190 */   198,  245,  117,   29,   85,   27,   26,   33,  209,  210,
 /*  1200 */   297,   76,  127,   94,  301,  256,   26,  231,  232,   59,
 /*  1210 */    42,  153,   87,  155,   89,  231,  232,   92,   31,   70,
 /*  1220 */   244,   71,   26,  231,  232,  114,   39,   78,  244,   65,
 /*  1230 */    81,   63,  111,  233,  137,   85,  244,  112,  113,  114,
 /*  1240 */    90,   22,   59,   24,   95,  201,  202,   97,  127,  187,
 /*  1250 */   139,  142,  187,  103,  104,   19,   20,  187,   22,  187,
 /*  1260 */   110,  187,  112,  113,  114,  187,  141,  117,  141,  187,
 /*  1270 */    23,  187,   36,   26,  209,  210,  134,  135,  129,  209,
 /*  1280 */   210,  209,  210,  134,  135,   22,  159,  209,  210,  187,
 /*  1290 */   187,  209,  210,  209,  210,   59,  113,  187,  148,  149,
 /*  1300 */   150,  151,  152,  289,  290,  187,  157,   71,  248,  249,
 /*  1310 */   114,  141,  209,  210,   46,  125,  116,  117,  138,   19,
 /*  1320 */    20,   85,   22,  148,   61,  150,   90,  209,  210,   23,
 /*  1330 */     7,    8,   26,   97,  187,  139,   36,  147,   59,  103,
 /*  1340 */   104,   19,   20,  187,   22,   59,  110,  187,  112,  113,
 /*  1350 */   114,    1,    2,  117,  187,    5,  209,  210,   36,   59,
 /*  1360 */    10,   11,   12,   13,   14,  209,  210,   17,   59,  209,
 /*  1370 */   210,   71,  187,  148,  250,  150,  209,  210,  187,  111,
 /*  1380 */    30,   59,   32,   22,  148,  149,  150,  151,  152,  187,
 /*  1390 */    40,  187,  113,   71,  209,  210,  187,   97,  187,  113,
 /*  1400 */   209,  210,  187,  103,  104,  105,   23,  187,  187,   26,
 /*  1410 */   110,  187,  112,  113,  114,   83,   84,  117,   23,   97,
 /*  1420 */    70,   26,  113,  218,  187,  103,  104,  187,   78,  209,
 /*  1430 */   210,   81,  110,  187,  112,  113,  114,   19,   20,  117,
 /*  1440 */    22,  187,  187,  187,  187,   95,  209,  210,  148,  149,
 /*  1450 */   150,  151,  152,  187,   36,   23,  187,  187,   26,  187,
 /*  1460 */   187,  187,  187,  209,  210,  209,  210,  187,  187,  218,
 /*  1470 */   148,  149,  150,  151,  152,  209,  210,   59,  187,  129,
 /*  1480 */   187,  209,  210,  187,  134,  135,  187,  306,  187,   71,
 /*  1490 */   209,  210,   23,  228,  187,   26,   23,  187,  137,   26,
 /*  1500 */   209,  210,  209,  210,  187,  209,  210,  157,  209,  210,
 /*  1510 */   209,  210,  218,  187,  187,   97,  209,  210,  187,  278,
 /*  1520 */    23,  103,  104,   26,  187,  187,  187,  187,  110,  187,
 /*  1530 */   112,  113,  114,  187,  187,  117,    5,  247,  187,  187,
 /*  1540 */   187,   10,   11,   12,   13,   14,  209,  210,   17,  209,
 /*  1550 */   210,  209,  210,  187,  187,  187,  209,  210,  187,   23,
 /*  1560 */   187,   30,   26,   32,  187,  187,  148,  149,  150,  151,
 /*  1570 */   152,   40,  187,  187,  187,  209,  210,  209,  210,  187,
 /*  1580 */   209,  210,  209,  210,  187,  187,  209,  210,  187,  277,
 /*  1590 */   234,  187,  247,  187,  209,  210,  209,  210,  187,  235,
 /*  1600 */   187,   70,  187,  207,  247,  247,  247,  209,  210,   78,
 /*  1610 */   209,  210,   81,  209,  210,  209,  210,  187,  185,  238,
 /*  1620 */   209,  210,  209,  210,  209,  210,   95,  251,  287,  238,
 /*  1630 */   251,   23,   23,   23,   26,   26,   26,  283,   23,  209,
 /*  1640 */   210,   26,  213,  238,  221,  283,  212,  212,  217,  212,
 /*  1650 */   251,  241,  270,  241,  237,  235,  190,   60,  137,  287,
 /*  1660 */   129,  194,  194,   38,  284,  134,  135,  273,  284,  194,
 /*  1670 */   146,  111,   22,   43,  226,  145,  262,  261,  238,   18,
 /*  1680 */   229,  229,  229,  229,  194,   18,  193,  238,  157,  262,
 /*  1690 */   226,  226,  194,  238,  238,  193,  153,  261,   62,  280,
 /*  1700 */   279,  194,  193,   22,  194,  214,  193,  111,  194,  193,
 /*  1710 */   214,  211,  211,   64,  211,  211,  122,  211,  219,  214,
 /*  1720 */   109,  213,  160,  211,  300,  140,  211,  253,  111,  272,
 /*  1730 */   219,  214,  272,  214,  252,  194,  253,  252,   91,  253,
 /*  1740 */   252,   82,  253,  305,  252,  144,  305,  141,   22,  194,
 /*  1750 */   269,  257,  257,  153,  267,  143,  142,   25,  197,   26,
 /*  1760 */   242,  241,  196,  240,  242,  239,  238,   13,  188,  188,
 /*  1770 */     6,  293,  186,  186,  186,  200,  206,  293,  206,  206,
 /*  1780 */   206,    4,  200,  215,  215,  207,  207,    3,   22,  206,
 /*  1790 */   200,  158,   96,   15,   23,   16,   23,  135,  146,  126,
 /*  1800 */    24,  138,   20,   16,  140,    1,  138,  147,  126,   61,
 /*  1810 */    53,   37,  146,   53,   53,  290,   53,  126,  112,   34,
 /*  1820 */   137,    1,    5,   22,  111,  156,   68,   68,   26,   41,
 /*  1830 */    75,  137,  111,   24,   20,   19,  127,  121,   23,   67,
 /*  1840 */    22,   22,   67,   22,   22,   37,   22,   67,   23,  145,
 /*  1850 */    22,   28,   23,   23,   23,  137,   23,   22,   26,   22,
 /*  1860 */    24,   23,  112,   24,   23,   23,   22,  139,   34,   26,
 /*  1870 */    75,   88,   86,   75,   34,   23,   22,   24,   22,   34,
 /*  1880 */    34,   34,   93,   34,   26,   26,   23,   23,   23,   34,
 /*  1890 */    23,   23,   44,   23,   11,   26,   22,   22,   26,   23,
 /*  1900 */    23,   22,   22,   15,  137,  137,  137,  137,   23,    1,
 /*  1910 */   307,  307,  131,  307,  307,  307,  307,  307,  307,  307,
 /*  1920 */   307,  307,  307,  307,  307,  307,  307,  307,  307,  307,
 /*  1930 */   307,  307,  307,  307,  307,  307,  307,  307,  307,  307,
 /*  1940 */   307,  307,  307,  307,  307,  307,  307,  307,  307,  307,
 /*  1950 */   307,  307,  307,  307,  307,  307,  307,  307,  307,  307,
 /*  1960 */   307,  307,  307,  307,  307,  307,  307,  307,  307,  307,
 /*  1970 */   307,  307,  307,  307,  307,  307,  307,  307,  307,  307,
 /*  1980 */   307,  307,  307,  307,  307,  307,  307,  307,  307,  307,
 /*  1990 */   307,  307,  307,  307,  307,  307,  307,  307,  307,  307,
 /*  2000 */   307,  307,  307,  307,  307,  307,  307,  307,  307,  307,
 /*  2010 */   307,  307,  307,  307,  307,  307,  307,  307,  307,  307,
 /*  2020 */   307,  307,  307,  307,  307,  307,  307,  307,  307,  307,
 /*  2030 */   307,  307,  307,  307,  307,  307,  307,  307,  307,  307,
 /*  2040 */   307,  307,  307,  307,  307,  307,  307,  307,  307,  307,
 /*  2050 */   307,  307,  307,  307,  307,  307,  307,  307,  307,  307,
 /*  2060 */   307,  307,  307,  307,  307,  307,  307,  307,  307,  307,
 /*  2070 */   307,  307,  307,  307,  307,  307,  307,  307,  307,  307,
 /*  2080 */   307,  307,  307,  307,  307,  307,  307,  307,  307,  307,
 /*  2090 */   307,  307,
};
#define YY_SHIFT_COUNT    (542)
#define YY_SHIFT_MIN      (0)
#define YY_SHIFT_MAX      (1908)
static const unsigned short int yy_shift_ofst[] = {
 /*     0 */  1350, 1149, 1531,  939,  939,  548,  996, 1150, 1236, 1322,
 /*    10 */  1322, 1322,  334,    0,    0,  178,  777, 1322, 1322, 1322,
 /*    20 */  1322, 1322, 1322, 1322, 1322, 1322, 1322, 1322, 1322, 1322,
 /*    30 */   991,  991, 1125, 1125,  447,  597,  548,  548,  548,  548,
 /*    40 */   548,  548,   40,  108,  217,  284,  323,  390,  429,  496,
 /*    50 */   535,  602,  641,  757,  777,  777,  777,  777,  777,  777,
 /*    60 */   777,  777,  777,  777,  777,  777,  777,  777,  777,  777,
 /*    70 */   777,  777,  796,  777,  887,  900,  900, 1300, 1322, 1322,
 /*    80 */  1322, 1322, 1322, 1322, 1322, 1322, 1322, 1322, 1322, 1322,
 /*    90 */  1322, 1322, 1322, 1322, 1322, 1322, 1322, 1322, 1322, 1322,
 /*   100 */  1322, 1322, 1322, 1322, 1322, 1322, 1322, 1322, 1322, 1322,
 /*   110 */  1418, 1322, 1322, 1322, 1322, 1322, 1322, 1322, 1322, 1322,
 /*   120 */  1322, 1322, 1322, 1322,  147,  254,  254,  254,  254,  254,
 /*   130 */    84,  185,   66,  853,  958, 1121,  853,   92,   92,  853,
 /*   140 */   321,  321,  321,  321,  325,  350,  350,  461,  150, 1913,
 /*   150 */  1913,  285,  285,  285,  236,  184,  349,  184,  184,  712,
 /*   160 */   712,  433,  553,  771,  899,  853,  853,  853,  853,  853,
 /*   170 */   853,  853,  853,  853,  853,  853,  853,  853,  853,  853,
 /*   180 */   853,  853,  853,  853,  853,  853,   46,   46,  853,  113,
 /*   190 */   223,  223, 1183, 1183, 1127, 1142, 1913, 1913, 1913,  459,
 /*   200 */   514,  514,  653,  495,  657,  305,  705,  560,  622,  776,
 /*   210 */   853,  853,  853,  853,  853,  853,  853,  853,  853,  545,
 /*   220 */   853,  853,  853,  853,  853,  853,  853,  853,  853,  853,
 /*   230 */   853,  853, 1002, 1002, 1002,  853,  853,  853,  853, 1111,
 /*   240 */   853,  853,  853, 1006, 1109,  853,  853, 1168,  853,  853,
 /*   250 */   853,  853,  853,  853,  853,  853,  845, 1164,  738,  953,
 /*   260 */   953,  953,  953, 1196,  738,  738,   45,   96,  964,  179,
 /*   270 */   580,  907,  907, 1073,  580,  580, 1073,  498,  388, 1268,
 /*   280 */  1187, 1187, 1187,  907, 1170, 1170, 1058, 1180,  328, 1219,
 /*   290 */  1597, 1521, 1521, 1625, 1625, 1521, 1524, 1560, 1650, 1630,
 /*   300 */  1530, 1661, 1661, 1661, 1661, 1521, 1667, 1530, 1530, 1560,
 /*   310 */  1650, 1630, 1630, 1530, 1521, 1667, 1543, 1636, 1521, 1667,
 /*   320 */  1681, 1521, 1667, 1521, 1667, 1681, 1596, 1596, 1596, 1649,
 /*   330 */  1681, 1596, 1594, 1596, 1649, 1596, 1596, 1562, 1681, 1611,
 /*   340 */  1611, 1681, 1585, 1617, 1585, 1617, 1585, 1617, 1585, 1617,
 /*   350 */  1521, 1647, 1647, 1659, 1659, 1601, 1606, 1726, 1521, 1600,
 /*   360 */  1601, 1612, 1614, 1530, 1732, 1733, 1754, 1754, 1764, 1764,
 /*   370 */  1764, 1913, 1913, 1913, 1913, 1913, 1913, 1913, 1913, 1913,
 /*   380 */  1913, 1913, 1913, 1913, 1913, 1913,  673,  901,  283,  740,
 /*   390 */   707,  973,  655, 1247, 1048, 1097, 1190, 1306, 1263, 1383,
 /*   400 */  1395, 1432, 1469, 1473, 1497, 1279, 1200, 1323, 1075, 1286,
 /*   410 */  1536, 1608, 1332, 1609, 1175, 1225, 1610, 1615, 1309, 1361,
 /*   420 */  1777, 1784, 1766, 1633, 1778, 1696, 1779, 1771, 1773, 1662,
 /*   430 */  1652, 1673, 1776, 1663, 1782, 1664, 1787, 1804, 1668, 1660,
 /*   440 */  1682, 1748, 1774, 1666, 1757, 1760, 1761, 1763, 1691, 1706,
 /*   450 */  1785, 1683, 1820, 1817, 1801, 1713, 1669, 1758, 1802, 1759,
 /*   460 */  1755, 1788, 1694, 1721, 1809, 1814, 1816, 1709, 1716, 1818,
 /*   470 */  1772, 1819, 1821, 1815, 1822, 1775, 1823, 1824, 1780, 1808,
 /*   480 */  1825, 1704, 1828, 1829, 1830, 1831, 1832, 1833, 1835, 1836,
 /*   490 */  1838, 1837, 1839, 1718, 1841, 1842, 1750, 1834, 1844, 1728,
 /*   500 */  1843, 1840, 1845, 1846, 1847, 1783, 1795, 1786, 1848, 1798,
 /*   510 */  1789, 1849, 1852, 1854, 1853, 1858, 1859, 1855, 1863, 1843,
 /*   520 */  1864, 1865, 1867, 1868, 1869, 1870, 1856, 1883, 1874, 1875,
 /*   530 */  1876, 1877, 1879, 1880, 1872, 1781, 1767, 1768, 1769, 1770,
 /*   540 */  1885, 1888, 1908,
};
#define YY_REDUCE_COUNT (385)
#define YY_REDUCE_MIN   (-256)
#define YY_REDUCE_MAX   (1590)
static const short yy_reduce_ofst[] = {
 /*     0 */   880, -121,  269,  528,  933, -119, -187, -185, -182, -180,
 /*    10 */  -176, -174,  -62,  -46,  131, -248, -133,  407,  568,  700,
 /*    20 */   704,  278,  706,  824,  542,  830,  948,  773,  943,  946,
 /*    30 */    71,  650,  211,  267,  826,  272,  676,  732,  885,  976,
 /*    40 */   984,  992, -256, -256, -256, -256, -256, -256, -256, -256,
 /*    50 */  -256, -256, -256, -256, -256, -256, -256, -256, -256, -256,
 /*    60 */  -256, -256, -256, -256, -256, -256, -256, -256, -256, -256,
 /*    70 */  -256, -256, -256, -256, -256, -256, -256,  989, 1065, 1070,
 /*    80 */  1072, 1078, 1082, 1084, 1103, 1118, 1147, 1156, 1160, 1167,
 /*    90 */  1185, 1191, 1220, 1237, 1254, 1256, 1266, 1272, 1281, 1291,
 /*   100 */  1293, 1296, 1299, 1301, 1307, 1337, 1340, 1342, 1347, 1366,
 /*   110 */  1368, 1371, 1373, 1377, 1385, 1387, 1398, 1401, 1404, 1406,
 /*   120 */  1411, 1413, 1415, 1430, -256, -256, -256, -256, -256, -256,
 /*   130 */  -256, -256, -256, -172,  508, -213,   57, -163,  -25,  593,
 /*   140 */    69,  486,   69,  486, -200,  573,  722, -256, -256, -256,
 /*   150 */  -256, -141, -141, -141, -105, -161, -167,  157,  212,  405,
 /*   160 */   530,  220,  233,  735,  735,  115,  318,  406,  612,  541,
 /*   170 */  -166,  441,  688,  794,  629,  368,  741,  775,  867,  797,
 /*   180 */   871,  842, -186, 1000,  858,  949,  379,  783,   70,  296,
 /*   190 */   821,  903,  924, 1044,  651,  282, 1014, 1060,  937, -195,
 /*   200 */  -177,  413,  439,  511,  566,  787,  827,  848,  898,  945,
 /*   210 */  1062, 1074, 1102, 1110, 1202, 1204, 1209, 1211, 1215,  529,
 /*   220 */  1221, 1224, 1240, 1246, 1255, 1257, 1269, 1270, 1273, 1274,
 /*   230 */  1275, 1280, 1205, 1251, 1294, 1310, 1317, 1326, 1327, 1124,
 /*   240 */  1331, 1338, 1339, 1290, 1181, 1346, 1351, 1265, 1352,  787,
 /*   250 */  1353, 1367, 1378, 1386, 1392, 1397, 1241, 1312, 1356, 1345,
 /*   260 */  1357, 1358, 1359, 1124, 1356, 1356, 1364, 1396, 1433, 1341,
 /*   270 */  1381, 1376, 1379, 1354, 1391, 1405, 1362, 1429, 1423, 1431,
 /*   280 */  1434, 1435, 1437, 1399, 1410, 1412, 1382, 1417, 1420, 1466,
 /*   290 */  1372, 1467, 1468, 1380, 1384, 1475, 1394, 1414, 1416, 1448,
 /*   300 */  1440, 1451, 1452, 1453, 1454, 1490, 1493, 1449, 1455, 1427,
 /*   310 */  1436, 1464, 1465, 1456, 1498, 1502, 1419, 1421, 1507, 1509,
 /*   320 */  1491, 1510, 1513, 1514, 1516, 1496, 1500, 1501, 1503, 1499,
 /*   330 */  1505, 1504, 1508, 1506, 1511, 1512, 1515, 1424, 1517, 1457,
 /*   340 */  1460, 1519, 1474, 1482, 1483, 1485, 1486, 1488, 1489, 1492,
 /*   350 */  1541, 1438, 1441, 1494, 1495, 1518, 1520, 1487, 1555, 1481,
 /*   360 */  1522, 1523, 1526, 1528, 1561, 1566, 1580, 1581, 1586, 1587,
 /*   370 */  1588, 1478, 1484, 1525, 1575, 1570, 1572, 1573, 1574, 1582,
 /*   380 */  1568, 1569, 1578, 1579, 1583, 1590,
};
static const YYACTIONTYPE yy_default[] = {
 /*     0 */  1554, 1554, 1554, 1392, 1171, 1278, 1171, 1171, 1171, 1392,
 /*    10 */  1392, 1392, 1171, 1308, 1308, 1445, 1202, 1171, 1171, 1171,
 /*    20 */  1171, 1171, 1171, 1171, 1171, 1171, 1171, 1391, 1171, 1171,
 /*    30 */  1171, 1171, 1475, 1475, 1171, 1171, 1171, 1171, 1171, 1171,
 /*    40 */  1171, 1171, 1171, 1317, 1171, 1171, 1171, 1171, 1171, 1393,
 /*    50 */  1394, 1171, 1171, 1171, 1444, 1446, 1409, 1327, 1326, 1325,
 /*    60 */  1324, 1427, 1295, 1322, 1315, 1319, 1387, 1388, 1386, 1390,
 /*    70 */  1394, 1393, 1171, 1318, 1358, 1372, 1357, 1171, 1171, 1171,
 /*    80 */  1171, 1171, 1171, 1171, 1171, 1171, 1171, 1171, 1171, 1171,
 /*    90 */  1171, 1171, 1171, 1171, 1171, 1171, 1171, 1171, 1171, 1171,
 /*   100 */  1171, 1171, 1171, 1171, 1171, 1171, 1171, 1171, 1171, 1171,
 /*   110 */  1171, 1171, 1171, 1171, 1171, 1171, 1171, 1171, 1171, 1171,
 /*   120 */  1171, 1171, 1171, 1171, 1366, 1371, 1377, 1370, 1367, 1360,
 /*   130 */  1359, 1361, 1362, 1171, 1192, 1242, 1171, 1171, 1171, 1171,
 /*   140 */  1463, 1462, 1171, 1171, 1202, 1352, 1351, 1363, 1364, 1374,
 /*   150 */  1373, 1452, 1510, 1509, 1410, 1171, 1171, 1171, 1171, 1171,
 /*   160 */  1171, 1475, 1171, 1171, 1171, 1171, 1171, 1171, 1171, 1171,
 /*   170 */  1171, 1171, 1171, 1171, 1171, 1171, 1171, 1171, 1171, 1171,
 /*   180 */  1171, 1171, 1171, 1171, 1171, 1171, 1475, 1475, 1171, 1202,
 /*   190 */  1475, 1475, 1198, 1198, 1302, 1171, 1458, 1278, 1269, 1171,
 /*   200 */  1171, 1171, 1171, 1171, 1171, 1171, 1171, 1171, 1171, 1171,
 /*   210 */  1171, 1171, 1171, 1449, 1447, 1171, 1171, 1171, 1171, 1171,
 /*   220 */  1171, 1171, 1171, 1171, 1171, 1171, 1171, 1171, 1171, 1171,
 /*   230 */  1171, 1171, 1171, 1171, 1171, 1171, 1171, 1171, 1171, 1171,
 /*   240 */  1171, 1171, 1171, 1274, 1171, 1171, 1171, 1171, 1171, 1171,
 /*   250 */  1171, 1171, 1171, 1171, 1171, 1504, 1171, 1422, 1256, 1274,
 /*   260 */  1274, 1274, 1274, 1276, 1257, 1255, 1268, 1203, 1178, 1546,
 /*   270 */  1321, 1297, 1297, 1543, 1321, 1321, 1543, 1217, 1524, 1214,
 /*   280 */  1308, 1308, 1308, 1297, 1302, 1302, 1389, 1275, 1268, 1171,
 /*   290 */  1546, 1283, 1283, 1545, 1545, 1283, 1410, 1330, 1336, 1245,
 /*   300 */  1321, 1251, 1251, 1251, 1251, 1283, 1189, 1321, 1321, 1330,
 /*   310 */  1336, 1245, 1245, 1321, 1283, 1189, 1426, 1540, 1283, 1189,
 /*   320 */  1400, 1283, 1189, 1283, 1189, 1400, 1243, 1243, 1243, 1232,
 /*   330 */  1400, 1243, 1217, 1243, 1232, 1243, 1243, 1493, 1400, 1404,
 /*   340 */  1404, 1400, 1301, 1296, 1301, 1296, 1301, 1296, 1301, 1296,
 /*   350 */  1283, 1485, 1485, 1311, 1311, 1316, 1302, 1395, 1283, 1171,
 /*   360 */  1316, 1314, 1312, 1321, 1195, 1235, 1507, 1507, 1503, 1503,
 /*   370 */  1503, 1551, 1551, 1458, 1519, 1202, 1202, 1202, 1202, 1519,
 /*   380 */  1219, 1219, 1203, 1203, 1202, 1519, 1171, 1171, 1171, 1171,
 /*   390 */  1171, 1171, 1514, 1171, 1411, 1287, 1171, 1171, 1171, 1171,
 /*   400 */  1171, 1171, 1171, 1171, 1171, 1171, 1171, 1171, 1171, 1171,
 /*   410 */  1171, 1171, 1171, 1171, 1171, 1171, 1171, 1171, 1171, 1341,
 /*   420 */  1171, 1174, 1455, 1171, 1171, 1453, 1171, 1171, 1171, 1171,
 /*   430 */  1171, 1171, 1288, 1171, 1171, 1171, 1171, 1171, 1171, 1171,
 /*   440 */  1171, 1171, 1171, 1171, 1171, 1171, 1171, 1171, 1171, 1171,
 /*   450 */  1171, 1542, 1171, 1171, 1171, 1171, 1171, 1171, 1425, 1424,
 /*   460 */  1171, 1171, 1285, 1171, 1171, 1171, 1171, 1171, 1171, 1171,
 /*   470 */  1171, 1171, 1171, 1171, 1171, 1171, 1171, 1171, 1171, 1171,
 /*   480 */  1171, 1171, 1171, 1171, 1171, 1171, 1171, 1171, 1171, 1171,
 /*   490 */  1171, 1171, 1171, 1171, 1171, 1171, 1171, 1171, 1171, 1171,
 /*   500 */  1313, 1171, 1171, 1171, 1171, 1171, 1171, 1171, 1171, 1171,
 /*   510 */  1171, 1171, 1171, 1171, 1171, 1490, 1303, 1171, 1171, 1533,
 /*   520 */  1171, 1171, 1171, 1171, 1171, 1171, 1171, 1171, 1171, 1171,
 /*   530 */  1171, 1171, 1171, 1171, 1528, 1259, 1343, 1171, 1342, 1346,
 /*   540 */  1171, 1183, 1171,
};
/********** End of lemon-generated parsing tables *****************************/

/* The next table maps tokens (terminal symbols) into fallback tokens.  
** If a construct like the following:
** 
**      %fallback ID X Y Z.
**
** appears in the grammar, then ID becomes a fallback token for X, Y,
** and Z.  Whenever one of the tokens X, Y, or Z is input to the parser
** but it does not parse, the type of the token is changed to ID and
** the parse is retried before an error is thrown.
**
** This feature can be used, for example, to cause some keywords in a language
** to revert to identifiers if they keyword does not apply in the context where
** it appears.
*/
#ifdef YYFALLBACK
static const YYCODETYPE yyFallback[] = {
    0,  /*          $ => nothing */
    0,  /*       SEMI => nothing */
   59,  /*    EXPLAIN => ID */
   59,  /*      QUERY => ID */
   59,  /*       PLAN => ID */
   59,  /*      BEGIN => ID */
    0,  /* TRANSACTION => nothing */
   59,  /*   DEFERRED => ID */
   59,  /*  IMMEDIATE => ID */
   59,  /*  EXCLUSIVE => ID */
    0,  /*     COMMIT => nothing */
   59,  /*        END => ID */
   59,  /*   ROLLBACK => ID */
   59,  /*  SAVEPOINT => ID */
   59,  /*    RELEASE => ID */
    0,  /*         TO => nothing */
    0,  /*      TABLE => nothing */
    0,  /*     CREATE => nothing */
   59,  /*         IF => ID */
    0,  /*        NOT => nothing */
    0,  /*     EXISTS => nothing */
   59,  /*       TEMP => ID */
    0,  /*         LP => nothing */
    0,  /*         RP => nothing */
    0,  /*         AS => nothing */
   59,  /*    WITHOUT => ID */
    0,  /*      COMMA => nothing */
   59,  /*      ABORT => ID */
   59,  /*     ACTION => ID */
   59,  /*      AFTER => ID */
   59,  /*    ANALYZE => ID */
   59,  /*        ASC => ID */
   59,  /*     ATTACH => ID */
   59,  /*     BEFORE => ID */
   59,  /*         BY => ID */
   59,  /*    CASCADE => ID */
   59,  /*       CAST => ID */
   59,  /*   CONFLICT => ID */
   59,  /*   DATABASE => ID */
   59,  /*       DESC => ID */
   59,  /*     DETACH => ID */
   59,  /*       EACH => ID */
   59,  /*       FAIL => ID */
    0,  /*         OR => nothing */
    0,  /*        AND => nothing */
    0,  /*         IS => nothing */
   59,  /*      MATCH => ID */
   59,  /*    LIKE_KW => ID */
    0,  /*    BETWEEN => nothing */
    0,  /*         IN => nothing */
    0,  /*     ISNULL => nothing */
    0,  /*    NOTNULL => nothing */
    0,  /*         NE => nothing */
    0,  /*         EQ => nothing */
    0,  /*         GT => nothing */
    0,  /*         LE => nothing */
    0,  /*         LT => nothing */
    0,  /*         GE => nothing */
    0,  /*     ESCAPE => nothing */
    0,  /*         ID => nothing */
   59,  /*   COLUMNKW => ID */
   59,  /*         DO => ID */
   59,  /*        FOR => ID */
   59,  /*     IGNORE => ID */
   59,  /*  INITIALLY => ID */
   59,  /*    INSTEAD => ID */
   59,  /*         NO => ID */
   59,  /*        KEY => ID */
   59,  /*         OF => ID */
   59,  /*     OFFSET => ID */
   59,  /*     PRAGMA => ID */
   59,  /*      RAISE => ID */
   59,  /*  RECURSIVE => ID */
   59,  /*    REPLACE => ID */
   59,  /*   RESTRICT => ID */
   59,  /*        ROW => ID */
   59,  /*       ROWS => ID */
   59,  /*    TRIGGER => ID */
   59,  /*     VACUUM => ID */
   59,  /*       VIEW => ID */
   59,  /*    VIRTUAL => ID */
   59,  /*       WITH => ID */
   59,  /*      NULLS => ID */
   59,  /*      FIRST => ID */
   59,  /*       LAST => ID */
   59,  /*    CURRENT => ID */
   59,  /*  FOLLOWING => ID */
   59,  /*  PARTITION => ID */
   59,  /*  PRECEDING => ID */
   59,  /*      RANGE => ID */
   59,  /*  UNBOUNDED => ID */
   59,  /*    EXCLUDE => ID */
   59,  /*     GROUPS => ID */
   59,  /*     OTHERS => ID */
   59,  /*       TIES => ID */
   59,  /*    REINDEX => ID */
   59,  /*     RENAME => ID */
   59,  /*   CTIME_KW => ID */
    0,  /*        ANY => nothing */
    0,  /*     BITAND => nothing */
    0,  /*      BITOR => nothing */
    0,  /*     LSHIFT => nothing */
    0,  /*     RSHIFT => nothing */
    0,  /*       PLUS => nothing */
    0,  /*      MINUS => nothing */
    0,  /*       STAR => nothing */
    0,  /*      SLASH => nothing */
    0,  /*        REM => nothing */
    0,  /*     CONCAT => nothing */
    0,  /*    COLLATE => nothing */
    0,  /*     BITNOT => nothing */
    0,  /*         ON => nothing */
    0,  /*    INDEXED => nothing */
    0,  /*     STRING => nothing */
    0,  /*    JOIN_KW => nothing */
    0,  /* CONSTRAINT => nothing */
    0,  /*    DEFAULT => nothing */
    0,  /*       NULL => nothing */
    0,  /*    PRIMARY => nothing */
    0,  /*     UNIQUE => nothing */
    0,  /*      CHECK => nothing */
    0,  /* REFERENCES => nothing */
    0,  /*   AUTOINCR => nothing */
    0,  /*     INSERT => nothing */
    0,  /*     DELETE => nothing */
    0,  /*     UPDATE => nothing */
    0,  /*        SET => nothing */
    0,  /* DEFERRABLE => nothing */
    0,  /*    FOREIGN => nothing */
    0,  /*       DROP => nothing */
    0,  /*      UNION => nothing */
    0,  /*        ALL => nothing */
    0,  /*     EXCEPT => nothing */
    0,  /*  INTERSECT => nothing */
    0,  /*     SELECT => nothing */
    0,  /*     VALUES => nothing */
    0,  /*   DISTINCT => nothing */
    0,  /*        DOT => nothing */
    0,  /*       FROM => nothing */
    0,  /*       JOIN => nothing */
    0,  /*      USING => nothing */
    0,  /*      ORDER => nothing */
    0,  /*      GROUP => nothing */
    0,  /*     HAVING => nothing */
    0,  /*      LIMIT => nothing */
    0,  /*      WHERE => nothing */
    0,  /*       INTO => nothing */
    0,  /*    NOTHING => nothing */
    0,  /*      FLOAT => nothing */
    0,  /*       BLOB => nothing */
    0,  /*    INTEGER => nothing */
    0,  /*   VARIABLE => nothing */
    0,  /*       CASE => nothing */
    0,  /*       WHEN => nothing */
    0,  /*       THEN => nothing */
    0,  /*       ELSE => nothing */
    0,  /*      INDEX => nothing */
    0,  /*      ALTER => nothing */
    0,  /*        ADD => nothing */
    0,  /*     WINDOW => nothing */
    0,  /*       OVER => nothing */
    0,  /*     FILTER => nothing */
    0,  /*     COLUMN => nothing */
    0,  /* AGG_FUNCTION => nothing */
    0,  /* AGG_COLUMN => nothing */
    0,  /*  TRUEFALSE => nothing */
    0,  /*      ISNOT => nothing */
    0,  /*   FUNCTION => nothing */
    0,  /*     UMINUS => nothing */
    0,  /*      UPLUS => nothing */
    0,  /*      TRUTH => nothing */
    0,  /*   REGISTER => nothing */
    0,  /*     VECTOR => nothing */
    0,  /* SELECT_COLUMN => nothing */
    0,  /* IF_NULL_ROW => nothing */
    0,  /*   ASTERISK => nothing */
    0,  /*       SPAN => nothing */
    0,  /*      SPACE => nothing */
    0,  /*    ILLEGAL => nothing */
};
#endif /* YYFALLBACK */

/* The following structure represents a single element of the
** parser's stack.  Information stored includes:
**
**   +  The state number for the parser at this level of the stack.
**
**   +  The value of the token stored at this level of the stack.
**      (In other words, the "major" token.)
**
**   +  The semantic value stored at this level of the stack.  This is
**      the information used by the action routines in the grammar.
**      It is sometimes called the "minor" token.
**
** After the "shift" half of a SHIFTREDUCE action, the stateno field
** actually contains the reduce action for the second half of the
** SHIFTREDUCE.
*/
struct yyStackEntry {
  YYACTIONTYPE stateno;  /* The state-number, or reduce action in SHIFTREDUCE */
  YYCODETYPE major;      /* The major token value.  This is the code
                         ** number for the token at this stack level */
  YYMINORTYPE minor;     /* The user-supplied minor token value.  This
                         ** is the value of the token  */
};
typedef struct yyStackEntry yyStackEntry;

/* The state of the parser is completely contained in an instance of
** the following structure */
struct yyParser {
  yyStackEntry *yytos;          /* Pointer to top element of the stack */
#ifdef YYTRACKMAXSTACKDEPTH
  int yyhwm;                    /* High-water mark of the stack */
#endif
#ifndef YYNOERRORRECOVERY
  int yyerrcnt;                 /* Shifts left before out of the error */
#endif
  sqlite3ParserARG_SDECL                /* A place to hold %extra_argument */
  sqlite3ParserCTX_SDECL                /* A place to hold %extra_context */
#if YYSTACKDEPTH<=0
  int yystksz;                  /* Current side of the stack */
  yyStackEntry *yystack;        /* The parser's stack */
  yyStackEntry yystk0;          /* First stack entry */
#else
  yyStackEntry yystack[YYSTACKDEPTH];  /* The parser's stack */
  yyStackEntry *yystackEnd;            /* Last entry in the stack */
#endif
};
typedef struct yyParser yyParser;

#ifndef NDEBUG
#include <stdio.h>
static FILE *yyTraceFILE = 0;
static char *yyTracePrompt = 0;
#endif /* NDEBUG */

#ifndef NDEBUG
/* 
** Turn parser tracing on by giving a stream to which to write the trace
** and a prompt to preface each trace message.  Tracing is turned off
** by making either argument NULL 
**
** Inputs:
** <ul>
** <li> A FILE* to which trace output should be written.
**      If NULL, then tracing is turned off.
** <li> A prefix string written at the beginning of every
**      line of trace output.  If NULL, then tracing is
**      turned off.
** </ul>
**
** Outputs:
** None.
*/
void sqlite3ParserTrace(FILE *TraceFILE, char *zTracePrompt){
  yyTraceFILE = TraceFILE;
  yyTracePrompt = zTracePrompt;
  if( yyTraceFILE==0 ) yyTracePrompt = 0;
  else if( yyTracePrompt==0 ) yyTraceFILE = 0;
}
#endif /* NDEBUG */

#if defined(YYCOVERAGE) || !defined(NDEBUG)
/* For tracing shifts, the names of all terminals and nonterminals
** are required.  The following table supplies these names */
static const char *const yyTokenName[] = { 
  /*    0 */ "$",
  /*    1 */ "SEMI",
  /*    2 */ "EXPLAIN",
  /*    3 */ "QUERY",
  /*    4 */ "PLAN",
  /*    5 */ "BEGIN",
  /*    6 */ "TRANSACTION",
  /*    7 */ "DEFERRED",
  /*    8 */ "IMMEDIATE",
  /*    9 */ "EXCLUSIVE",
  /*   10 */ "COMMIT",
  /*   11 */ "END",
  /*   12 */ "ROLLBACK",
  /*   13 */ "SAVEPOINT",
  /*   14 */ "RELEASE",
  /*   15 */ "TO",
  /*   16 */ "TABLE",
  /*   17 */ "CREATE",
  /*   18 */ "IF",
  /*   19 */ "NOT",
  /*   20 */ "EXISTS",
  /*   21 */ "TEMP",
  /*   22 */ "LP",
  /*   23 */ "RP",
  /*   24 */ "AS",
  /*   25 */ "WITHOUT",
  /*   26 */ "COMMA",
  /*   27 */ "ABORT",
  /*   28 */ "ACTION",
  /*   29 */ "AFTER",
  /*   30 */ "ANALYZE",
  /*   31 */ "ASC",
  /*   32 */ "ATTACH",
  /*   33 */ "BEFORE",
  /*   34 */ "BY",
  /*   35 */ "CASCADE",
  /*   36 */ "CAST",
  /*   37 */ "CONFLICT",
  /*   38 */ "DATABASE",
  /*   39 */ "DESC",
  /*   40 */ "DETACH",
  /*   41 */ "EACH",
  /*   42 */ "FAIL",
  /*   43 */ "OR",
  /*   44 */ "AND",
  /*   45 */ "IS",
  /*   46 */ "MATCH",
  /*   47 */ "LIKE_KW",
  /*   48 */ "BETWEEN",
  /*   49 */ "IN",
  /*   50 */ "ISNULL",
  /*   51 */ "NOTNULL",
  /*   52 */ "NE",
  /*   53 */ "EQ",
  /*   54 */ "GT",
  /*   55 */ "LE",
  /*   56 */ "LT",
  /*   57 */ "GE",
  /*   58 */ "ESCAPE",
  /*   59 */ "ID",
  /*   60 */ "COLUMNKW",
  /*   61 */ "DO",
  /*   62 */ "FOR",
  /*   63 */ "IGNORE",
  /*   64 */ "INITIALLY",
  /*   65 */ "INSTEAD",
  /*   66 */ "NO",
  /*   67 */ "KEY",
  /*   68 */ "OF",
  /*   69 */ "OFFSET",
  /*   70 */ "PRAGMA",
  /*   71 */ "RAISE",
  /*   72 */ "RECURSIVE",
  /*   73 */ "REPLACE",
  /*   74 */ "RESTRICT",
  /*   75 */ "ROW",
  /*   76 */ "ROWS",
  /*   77 */ "TRIGGER",
  /*   78 */ "VACUUM",
  /*   79 */ "VIEW",
  /*   80 */ "VIRTUAL",
  /*   81 */ "WITH",
  /*   82 */ "NULLS",
  /*   83 */ "FIRST",
  /*   84 */ "LAST",
  /*   85 */ "CURRENT",
  /*   86 */ "FOLLOWING",
  /*   87 */ "PARTITION",
  /*   88 */ "PRECEDING",
  /*   89 */ "RANGE",
  /*   90 */ "UNBOUNDED",
  /*   91 */ "EXCLUDE",
  /*   92 */ "GROUPS",
  /*   93 */ "OTHERS",
  /*   94 */ "TIES",
  /*   95 */ "REINDEX",
  /*   96 */ "RENAME",
  /*   97 */ "CTIME_KW",
  /*   98 */ "ANY",
  /*   99 */ "BITAND",
  /*  100 */ "BITOR",
  /*  101 */ "LSHIFT",
  /*  102 */ "RSHIFT",
  /*  103 */ "PLUS",
  /*  104 */ "MINUS",
  /*  105 */ "STAR",
  /*  106 */ "SLASH",
  /*  107 */ "REM",
  /*  108 */ "CONCAT",
  /*  109 */ "COLLATE",
  /*  110 */ "BITNOT",
  /*  111 */ "ON",
  /*  112 */ "INDEXED",
  /*  113 */ "STRING",
  /*  114 */ "JOIN_KW",
  /*  115 */ "CONSTRAINT",
  /*  116 */ "DEFAULT",
  /*  117 */ "NULL",
  /*  118 */ "PRIMARY",
  /*  119 */ "UNIQUE",
  /*  120 */ "CHECK",
  /*  121 */ "REFERENCES",
  /*  122 */ "AUTOINCR",
  /*  123 */ "INSERT",
  /*  124 */ "DELETE",
  /*  125 */ "UPDATE",
  /*  126 */ "SET",
  /*  127 */ "DEFERRABLE",
  /*  128 */ "FOREIGN",
  /*  129 */ "DROP",
  /*  130 */ "UNION",
  /*  131 */ "ALL",
  /*  132 */ "EXCEPT",
  /*  133 */ "INTERSECT",
  /*  134 */ "SELECT",
  /*  135 */ "VALUES",
  /*  136 */ "DISTINCT",
  /*  137 */ "DOT",
  /*  138 */ "FROM",
  /*  139 */ "JOIN",
  /*  140 */ "USING",
  /*  141 */ "ORDER",
  /*  142 */ "GROUP",
  /*  143 */ "HAVING",
  /*  144 */ "LIMIT",
  /*  145 */ "WHERE",
  /*  146 */ "INTO",
  /*  147 */ "NOTHING",
  /*  148 */ "FLOAT",
  /*  149 */ "BLOB",
  /*  150 */ "INTEGER",
  /*  151 */ "VARIABLE",
  /*  152 */ "CASE",
  /*  153 */ "WHEN",
  /*  154 */ "THEN",
  /*  155 */ "ELSE",
  /*  156 */ "INDEX",
  /*  157 */ "ALTER",
  /*  158 */ "ADD",
  /*  159 */ "WINDOW",
  /*  160 */ "OVER",
  /*  161 */ "FILTER",
  /*  162 */ "COLUMN",
  /*  163 */ "AGG_FUNCTION",
  /*  164 */ "AGG_COLUMN",
  /*  165 */ "TRUEFALSE",
  /*  166 */ "ISNOT",
  /*  167 */ "FUNCTION",
  /*  168 */ "UMINUS",
  /*  169 */ "UPLUS",
  /*  170 */ "TRUTH",
  /*  171 */ "REGISTER",
  /*  172 */ "VECTOR",
  /*  173 */ "SELECT_COLUMN",
  /*  174 */ "IF_NULL_ROW",
  /*  175 */ "ASTERISK",
  /*  176 */ "SPAN",
  /*  177 */ "SPACE",
  /*  178 */ "ILLEGAL",
  /*  179 */ "input",
  /*  180 */ "cmdlist",
  /*  181 */ "ecmd",
  /*  182 */ "cmdx",
  /*  183 */ "explain",
  /*  184 */ "cmd",
  /*  185 */ "transtype",
  /*  186 */ "trans_opt",
  /*  187 */ "nm",
  /*  188 */ "savepoint_opt",
  /*  189 */ "create_table",
  /*  190 */ "create_table_args",
  /*  191 */ "createkw",
  /*  192 */ "temp",
  /*  193 */ "ifnotexists",
  /*  194 */ "dbnm",
  /*  195 */ "columnlist",
  /*  196 */ "conslist_opt",
  /*  197 */ "table_options",
  /*  198 */ "select",
  /*  199 */ "columnname",
  /*  200 */ "carglist",
  /*  201 */ "typetoken",
  /*  202 */ "typename",
  /*  203 */ "signed",
  /*  204 */ "plus_num",
  /*  205 */ "minus_num",
  /*  206 */ "scanpt",
  /*  207 */ "scantok",
  /*  208 */ "ccons",
  /*  209 */ "term",
  /*  210 */ "expr",
  /*  211 */ "onconf",
  /*  212 */ "sortorder",
  /*  213 */ "autoinc",
  /*  214 */ "eidlist_opt",
  /*  215 */ "refargs",
  /*  216 */ "defer_subclause",
  /*  217 */ "refarg",
  /*  218 */ "refact",
  /*  219 */ "init_deferred_pred_opt",
  /*  220 */ "conslist",
  /*  221 */ "tconscomma",
  /*  222 */ "tcons",
  /*  223 */ "sortlist",
  /*  224 */ "eidlist",
  /*  225 */ "defer_subclause_opt",
  /*  226 */ "orconf",
  /*  227 */ "resolvetype",
  /*  228 */ "raisetype",
  /*  229 */ "ifexists",
  /*  230 */ "fullname",
  /*  231 */ "selectnowith",
  /*  232 */ "oneselect",
  /*  233 */ "wqlist",
  /*  234 */ "multiselect_op",
  /*  235 */ "distinct",
  /*  236 */ "selcollist",
  /*  237 */ "from",
  /*  238 */ "where_opt",
  /*  239 */ "groupby_opt",
  /*  240 */ "having_opt",
  /*  241 */ "orderby_opt",
  /*  242 */ "limit_opt",
  /*  243 */ "window_clause",
  /*  244 */ "values",
  /*  245 */ "nexprlist",
  /*  246 */ "sclp",
  /*  247 */ "as",
  /*  248 */ "seltablist",
  /*  249 */ "stl_prefix",
  /*  250 */ "joinop",
  /*  251 */ "indexed_opt",
  /*  252 */ "on_opt",
  /*  253 */ "using_opt",
  /*  254 */ "exprlist",
  /*  255 */ "xfullname",
  /*  256 */ "idlist",
  /*  257 */ "nulls",
  /*  258 */ "with",
  /*  259 */ "setlist",
  /*  260 */ "insert_cmd",
  /*  261 */ "idlist_opt",
  /*  262 */ "upsert",
  /*  263 */ "filter_over",
  /*  264 */ "likeop",
  /*  265 */ "between_op",
  /*  266 */ "in_op",
  /*  267 */ "paren_exprlist",
  /*  268 */ "case_operand",
  /*  269 */ "case_exprlist",
  /*  270 */ "case_else",
  /*  271 */ "uniqueflag",
  /*  272 */ "collate",
  /*  273 */ "vinto",
  /*  274 */ "nmnum",
  /*  275 */ "trigger_decl",
  /*  276 */ "trigger_cmd_list",
  /*  277 */ "trigger_time",
  /*  278 */ "trigger_event",
  /*  279 */ "foreach_clause",
  /*  280 */ "when_clause",
  /*  281 */ "trigger_cmd",
  /*  282 */ "trnm",
  /*  283 */ "tridxby",
  /*  284 */ "database_kw_opt",
  /*  285 */ "key_opt",
  /*  286 */ "add_column_fullname",
  /*  287 */ "kwcolumn_opt",
  /*  288 */ "create_vtab",
  /*  289 */ "vtabarglist",
  /*  290 */ "vtabarg",
  /*  291 */ "vtabargtoken",
  /*  292 */ "lp",
  /*  293 */ "anylist",
  /*  294 */ "windowdefn_list",
  /*  295 */ "windowdefn",
  /*  296 */ "window",
  /*  297 */ "frame_opt",
  /*  298 */ "part_opt",
  /*  299 */ "filter_clause",
  /*  300 */ "over_clause",
  /*  301 */ "range_or_rows",
  /*  302 */ "frame_bound",
  /*  303 */ "frame_bound_s",
  /*  304 */ "frame_bound_e",
  /*  305 */ "frame_exclude_opt",
  /*  306 */ "frame_exclude",
};
#endif /* defined(YYCOVERAGE) || !defined(NDEBUG) */

#ifndef NDEBUG
/* For tracing reduce actions, the names of all rules are required.
*/
static const char *const yyRuleName[] = {
 /*   0 */ "explain ::= EXPLAIN",
 /*   1 */ "explain ::= EXPLAIN QUERY PLAN",
 /*   2 */ "cmdx ::= cmd",
 /*   3 */ "cmd ::= BEGIN transtype trans_opt",
 /*   4 */ "transtype ::=",
 /*   5 */ "transtype ::= DEFERRED",
 /*   6 */ "transtype ::= IMMEDIATE",
 /*   7 */ "transtype ::= EXCLUSIVE",
 /*   8 */ "cmd ::= COMMIT|END trans_opt",
 /*   9 */ "cmd ::= ROLLBACK trans_opt",
 /*  10 */ "cmd ::= SAVEPOINT nm",
 /*  11 */ "cmd ::= RELEASE savepoint_opt nm",
 /*  12 */ "cmd ::= ROLLBACK trans_opt TO savepoint_opt nm",
 /*  13 */ "create_table ::= createkw temp TABLE ifnotexists nm dbnm",
 /*  14 */ "createkw ::= CREATE",
 /*  15 */ "ifnotexists ::=",
 /*  16 */ "ifnotexists ::= IF NOT EXISTS",
 /*  17 */ "temp ::= TEMP",
 /*  18 */ "temp ::=",
 /*  19 */ "create_table_args ::= LP columnlist conslist_opt RP table_options",
 /*  20 */ "create_table_args ::= AS select",
 /*  21 */ "table_options ::=",
 /*  22 */ "table_options ::= WITHOUT nm",
 /*  23 */ "columnname ::= nm typetoken",
 /*  24 */ "typetoken ::=",
 /*  25 */ "typetoken ::= typename LP signed RP",
 /*  26 */ "typetoken ::= typename LP signed COMMA signed RP",
 /*  27 */ "typename ::= typename ID|STRING",
 /*  28 */ "scanpt ::=",
 /*  29 */ "scantok ::=",
 /*  30 */ "ccons ::= CONSTRAINT nm",
 /*  31 */ "ccons ::= DEFAULT scantok term",
 /*  32 */ "ccons ::= DEFAULT LP expr RP",
 /*  33 */ "ccons ::= DEFAULT PLUS scantok term",
 /*  34 */ "ccons ::= DEFAULT MINUS scantok term",
 /*  35 */ "ccons ::= DEFAULT scantok ID|INDEXED",
 /*  36 */ "ccons ::= NOT NULL onconf",
 /*  37 */ "ccons ::= PRIMARY KEY sortorder onconf autoinc",
 /*  38 */ "ccons ::= UNIQUE onconf",
 /*  39 */ "ccons ::= CHECK LP expr RP",
 /*  40 */ "ccons ::= REFERENCES nm eidlist_opt refargs",
 /*  41 */ "ccons ::= defer_subclause",
 /*  42 */ "ccons ::= COLLATE ID|STRING",
 /*  43 */ "autoinc ::=",
 /*  44 */ "autoinc ::= AUTOINCR",
 /*  45 */ "refargs ::=",
 /*  46 */ "refargs ::= refargs refarg",
 /*  47 */ "refarg ::= MATCH nm",
 /*  48 */ "refarg ::= ON INSERT refact",
 /*  49 */ "refarg ::= ON DELETE refact",
 /*  50 */ "refarg ::= ON UPDATE refact",
 /*  51 */ "refact ::= SET NULL",
 /*  52 */ "refact ::= SET DEFAULT",
 /*  53 */ "refact ::= CASCADE",
 /*  54 */ "refact ::= RESTRICT",
 /*  55 */ "refact ::= NO ACTION",
 /*  56 */ "defer_subclause ::= NOT DEFERRABLE init_deferred_pred_opt",
 /*  57 */ "defer_subclause ::= DEFERRABLE init_deferred_pred_opt",
 /*  58 */ "init_deferred_pred_opt ::=",
 /*  59 */ "init_deferred_pred_opt ::= INITIALLY DEFERRED",
 /*  60 */ "init_deferred_pred_opt ::= INITIALLY IMMEDIATE",
 /*  61 */ "conslist_opt ::=",
 /*  62 */ "tconscomma ::= COMMA",
 /*  63 */ "tcons ::= CONSTRAINT nm",
 /*  64 */ "tcons ::= PRIMARY KEY LP sortlist autoinc RP onconf",
 /*  65 */ "tcons ::= UNIQUE LP sortlist RP onconf",
 /*  66 */ "tcons ::= CHECK LP expr RP onconf",
 /*  67 */ "tcons ::= FOREIGN KEY LP eidlist RP REFERENCES nm eidlist_opt refargs defer_subclause_opt",
 /*  68 */ "defer_subclause_opt ::=",
 /*  69 */ "onconf ::=",
 /*  70 */ "onconf ::= ON CONFLICT resolvetype",
 /*  71 */ "orconf ::=",
 /*  72 */ "orconf ::= OR resolvetype",
 /*  73 */ "resolvetype ::= IGNORE",
 /*  74 */ "resolvetype ::= REPLACE",
 /*  75 */ "cmd ::= DROP TABLE ifexists fullname",
 /*  76 */ "ifexists ::= IF EXISTS",
 /*  77 */ "ifexists ::=",
 /*  78 */ "cmd ::= createkw temp VIEW ifnotexists nm dbnm eidlist_opt AS select",
 /*  79 */ "cmd ::= DROP VIEW ifexists fullname",
 /*  80 */ "cmd ::= select",
 /*  81 */ "select ::= WITH wqlist selectnowith",
 /*  82 */ "select ::= WITH RECURSIVE wqlist selectnowith",
 /*  83 */ "select ::= selectnowith",
 /*  84 */ "selectnowith ::= selectnowith multiselect_op oneselect",
 /*  85 */ "multiselect_op ::= UNION",
 /*  86 */ "multiselect_op ::= UNION ALL",
 /*  87 */ "multiselect_op ::= EXCEPT|INTERSECT",
 /*  88 */ "oneselect ::= SELECT distinct selcollist from where_opt groupby_opt having_opt orderby_opt limit_opt",
 /*  89 */ "oneselect ::= SELECT distinct selcollist from where_opt groupby_opt having_opt window_clause orderby_opt limit_opt",
 /*  90 */ "values ::= VALUES LP nexprlist RP",
 /*  91 */ "values ::= values COMMA LP nexprlist RP",
 /*  92 */ "distinct ::= DISTINCT",
 /*  93 */ "distinct ::= ALL",
 /*  94 */ "distinct ::=",
 /*  95 */ "sclp ::=",
 /*  96 */ "selcollist ::= sclp scanpt expr scanpt as",
 /*  97 */ "selcollist ::= sclp scanpt STAR",
 /*  98 */ "selcollist ::= sclp scanpt nm DOT STAR",
 /*  99 */ "as ::= AS nm",
 /* 100 */ "as ::=",
 /* 101 */ "from ::=",
 /* 102 */ "from ::= FROM seltablist",
 /* 103 */ "stl_prefix ::= seltablist joinop",
 /* 104 */ "stl_prefix ::=",
 /* 105 */ "seltablist ::= stl_prefix nm dbnm as indexed_opt on_opt using_opt",
 /* 106 */ "seltablist ::= stl_prefix nm dbnm LP exprlist RP as on_opt using_opt",
 /* 107 */ "seltablist ::= stl_prefix LP select RP as on_opt using_opt",
 /* 108 */ "seltablist ::= stl_prefix LP seltablist RP as on_opt using_opt",
 /* 109 */ "dbnm ::=",
 /* 110 */ "dbnm ::= DOT nm",
 /* 111 */ "fullname ::= nm",
 /* 112 */ "fullname ::= nm DOT nm",
 /* 113 */ "xfullname ::= nm",
 /* 114 */ "xfullname ::= nm DOT nm",
 /* 115 */ "xfullname ::= nm DOT nm AS nm",
 /* 116 */ "xfullname ::= nm AS nm",
 /* 117 */ "joinop ::= COMMA|JOIN",
 /* 118 */ "joinop ::= JOIN_KW JOIN",
 /* 119 */ "joinop ::= JOIN_KW nm JOIN",
 /* 120 */ "joinop ::= JOIN_KW nm nm JOIN",
 /* 121 */ "on_opt ::= ON expr",
 /* 122 */ "on_opt ::=",
 /* 123 */ "indexed_opt ::=",
 /* 124 */ "indexed_opt ::= INDEXED BY nm",
 /* 125 */ "indexed_opt ::= NOT INDEXED",
 /* 126 */ "using_opt ::= USING LP idlist RP",
 /* 127 */ "using_opt ::=",
 /* 128 */ "orderby_opt ::=",
 /* 129 */ "orderby_opt ::= ORDER BY sortlist",
 /* 130 */ "sortlist ::= sortlist COMMA expr sortorder nulls",
 /* 131 */ "sortlist ::= expr sortorder nulls",
 /* 132 */ "sortorder ::= ASC",
 /* 133 */ "sortorder ::= DESC",
 /* 134 */ "sortorder ::=",
 /* 135 */ "nulls ::= NULLS FIRST",
 /* 136 */ "nulls ::= NULLS LAST",
 /* 137 */ "nulls ::=",
 /* 138 */ "groupby_opt ::=",
 /* 139 */ "groupby_opt ::= GROUP BY nexprlist",
 /* 140 */ "having_opt ::=",
 /* 141 */ "having_opt ::= HAVING expr",
 /* 142 */ "limit_opt ::=",
 /* 143 */ "limit_opt ::= LIMIT expr",
 /* 144 */ "limit_opt ::= LIMIT expr OFFSET expr",
 /* 145 */ "limit_opt ::= LIMIT expr COMMA expr",
 /* 146 */ "cmd ::= with DELETE FROM xfullname indexed_opt where_opt",
 /* 147 */ "where_opt ::=",
 /* 148 */ "where_opt ::= WHERE expr",
 /* 149 */ "cmd ::= with UPDATE orconf xfullname indexed_opt SET setlist where_opt",
 /* 150 */ "setlist ::= setlist COMMA nm EQ expr",
 /* 151 */ "setlist ::= setlist COMMA LP idlist RP EQ expr",
 /* 152 */ "setlist ::= nm EQ expr",
 /* 153 */ "setlist ::= LP idlist RP EQ expr",
 /* 154 */ "cmd ::= with insert_cmd INTO xfullname idlist_opt select upsert",
 /* 155 */ "cmd ::= with insert_cmd INTO xfullname idlist_opt DEFAULT VALUES",
 /* 156 */ "upsert ::=",
 /* 157 */ "upsert ::= ON CONFLICT LP sortlist RP where_opt DO UPDATE SET setlist where_opt",
 /* 158 */ "upsert ::= ON CONFLICT LP sortlist RP where_opt DO NOTHING",
 /* 159 */ "upsert ::= ON CONFLICT DO NOTHING",
 /* 160 */ "insert_cmd ::= INSERT orconf",
 /* 161 */ "insert_cmd ::= REPLACE",
 /* 162 */ "idlist_opt ::=",
 /* 163 */ "idlist_opt ::= LP idlist RP",
 /* 164 */ "idlist ::= idlist COMMA nm",
 /* 165 */ "idlist ::= nm",
 /* 166 */ "expr ::= LP expr RP",
 /* 167 */ "expr ::= ID|INDEXED",
 /* 168 */ "expr ::= JOIN_KW",
 /* 169 */ "expr ::= nm DOT nm",
 /* 170 */ "expr ::= nm DOT nm DOT nm",
 /* 171 */ "term ::= NULL|FLOAT|BLOB",
 /* 172 */ "term ::= STRING",
 /* 173 */ "term ::= INTEGER",
 /* 174 */ "expr ::= VARIABLE",
 /* 175 */ "expr ::= expr COLLATE ID|STRING",
 /* 176 */ "expr ::= CAST LP expr AS typetoken RP",
 /* 177 */ "expr ::= ID|INDEXED LP distinct exprlist RP",
 /* 178 */ "expr ::= ID|INDEXED LP STAR RP",
 /* 179 */ "expr ::= ID|INDEXED LP distinct exprlist RP filter_over",
 /* 180 */ "expr ::= ID|INDEXED LP STAR RP filter_over",
 /* 181 */ "term ::= CTIME_KW",
 /* 182 */ "expr ::= LP nexprlist COMMA expr RP",
 /* 183 */ "expr ::= expr AND expr",
 /* 184 */ "expr ::= expr OR expr",
 /* 185 */ "expr ::= expr LT|GT|GE|LE expr",
 /* 186 */ "expr ::= expr EQ|NE expr",
 /* 187 */ "expr ::= expr BITAND|BITOR|LSHIFT|RSHIFT expr",
 /* 188 */ "expr ::= expr PLUS|MINUS expr",
 /* 189 */ "expr ::= expr STAR|SLASH|REM expr",
 /* 190 */ "expr ::= expr CONCAT expr",
 /* 191 */ "likeop ::= NOT LIKE_KW|MATCH",
 /* 192 */ "expr ::= expr likeop expr",
 /* 193 */ "expr ::= expr likeop expr ESCAPE expr",
 /* 194 */ "expr ::= expr ISNULL|NOTNULL",
 /* 195 */ "expr ::= expr NOT NULL",
 /* 196 */ "expr ::= expr IS expr",
 /* 197 */ "expr ::= expr IS NOT expr",
 /* 198 */ "expr ::= NOT expr",
 /* 199 */ "expr ::= BITNOT expr",
 /* 200 */ "expr ::= PLUS|MINUS expr",
 /* 201 */ "between_op ::= BETWEEN",
 /* 202 */ "between_op ::= NOT BETWEEN",
 /* 203 */ "expr ::= expr between_op expr AND expr",
 /* 204 */ "in_op ::= IN",
 /* 205 */ "in_op ::= NOT IN",
 /* 206 */ "expr ::= expr in_op LP exprlist RP",
 /* 207 */ "expr ::= LP select RP",
 /* 208 */ "expr ::= expr in_op LP select RP",
 /* 209 */ "expr ::= expr in_op nm dbnm paren_exprlist",
 /* 210 */ "expr ::= EXISTS LP select RP",
 /* 211 */ "expr ::= CASE case_operand case_exprlist case_else END",
 /* 212 */ "case_exprlist ::= case_exprlist WHEN expr THEN expr",
 /* 213 */ "case_exprlist ::= WHEN expr THEN expr",
 /* 214 */ "case_else ::= ELSE expr",
 /* 215 */ "case_else ::=",
 /* 216 */ "case_operand ::= expr",
 /* 217 */ "case_operand ::=",
 /* 218 */ "exprlist ::=",
 /* 219 */ "nexprlist ::= nexprlist COMMA expr",
 /* 220 */ "nexprlist ::= expr",
 /* 221 */ "paren_exprlist ::=",
 /* 222 */ "paren_exprlist ::= LP exprlist RP",
 /* 223 */ "cmd ::= createkw uniqueflag INDEX ifnotexists nm dbnm ON nm LP sortlist RP where_opt",
 /* 224 */ "uniqueflag ::= UNIQUE",
 /* 225 */ "uniqueflag ::=",
 /* 226 */ "eidlist_opt ::=",
 /* 227 */ "eidlist_opt ::= LP eidlist RP",
 /* 228 */ "eidlist ::= eidlist COMMA nm collate sortorder",
 /* 229 */ "eidlist ::= nm collate sortorder",
 /* 230 */ "collate ::=",
 /* 231 */ "collate ::= COLLATE ID|STRING",
 /* 232 */ "cmd ::= DROP INDEX ifexists fullname",
 /* 233 */ "cmd ::= VACUUM vinto",
 /* 234 */ "cmd ::= VACUUM nm vinto",
 /* 235 */ "vinto ::= INTO expr",
 /* 236 */ "vinto ::=",
 /* 237 */ "cmd ::= PRAGMA nm dbnm",
 /* 238 */ "cmd ::= PRAGMA nm dbnm EQ nmnum",
 /* 239 */ "cmd ::= PRAGMA nm dbnm LP nmnum RP",
 /* 240 */ "cmd ::= PRAGMA nm dbnm EQ minus_num",
 /* 241 */ "cmd ::= PRAGMA nm dbnm LP minus_num RP",
 /* 242 */ "plus_num ::= PLUS INTEGER|FLOAT",
 /* 243 */ "minus_num ::= MINUS INTEGER|FLOAT",
 /* 244 */ "cmd ::= createkw trigger_decl BEGIN trigger_cmd_list END",
 /* 245 */ "trigger_decl ::= temp TRIGGER ifnotexists nm dbnm trigger_time trigger_event ON fullname foreach_clause when_clause",
 /* 246 */ "trigger_time ::= BEFORE|AFTER",
 /* 247 */ "trigger_time ::= INSTEAD OF",
 /* 248 */ "trigger_time ::=",
 /* 249 */ "trigger_event ::= DELETE|INSERT",
 /* 250 */ "trigger_event ::= UPDATE",
 /* 251 */ "trigger_event ::= UPDATE OF idlist",
 /* 252 */ "when_clause ::=",
 /* 253 */ "when_clause ::= WHEN expr",
 /* 254 */ "trigger_cmd_list ::= trigger_cmd_list trigger_cmd SEMI",
 /* 255 */ "trigger_cmd_list ::= trigger_cmd SEMI",
 /* 256 */ "trnm ::= nm DOT nm",
 /* 257 */ "tridxby ::= INDEXED BY nm",
 /* 258 */ "tridxby ::= NOT INDEXED",
 /* 259 */ "trigger_cmd ::= UPDATE orconf trnm tridxby SET setlist where_opt scanpt",
 /* 260 */ "trigger_cmd ::= scanpt insert_cmd INTO trnm idlist_opt select upsert scanpt",
 /* 261 */ "trigger_cmd ::= DELETE FROM trnm tridxby where_opt scanpt",
 /* 262 */ "trigger_cmd ::= scanpt select scanpt",
 /* 263 */ "expr ::= RAISE LP IGNORE RP",
 /* 264 */ "expr ::= RAISE LP raisetype COMMA nm RP",
 /* 265 */ "raisetype ::= ROLLBACK",
 /* 266 */ "raisetype ::= ABORT",
 /* 267 */ "raisetype ::= FAIL",
 /* 268 */ "cmd ::= DROP TRIGGER ifexists fullname",
 /* 269 */ "cmd ::= ATTACH database_kw_opt expr AS expr key_opt",
 /* 270 */ "cmd ::= DETACH database_kw_opt expr",
 /* 271 */ "key_opt ::=",
 /* 272 */ "key_opt ::= KEY expr",
 /* 273 */ "cmd ::= REINDEX",
 /* 274 */ "cmd ::= REINDEX nm dbnm",
 /* 275 */ "cmd ::= ANALYZE",
 /* 276 */ "cmd ::= ANALYZE nm dbnm",
 /* 277 */ "cmd ::= ALTER TABLE fullname RENAME TO nm",
 /* 278 */ "cmd ::= ALTER TABLE add_column_fullname ADD kwcolumn_opt columnname carglist",
 /* 279 */ "add_column_fullname ::= fullname",
 /* 280 */ "cmd ::= ALTER TABLE fullname RENAME kwcolumn_opt nm TO nm",
 /* 281 */ "cmd ::= create_vtab",
 /* 282 */ "cmd ::= create_vtab LP vtabarglist RP",
 /* 283 */ "create_vtab ::= createkw VIRTUAL TABLE ifnotexists nm dbnm USING nm",
 /* 284 */ "vtabarg ::=",
 /* 285 */ "vtabargtoken ::= ANY",
 /* 286 */ "vtabargtoken ::= lp anylist RP",
 /* 287 */ "lp ::= LP",
 /* 288 */ "with ::= WITH wqlist",
 /* 289 */ "with ::= WITH RECURSIVE wqlist",
 /* 290 */ "wqlist ::= nm eidlist_opt AS LP select RP",
 /* 291 */ "wqlist ::= wqlist COMMA nm eidlist_opt AS LP select RP",
 /* 292 */ "windowdefn_list ::= windowdefn",
 /* 293 */ "windowdefn_list ::= windowdefn_list COMMA windowdefn",
 /* 294 */ "windowdefn ::= nm AS LP window RP",
 /* 295 */ "window ::= PARTITION BY nexprlist orderby_opt frame_opt",
 /* 296 */ "window ::= nm PARTITION BY nexprlist orderby_opt frame_opt",
 /* 297 */ "window ::= ORDER BY sortlist frame_opt",
 /* 298 */ "window ::= nm ORDER BY sortlist frame_opt",
 /* 299 */ "window ::= frame_opt",
 /* 300 */ "window ::= nm frame_opt",
 /* 301 */ "frame_opt ::=",
 /* 302 */ "frame_opt ::= range_or_rows frame_bound_s frame_exclude_opt",
 /* 303 */ "frame_opt ::= range_or_rows BETWEEN frame_bound_s AND frame_bound_e frame_exclude_opt",
 /* 304 */ "range_or_rows ::= RANGE|ROWS|GROUPS",
 /* 305 */ "frame_bound_s ::= frame_bound",
 /* 306 */ "frame_bound_s ::= UNBOUNDED PRECEDING",
 /* 307 */ "frame_bound_e ::= frame_bound",
 /* 308 */ "frame_bound_e ::= UNBOUNDED FOLLOWING",
 /* 309 */ "frame_bound ::= expr PRECEDING|FOLLOWING",
 /* 310 */ "frame_bound ::= CURRENT ROW",
 /* 311 */ "frame_exclude_opt ::=",
 /* 312 */ "frame_exclude_opt ::= EXCLUDE frame_exclude",
 /* 313 */ "frame_exclude ::= NO OTHERS",
 /* 314 */ "frame_exclude ::= CURRENT ROW",
 /* 315 */ "frame_exclude ::= GROUP|TIES",
 /* 316 */ "window_clause ::= WINDOW windowdefn_list",
 /* 317 */ "filter_over ::= filter_clause over_clause",
 /* 318 */ "filter_over ::= over_clause",
 /* 319 */ "filter_over ::= filter_clause",
 /* 320 */ "over_clause ::= OVER LP window RP",
 /* 321 */ "over_clause ::= OVER nm",
 /* 322 */ "filter_clause ::= FILTER LP WHERE expr RP",
 /* 323 */ "input ::= cmdlist",
 /* 324 */ "cmdlist ::= cmdlist ecmd",
 /* 325 */ "cmdlist ::= ecmd",
 /* 326 */ "ecmd ::= SEMI",
 /* 327 */ "ecmd ::= cmdx SEMI",
 /* 328 */ "ecmd ::= explain cmdx",
 /* 329 */ "trans_opt ::=",
 /* 330 */ "trans_opt ::= TRANSACTION",
 /* 331 */ "trans_opt ::= TRANSACTION nm",
 /* 332 */ "savepoint_opt ::= SAVEPOINT",
 /* 333 */ "savepoint_opt ::=",
 /* 334 */ "cmd ::= create_table create_table_args",
 /* 335 */ "columnlist ::= columnlist COMMA columnname carglist",
 /* 336 */ "columnlist ::= columnname carglist",
 /* 337 */ "nm ::= ID|INDEXED",
 /* 338 */ "nm ::= STRING",
 /* 339 */ "nm ::= JOIN_KW",
 /* 340 */ "typetoken ::= typename",
 /* 341 */ "typename ::= ID|STRING",
 /* 342 */ "signed ::= plus_num",
 /* 343 */ "signed ::= minus_num",
 /* 344 */ "carglist ::= carglist ccons",
 /* 345 */ "carglist ::=",
 /* 346 */ "ccons ::= NULL onconf",
 /* 347 */ "conslist_opt ::= COMMA conslist",
 /* 348 */ "conslist ::= conslist tconscomma tcons",
 /* 349 */ "conslist ::= tcons",
 /* 350 */ "tconscomma ::=",
 /* 351 */ "defer_subclause_opt ::= defer_subclause",
 /* 352 */ "resolvetype ::= raisetype",
 /* 353 */ "selectnowith ::= oneselect",
 /* 354 */ "oneselect ::= values",
 /* 355 */ "sclp ::= selcollist COMMA",
 /* 356 */ "as ::= ID|STRING",
 /* 357 */ "expr ::= term",
 /* 358 */ "likeop ::= LIKE_KW|MATCH",
 /* 359 */ "exprlist ::= nexprlist",
 /* 360 */ "nmnum ::= plus_num",
 /* 361 */ "nmnum ::= nm",
 /* 362 */ "nmnum ::= ON",
 /* 363 */ "nmnum ::= DELETE",
 /* 364 */ "nmnum ::= DEFAULT",
 /* 365 */ "plus_num ::= INTEGER|FLOAT",
 /* 366 */ "foreach_clause ::=",
 /* 367 */ "foreach_clause ::= FOR EACH ROW",
 /* 368 */ "trnm ::= nm",
 /* 369 */ "tridxby ::=",
 /* 370 */ "database_kw_opt ::= DATABASE",
 /* 371 */ "database_kw_opt ::=",
 /* 372 */ "kwcolumn_opt ::=",
 /* 373 */ "kwcolumn_opt ::= COLUMNKW",
 /* 374 */ "vtabarglist ::= vtabarg",
 /* 375 */ "vtabarglist ::= vtabarglist COMMA vtabarg",
 /* 376 */ "vtabarg ::= vtabarg vtabargtoken",
 /* 377 */ "anylist ::=",
 /* 378 */ "anylist ::= anylist LP anylist RP",
 /* 379 */ "anylist ::= anylist ANY",
 /* 380 */ "with ::=",
};
#endif /* NDEBUG */


#if YYSTACKDEPTH<=0
/*
** Try to increase the size of the parser stack.  Return the number
** of errors.  Return 0 on success.
*/
static int yyGrowStack(yyParser *p){
  int newSize;
  int idx;
  yyStackEntry *pNew;

  newSize = p->yystksz*2 + 100;
  idx = p->yytos ? (int)(p->yytos - p->yystack) : 0;
  if( p->yystack==&p->yystk0 ){
    pNew = malloc(newSize*sizeof(pNew[0]));
    if( pNew ) pNew[0] = p->yystk0;
  }else{
    pNew = realloc(p->yystack, newSize*sizeof(pNew[0]));
  }
  if( pNew ){
    p->yystack = pNew;
    p->yytos = &p->yystack[idx];
#ifndef NDEBUG
    if( yyTraceFILE ){
      fprintf(yyTraceFILE,"%sStack grows from %d to %d entries.\n",
              yyTracePrompt, p->yystksz, newSize);
    }
#endif
    p->yystksz = newSize;
  }
  return pNew==0; 
}
#endif

/* Datatype of the argument to the memory allocated passed as the
** second argument to sqlite3ParserAlloc() below.  This can be changed by
** putting an appropriate #define in the %include section of the input
** grammar.
*/
#ifndef YYMALLOCARGTYPE
# define YYMALLOCARGTYPE size_t
#endif

/* Initialize a new parser that has already been allocated.
*/
void sqlite3ParserInit(void *yypRawParser sqlite3ParserCTX_PDECL){
  yyParser *yypParser = (yyParser*)yypRawParser;
  sqlite3ParserCTX_STORE
#ifdef YYTRACKMAXSTACKDEPTH
  yypParser->yyhwm = 0;
#endif
#if YYSTACKDEPTH<=0
  yypParser->yytos = NULL;
  yypParser->yystack = NULL;
  yypParser->yystksz = 0;
  if( yyGrowStack(yypParser) ){
    yypParser->yystack = &yypParser->yystk0;
    yypParser->yystksz = 1;
  }
#endif
#ifndef YYNOERRORRECOVERY
  yypParser->yyerrcnt = -1;
#endif
  yypParser->yytos = yypParser->yystack;
  yypParser->yystack[0].stateno = 0;
  yypParser->yystack[0].major = 0;
#if YYSTACKDEPTH>0
  yypParser->yystackEnd = &yypParser->yystack[YYSTACKDEPTH-1];
#endif
}

#ifndef sqlite3Parser_ENGINEALWAYSONSTACK
/* 
** This function allocates a new parser.
** The only argument is a pointer to a function which works like
** malloc.
**
** Inputs:
** A pointer to the function used to allocate memory.
**
** Outputs:
** A pointer to a parser.  This pointer is used in subsequent calls
** to sqlite3Parser and sqlite3ParserFree.
*/
void *sqlite3ParserAlloc(void *(*mallocProc)(YYMALLOCARGTYPE) sqlite3ParserCTX_PDECL){
  yyParser *yypParser;
  yypParser = (yyParser*)(*mallocProc)( (YYMALLOCARGTYPE)sizeof(yyParser) );
  if( yypParser ){
    sqlite3ParserCTX_STORE
    sqlite3ParserInit(yypParser sqlite3ParserCTX_PARAM);
  }
  return (void*)yypParser;
}
#endif /* sqlite3Parser_ENGINEALWAYSONSTACK */


/* The following function deletes the "minor type" or semantic value
** associated with a symbol.  The symbol can be either a terminal
** or nonterminal. "yymajor" is the symbol code, and "yypminor" is
** a pointer to the value to be deleted.  The code used to do the 
** deletions is derived from the %destructor and/or %token_destructor
** directives of the input grammar.
*/
static void yy_destructor(
  yyParser *yypParser,    /* The parser */
  YYCODETYPE yymajor,     /* Type code for object to destroy */
  YYMINORTYPE *yypminor   /* The object to be destroyed */
){
  sqlite3ParserARG_FETCH
  sqlite3ParserCTX_FETCH
  switch( yymajor ){
    /* Here is inserted the actions which take place when a
    ** terminal or non-terminal is destroyed.  This can happen
    ** when the symbol is popped from the stack during a
    ** reduce or during error processing or when a parser is 
    ** being destroyed before it is finished parsing.
    **
    ** Note: during a reduce, the only symbols destroyed are those
    ** which appear on the RHS of the rule, but which are *not* used
    ** inside the C code.
    */
/********* Begin destructor definitions ***************************************/
    case 198: /* select */
    case 231: /* selectnowith */
    case 232: /* oneselect */
    case 244: /* values */
{
#line 449 "parse.y"
sqlite3SelectDelete(pParse->db, (yypminor->yy25));
#line 2051 "parse.c"
}
      break;
    case 209: /* term */
    case 210: /* expr */
    case 238: /* where_opt */
    case 240: /* having_opt */
    case 252: /* on_opt */
    case 268: /* case_operand */
    case 270: /* case_else */
    case 273: /* vinto */
    case 280: /* when_clause */
    case 285: /* key_opt */
    case 299: /* filter_clause */
{
#line 945 "parse.y"
sqlite3ExprDelete(pParse->db, (yypminor->yy46));
#line 2068 "parse.c"
}
      break;
    case 214: /* eidlist_opt */
    case 223: /* sortlist */
    case 224: /* eidlist */
    case 236: /* selcollist */
    case 239: /* groupby_opt */
    case 241: /* orderby_opt */
    case 245: /* nexprlist */
    case 246: /* sclp */
    case 254: /* exprlist */
    case 259: /* setlist */
    case 267: /* paren_exprlist */
    case 269: /* case_exprlist */
    case 298: /* part_opt */
{
#line 1307 "parse.y"
sqlite3ExprListDelete(pParse->db, (yypminor->yy138));
#line 2087 "parse.c"
}
      break;
    case 230: /* fullname */
    case 237: /* from */
    case 248: /* seltablist */
    case 249: /* stl_prefix */
    case 255: /* xfullname */
{
#line 696 "parse.y"
sqlite3SrcListDelete(pParse->db, (yypminor->yy609));
#line 2098 "parse.c"
}
      break;
    case 233: /* wqlist */
{
#line 1590 "parse.y"
sqlite3WithDelete(pParse->db, (yypminor->yy297));
#line 2105 "parse.c"
}
      break;
    case 243: /* window_clause */
    case 294: /* windowdefn_list */
{
#line 1710 "parse.y"
sqlite3WindowListDelete(pParse->db, (yypminor->yy455));
#line 2113 "parse.c"
}
      break;
    case 253: /* using_opt */
    case 256: /* idlist */
    case 261: /* idlist_opt */
{
#line 768 "parse.y"
sqlite3IdListDelete(pParse->db, (yypminor->yy406));
#line 2122 "parse.c"
}
      break;
    case 263: /* filter_over */
    case 295: /* windowdefn */
    case 296: /* window */
    case 297: /* frame_opt */
    case 300: /* over_clause */
{
#line 1647 "parse.y"
sqlite3WindowDelete(pParse->db, (yypminor->yy455));
#line 2133 "parse.c"
}
      break;
    case 276: /* trigger_cmd_list */
    case 281: /* trigger_cmd */
{
#line 1424 "parse.y"
sqlite3DeleteTriggerStep(pParse->db, (yypminor->yy527));
#line 2141 "parse.c"
}
      break;
    case 278: /* trigger_event */
{
#line 1410 "parse.y"
sqlite3IdListDelete(pParse->db, (yypminor->yy572).b);
#line 2148 "parse.c"
}
      break;
    case 302: /* frame_bound */
    case 303: /* frame_bound_s */
    case 304: /* frame_bound_e */
{
#line 1652 "parse.y"
sqlite3ExprDelete(pParse->db, (yypminor->yy57).pExpr);
#line 2157 "parse.c"
}
      break;
/********* End destructor definitions *****************************************/
    default:  break;   /* If no destructor action specified: do nothing */
  }
}

/*
** Pop the parser's stack once.
**
** If there is a destructor routine associated with the token which
** is popped from the stack, then call it.
*/
static void yy_pop_parser_stack(yyParser *pParser){
  yyStackEntry *yytos;
  assert( pParser->yytos!=0 );
  assert( pParser->yytos > pParser->yystack );
  yytos = pParser->yytos--;
#ifndef NDEBUG
  if( yyTraceFILE ){
    fprintf(yyTraceFILE,"%sPopping %s\n",
      yyTracePrompt,
      yyTokenName[yytos->major]);
  }
#endif
  yy_destructor(pParser, yytos->major, &yytos->minor);
}

/*
** Clear all secondary memory allocations from the parser
*/
void sqlite3ParserFinalize(void *p){
  yyParser *pParser = (yyParser*)p;
  while( pParser->yytos>pParser->yystack ) yy_pop_parser_stack(pParser);
#if YYSTACKDEPTH<=0
  if( pParser->yystack!=&pParser->yystk0 ) free(pParser->yystack);
#endif
}

#ifndef sqlite3Parser_ENGINEALWAYSONSTACK
/* 
** Deallocate and destroy a parser.  Destructors are called for
** all stack elements before shutting the parser down.
**
** If the YYPARSEFREENEVERNULL macro exists (for example because it
** is defined in a %include section of the input grammar) then it is
** assumed that the input pointer is never NULL.
*/
void sqlite3ParserFree(
  void *p,                    /* The parser to be deleted */
  void (*freeProc)(void*)     /* Function used to reclaim memory */
){
#ifndef YYPARSEFREENEVERNULL
  if( p==0 ) return;
#endif
  sqlite3ParserFinalize(p);
  (*freeProc)(p);
}
#endif /* sqlite3Parser_ENGINEALWAYSONSTACK */

/*
** Return the peak depth of the stack for a parser.
*/
#ifdef YYTRACKMAXSTACKDEPTH
int sqlite3ParserStackPeak(void *p){
  yyParser *pParser = (yyParser*)p;
  return pParser->yyhwm;
}
#endif

/* This array of booleans keeps track of the parser statement
** coverage.  The element yycoverage[X][Y] is set when the parser
** is in state X and has a lookahead token Y.  In a well-tested
** systems, every element of this matrix should end up being set.
*/
#if defined(YYCOVERAGE)
static unsigned char yycoverage[YYNSTATE][YYNTOKEN];
#endif

/*
** Write into out a description of every state/lookahead combination that
**
**   (1)  has not been used by the parser, and
**   (2)  is not a syntax error.
**
** Return the number of missed state/lookahead combinations.
*/
#if defined(YYCOVERAGE)
int sqlite3ParserCoverage(FILE *out){
  int stateno, iLookAhead, i;
  int nMissed = 0;
  for(stateno=0; stateno<YYNSTATE; stateno++){
    i = yy_shift_ofst[stateno];
    for(iLookAhead=0; iLookAhead<YYNTOKEN; iLookAhead++){
      if( yy_lookahead[i+iLookAhead]!=iLookAhead ) continue;
      if( yycoverage[stateno][iLookAhead]==0 ) nMissed++;
      if( out ){
        fprintf(out,"State %d lookahead %s %s\n", stateno,
                yyTokenName[iLookAhead],
                yycoverage[stateno][iLookAhead] ? "ok" : "missed");
      }
    }
  }
  return nMissed;
}
#endif

/*
** Find the appropriate action for a parser given the terminal
** look-ahead token iLookAhead.
*/
static YYACTIONTYPE yy_find_shift_action(
  YYCODETYPE iLookAhead,    /* The look-ahead token */
  YYACTIONTYPE stateno      /* Current state number */
){
  int i;

  if( stateno>YY_MAX_SHIFT ) return stateno;
  assert( stateno <= YY_SHIFT_COUNT );
#if defined(YYCOVERAGE)
  yycoverage[stateno][iLookAhead] = 1;
#endif
  do{
    i = yy_shift_ofst[stateno];
    assert( i>=0 );
    assert( i<=YY_ACTTAB_COUNT );
    assert( i+YYNTOKEN<=(int)YY_NLOOKAHEAD );
    assert( iLookAhead!=YYNOCODE );
    assert( iLookAhead < YYNTOKEN );
    i += iLookAhead;
    assert( i<(int)YY_NLOOKAHEAD );
    if( yy_lookahead[i]!=iLookAhead ){
#ifdef YYFALLBACK
      YYCODETYPE iFallback;            /* Fallback token */
      assert( iLookAhead<sizeof(yyFallback)/sizeof(yyFallback[0]) );
      iFallback = yyFallback[iLookAhead];
      if( iFallback!=0 ){
#ifndef NDEBUG
        if( yyTraceFILE ){
          fprintf(yyTraceFILE, "%sFALLBACK %s => %s\n",
             yyTracePrompt, yyTokenName[iLookAhead], yyTokenName[iFallback]);
        }
#endif
        assert( yyFallback[iFallback]==0 ); /* Fallback loop must terminate */
        iLookAhead = iFallback;
        continue;
      }
#endif
#ifdef YYWILDCARD
      {
        int j = i - iLookAhead + YYWILDCARD;
        assert( j<(int)(sizeof(yy_lookahead)/sizeof(yy_lookahead[0])) );
        if( yy_lookahead[j]==YYWILDCARD && iLookAhead>0 ){
#ifndef NDEBUG
          if( yyTraceFILE ){
            fprintf(yyTraceFILE, "%sWILDCARD %s => %s\n",
               yyTracePrompt, yyTokenName[iLookAhead],
               yyTokenName[YYWILDCARD]);
          }
#endif /* NDEBUG */
          return yy_action[j];
        }
      }
#endif /* YYWILDCARD */
      return yy_default[stateno];
    }else{
      assert( i>=0 && i<sizeof(yy_action)/sizeof(yy_action[0]) );
      return yy_action[i];
    }
  }while(1);
}

/*
** Find the appropriate action for a parser given the non-terminal
** look-ahead token iLookAhead.
*/
static YYACTIONTYPE yy_find_reduce_action(
  YYACTIONTYPE stateno,     /* Current state number */
  YYCODETYPE iLookAhead     /* The look-ahead token */
){
  int i;
#ifdef YYERRORSYMBOL
  if( stateno>YY_REDUCE_COUNT ){
    return yy_default[stateno];
  }
#else
  assert( stateno<=YY_REDUCE_COUNT );
#endif
  i = yy_reduce_ofst[stateno];
  assert( iLookAhead!=YYNOCODE );
  i += iLookAhead;
#ifdef YYERRORSYMBOL
  if( i<0 || i>=YY_ACTTAB_COUNT || yy_lookahead[i]!=iLookAhead ){
    return yy_default[stateno];
  }
#else
  assert( i>=0 && i<YY_ACTTAB_COUNT );
  assert( yy_lookahead[i]==iLookAhead );
#endif
  return yy_action[i];
}

/*
** The following routine is called if the stack overflows.
*/
static void yyStackOverflow(yyParser *yypParser){
   sqlite3ParserARG_FETCH
   sqlite3ParserCTX_FETCH
#ifndef NDEBUG
   if( yyTraceFILE ){
     fprintf(yyTraceFILE,"%sStack Overflow!\n",yyTracePrompt);
   }
#endif
   while( yypParser->yytos>yypParser->yystack ) yy_pop_parser_stack(yypParser);
   /* Here code is inserted which will execute if the parser
   ** stack every overflows */
/******** Begin %stack_overflow code ******************************************/
#line 41 "parse.y"

  sqlite3ErrorMsg(pParse, "parser stack overflow");
#line 2378 "parse.c"
/******** End %stack_overflow code ********************************************/
   sqlite3ParserARG_STORE /* Suppress warning about unused %extra_argument var */
   sqlite3ParserCTX_STORE
}

/*
** Print tracing information for a SHIFT action
*/
#ifndef NDEBUG
static void yyTraceShift(yyParser *yypParser, int yyNewState, const char *zTag){
  if( yyTraceFILE ){
    if( yyNewState<YYNSTATE ){
      fprintf(yyTraceFILE,"%s%s '%s', go to state %d\n",
         yyTracePrompt, zTag, yyTokenName[yypParser->yytos->major],
         yyNewState);
    }else{
      fprintf(yyTraceFILE,"%s%s '%s', pending reduce %d\n",
         yyTracePrompt, zTag, yyTokenName[yypParser->yytos->major],
         yyNewState - YY_MIN_REDUCE);
    }
  }
}
#else
# define yyTraceShift(X,Y,Z)
#endif

/*
** Perform a shift action.
*/
static void yy_shift(
  yyParser *yypParser,          /* The parser to be shifted */
  YYACTIONTYPE yyNewState,      /* The new state to shift in */
  YYCODETYPE yyMajor,           /* The major token to shift in */
  sqlite3ParserTOKENTYPE yyMinor        /* The minor token to shift in */
){
  yyStackEntry *yytos;
  yypParser->yytos++;
#ifdef YYTRACKMAXSTACKDEPTH
  if( (int)(yypParser->yytos - yypParser->yystack)>yypParser->yyhwm ){
    yypParser->yyhwm++;
    assert( yypParser->yyhwm == (int)(yypParser->yytos - yypParser->yystack) );
  }
#endif
#if YYSTACKDEPTH>0 
  if( yypParser->yytos>yypParser->yystackEnd ){
    yypParser->yytos--;
    yyStackOverflow(yypParser);
    return;
  }
#else
  if( yypParser->yytos>=&yypParser->yystack[yypParser->yystksz] ){
    if( yyGrowStack(yypParser) ){
      yypParser->yytos--;
      yyStackOverflow(yypParser);
      return;
    }
  }
#endif
  if( yyNewState > YY_MAX_SHIFT ){
    yyNewState += YY_MIN_REDUCE - YY_MIN_SHIFTREDUCE;
  }
  yytos = yypParser->yytos;
  yytos->stateno = yyNewState;
  yytos->major = yyMajor;
  yytos->minor.yy0 = yyMinor;
  yyTraceShift(yypParser, yyNewState, "Shift");
}

/* For rule J, yyRuleInfoLhs[J] contains the symbol on the left-hand side
** of that rule */
static const YYCODETYPE yyRuleInfoLhs[] = {
   183,  /* (0) explain ::= EXPLAIN */
   183,  /* (1) explain ::= EXPLAIN QUERY PLAN */
   182,  /* (2) cmdx ::= cmd */
   184,  /* (3) cmd ::= BEGIN transtype trans_opt */
   185,  /* (4) transtype ::= */
   185,  /* (5) transtype ::= DEFERRED */
   185,  /* (6) transtype ::= IMMEDIATE */
   185,  /* (7) transtype ::= EXCLUSIVE */
   184,  /* (8) cmd ::= COMMIT|END trans_opt */
   184,  /* (9) cmd ::= ROLLBACK trans_opt */
   184,  /* (10) cmd ::= SAVEPOINT nm */
   184,  /* (11) cmd ::= RELEASE savepoint_opt nm */
   184,  /* (12) cmd ::= ROLLBACK trans_opt TO savepoint_opt nm */
   189,  /* (13) create_table ::= createkw temp TABLE ifnotexists nm dbnm */
   191,  /* (14) createkw ::= CREATE */
   193,  /* (15) ifnotexists ::= */
   193,  /* (16) ifnotexists ::= IF NOT EXISTS */
   192,  /* (17) temp ::= TEMP */
   192,  /* (18) temp ::= */
   190,  /* (19) create_table_args ::= LP columnlist conslist_opt RP table_options */
   190,  /* (20) create_table_args ::= AS select */
   197,  /* (21) table_options ::= */
   197,  /* (22) table_options ::= WITHOUT nm */
   199,  /* (23) columnname ::= nm typetoken */
   201,  /* (24) typetoken ::= */
   201,  /* (25) typetoken ::= typename LP signed RP */
   201,  /* (26) typetoken ::= typename LP signed COMMA signed RP */
   202,  /* (27) typename ::= typename ID|STRING */
   206,  /* (28) scanpt ::= */
   207,  /* (29) scantok ::= */
   208,  /* (30) ccons ::= CONSTRAINT nm */
   208,  /* (31) ccons ::= DEFAULT scantok term */
   208,  /* (32) ccons ::= DEFAULT LP expr RP */
   208,  /* (33) ccons ::= DEFAULT PLUS scantok term */
   208,  /* (34) ccons ::= DEFAULT MINUS scantok term */
   208,  /* (35) ccons ::= DEFAULT scantok ID|INDEXED */
   208,  /* (36) ccons ::= NOT NULL onconf */
   208,  /* (37) ccons ::= PRIMARY KEY sortorder onconf autoinc */
   208,  /* (38) ccons ::= UNIQUE onconf */
   208,  /* (39) ccons ::= CHECK LP expr RP */
   208,  /* (40) ccons ::= REFERENCES nm eidlist_opt refargs */
   208,  /* (41) ccons ::= defer_subclause */
   208,  /* (42) ccons ::= COLLATE ID|STRING */
   213,  /* (43) autoinc ::= */
   213,  /* (44) autoinc ::= AUTOINCR */
   215,  /* (45) refargs ::= */
   215,  /* (46) refargs ::= refargs refarg */
   217,  /* (47) refarg ::= MATCH nm */
   217,  /* (48) refarg ::= ON INSERT refact */
   217,  /* (49) refarg ::= ON DELETE refact */
   217,  /* (50) refarg ::= ON UPDATE refact */
   218,  /* (51) refact ::= SET NULL */
   218,  /* (52) refact ::= SET DEFAULT */
   218,  /* (53) refact ::= CASCADE */
   218,  /* (54) refact ::= RESTRICT */
   218,  /* (55) refact ::= NO ACTION */
   216,  /* (56) defer_subclause ::= NOT DEFERRABLE init_deferred_pred_opt */
   216,  /* (57) defer_subclause ::= DEFERRABLE init_deferred_pred_opt */
   219,  /* (58) init_deferred_pred_opt ::= */
   219,  /* (59) init_deferred_pred_opt ::= INITIALLY DEFERRED */
   219,  /* (60) init_deferred_pred_opt ::= INITIALLY IMMEDIATE */
   196,  /* (61) conslist_opt ::= */
   221,  /* (62) tconscomma ::= COMMA */
   222,  /* (63) tcons ::= CONSTRAINT nm */
   222,  /* (64) tcons ::= PRIMARY KEY LP sortlist autoinc RP onconf */
   222,  /* (65) tcons ::= UNIQUE LP sortlist RP onconf */
   222,  /* (66) tcons ::= CHECK LP expr RP onconf */
   222,  /* (67) tcons ::= FOREIGN KEY LP eidlist RP REFERENCES nm eidlist_opt refargs defer_subclause_opt */
   225,  /* (68) defer_subclause_opt ::= */
   211,  /* (69) onconf ::= */
   211,  /* (70) onconf ::= ON CONFLICT resolvetype */
   226,  /* (71) orconf ::= */
   226,  /* (72) orconf ::= OR resolvetype */
   227,  /* (73) resolvetype ::= IGNORE */
   227,  /* (74) resolvetype ::= REPLACE */
   184,  /* (75) cmd ::= DROP TABLE ifexists fullname */
   229,  /* (76) ifexists ::= IF EXISTS */
   229,  /* (77) ifexists ::= */
   184,  /* (78) cmd ::= createkw temp VIEW ifnotexists nm dbnm eidlist_opt AS select */
   184,  /* (79) cmd ::= DROP VIEW ifexists fullname */
   184,  /* (80) cmd ::= select */
   198,  /* (81) select ::= WITH wqlist selectnowith */
   198,  /* (82) select ::= WITH RECURSIVE wqlist selectnowith */
   198,  /* (83) select ::= selectnowith */
   231,  /* (84) selectnowith ::= selectnowith multiselect_op oneselect */
   234,  /* (85) multiselect_op ::= UNION */
   234,  /* (86) multiselect_op ::= UNION ALL */
   234,  /* (87) multiselect_op ::= EXCEPT|INTERSECT */
   232,  /* (88) oneselect ::= SELECT distinct selcollist from where_opt groupby_opt having_opt orderby_opt limit_opt */
   232,  /* (89) oneselect ::= SELECT distinct selcollist from where_opt groupby_opt having_opt window_clause orderby_opt limit_opt */
   244,  /* (90) values ::= VALUES LP nexprlist RP */
   244,  /* (91) values ::= values COMMA LP nexprlist RP */
   235,  /* (92) distinct ::= DISTINCT */
   235,  /* (93) distinct ::= ALL */
   235,  /* (94) distinct ::= */
   246,  /* (95) sclp ::= */
   236,  /* (96) selcollist ::= sclp scanpt expr scanpt as */
   236,  /* (97) selcollist ::= sclp scanpt STAR */
   236,  /* (98) selcollist ::= sclp scanpt nm DOT STAR */
   247,  /* (99) as ::= AS nm */
   247,  /* (100) as ::= */
   237,  /* (101) from ::= */
   237,  /* (102) from ::= FROM seltablist */
   249,  /* (103) stl_prefix ::= seltablist joinop */
   249,  /* (104) stl_prefix ::= */
   248,  /* (105) seltablist ::= stl_prefix nm dbnm as indexed_opt on_opt using_opt */
   248,  /* (106) seltablist ::= stl_prefix nm dbnm LP exprlist RP as on_opt using_opt */
   248,  /* (107) seltablist ::= stl_prefix LP select RP as on_opt using_opt */
   248,  /* (108) seltablist ::= stl_prefix LP seltablist RP as on_opt using_opt */
   194,  /* (109) dbnm ::= */
   194,  /* (110) dbnm ::= DOT nm */
   230,  /* (111) fullname ::= nm */
   230,  /* (112) fullname ::= nm DOT nm */
   255,  /* (113) xfullname ::= nm */
   255,  /* (114) xfullname ::= nm DOT nm */
   255,  /* (115) xfullname ::= nm DOT nm AS nm */
   255,  /* (116) xfullname ::= nm AS nm */
   250,  /* (117) joinop ::= COMMA|JOIN */
   250,  /* (118) joinop ::= JOIN_KW JOIN */
   250,  /* (119) joinop ::= JOIN_KW nm JOIN */
   250,  /* (120) joinop ::= JOIN_KW nm nm JOIN */
   252,  /* (121) on_opt ::= ON expr */
   252,  /* (122) on_opt ::= */
   251,  /* (123) indexed_opt ::= */
   251,  /* (124) indexed_opt ::= INDEXED BY nm */
   251,  /* (125) indexed_opt ::= NOT INDEXED */
   253,  /* (126) using_opt ::= USING LP idlist RP */
   253,  /* (127) using_opt ::= */
   241,  /* (128) orderby_opt ::= */
   241,  /* (129) orderby_opt ::= ORDER BY sortlist */
   223,  /* (130) sortlist ::= sortlist COMMA expr sortorder nulls */
   223,  /* (131) sortlist ::= expr sortorder nulls */
   212,  /* (132) sortorder ::= ASC */
   212,  /* (133) sortorder ::= DESC */
   212,  /* (134) sortorder ::= */
   257,  /* (135) nulls ::= NULLS FIRST */
   257,  /* (136) nulls ::= NULLS LAST */
   257,  /* (137) nulls ::= */
   239,  /* (138) groupby_opt ::= */
   239,  /* (139) groupby_opt ::= GROUP BY nexprlist */
   240,  /* (140) having_opt ::= */
   240,  /* (141) having_opt ::= HAVING expr */
   242,  /* (142) limit_opt ::= */
   242,  /* (143) limit_opt ::= LIMIT expr */
   242,  /* (144) limit_opt ::= LIMIT expr OFFSET expr */
   242,  /* (145) limit_opt ::= LIMIT expr COMMA expr */
   184,  /* (146) cmd ::= with DELETE FROM xfullname indexed_opt where_opt */
   238,  /* (147) where_opt ::= */
   238,  /* (148) where_opt ::= WHERE expr */
   184,  /* (149) cmd ::= with UPDATE orconf xfullname indexed_opt SET setlist where_opt */
   259,  /* (150) setlist ::= setlist COMMA nm EQ expr */
   259,  /* (151) setlist ::= setlist COMMA LP idlist RP EQ expr */
   259,  /* (152) setlist ::= nm EQ expr */
   259,  /* (153) setlist ::= LP idlist RP EQ expr */
   184,  /* (154) cmd ::= with insert_cmd INTO xfullname idlist_opt select upsert */
   184,  /* (155) cmd ::= with insert_cmd INTO xfullname idlist_opt DEFAULT VALUES */
   262,  /* (156) upsert ::= */
   262,  /* (157) upsert ::= ON CONFLICT LP sortlist RP where_opt DO UPDATE SET setlist where_opt */
   262,  /* (158) upsert ::= ON CONFLICT LP sortlist RP where_opt DO NOTHING */
   262,  /* (159) upsert ::= ON CONFLICT DO NOTHING */
   260,  /* (160) insert_cmd ::= INSERT orconf */
   260,  /* (161) insert_cmd ::= REPLACE */
   261,  /* (162) idlist_opt ::= */
   261,  /* (163) idlist_opt ::= LP idlist RP */
   256,  /* (164) idlist ::= idlist COMMA nm */
   256,  /* (165) idlist ::= nm */
   210,  /* (166) expr ::= LP expr RP */
   210,  /* (167) expr ::= ID|INDEXED */
   210,  /* (168) expr ::= JOIN_KW */
   210,  /* (169) expr ::= nm DOT nm */
   210,  /* (170) expr ::= nm DOT nm DOT nm */
   209,  /* (171) term ::= NULL|FLOAT|BLOB */
   209,  /* (172) term ::= STRING */
   209,  /* (173) term ::= INTEGER */
   210,  /* (174) expr ::= VARIABLE */
   210,  /* (175) expr ::= expr COLLATE ID|STRING */
   210,  /* (176) expr ::= CAST LP expr AS typetoken RP */
   210,  /* (177) expr ::= ID|INDEXED LP distinct exprlist RP */
   210,  /* (178) expr ::= ID|INDEXED LP STAR RP */
   210,  /* (179) expr ::= ID|INDEXED LP distinct exprlist RP filter_over */
   210,  /* (180) expr ::= ID|INDEXED LP STAR RP filter_over */
   209,  /* (181) term ::= CTIME_KW */
   210,  /* (182) expr ::= LP nexprlist COMMA expr RP */
   210,  /* (183) expr ::= expr AND expr */
   210,  /* (184) expr ::= expr OR expr */
   210,  /* (185) expr ::= expr LT|GT|GE|LE expr */
   210,  /* (186) expr ::= expr EQ|NE expr */
   210,  /* (187) expr ::= expr BITAND|BITOR|LSHIFT|RSHIFT expr */
   210,  /* (188) expr ::= expr PLUS|MINUS expr */
   210,  /* (189) expr ::= expr STAR|SLASH|REM expr */
   210,  /* (190) expr ::= expr CONCAT expr */
   264,  /* (191) likeop ::= NOT LIKE_KW|MATCH */
   210,  /* (192) expr ::= expr likeop expr */
   210,  /* (193) expr ::= expr likeop expr ESCAPE expr */
   210,  /* (194) expr ::= expr ISNULL|NOTNULL */
   210,  /* (195) expr ::= expr NOT NULL */
   210,  /* (196) expr ::= expr IS expr */
   210,  /* (197) expr ::= expr IS NOT expr */
   210,  /* (198) expr ::= NOT expr */
   210,  /* (199) expr ::= BITNOT expr */
   210,  /* (200) expr ::= PLUS|MINUS expr */
   265,  /* (201) between_op ::= BETWEEN */
   265,  /* (202) between_op ::= NOT BETWEEN */
   210,  /* (203) expr ::= expr between_op expr AND expr */
   266,  /* (204) in_op ::= IN */
   266,  /* (205) in_op ::= NOT IN */
   210,  /* (206) expr ::= expr in_op LP exprlist RP */
   210,  /* (207) expr ::= LP select RP */
   210,  /* (208) expr ::= expr in_op LP select RP */
   210,  /* (209) expr ::= expr in_op nm dbnm paren_exprlist */
   210,  /* (210) expr ::= EXISTS LP select RP */
   210,  /* (211) expr ::= CASE case_operand case_exprlist case_else END */
   269,  /* (212) case_exprlist ::= case_exprlist WHEN expr THEN expr */
   269,  /* (213) case_exprlist ::= WHEN expr THEN expr */
   270,  /* (214) case_else ::= ELSE expr */
   270,  /* (215) case_else ::= */
   268,  /* (216) case_operand ::= expr */
   268,  /* (217) case_operand ::= */
   254,  /* (218) exprlist ::= */
   245,  /* (219) nexprlist ::= nexprlist COMMA expr */
   245,  /* (220) nexprlist ::= expr */
   267,  /* (221) paren_exprlist ::= */
   267,  /* (222) paren_exprlist ::= LP exprlist RP */
   184,  /* (223) cmd ::= createkw uniqueflag INDEX ifnotexists nm dbnm ON nm LP sortlist RP where_opt */
   271,  /* (224) uniqueflag ::= UNIQUE */
   271,  /* (225) uniqueflag ::= */
   214,  /* (226) eidlist_opt ::= */
   214,  /* (227) eidlist_opt ::= LP eidlist RP */
   224,  /* (228) eidlist ::= eidlist COMMA nm collate sortorder */
   224,  /* (229) eidlist ::= nm collate sortorder */
   272,  /* (230) collate ::= */
   272,  /* (231) collate ::= COLLATE ID|STRING */
   184,  /* (232) cmd ::= DROP INDEX ifexists fullname */
   184,  /* (233) cmd ::= VACUUM vinto */
   184,  /* (234) cmd ::= VACUUM nm vinto */
   273,  /* (235) vinto ::= INTO expr */
   273,  /* (236) vinto ::= */
   184,  /* (237) cmd ::= PRAGMA nm dbnm */
   184,  /* (238) cmd ::= PRAGMA nm dbnm EQ nmnum */
   184,  /* (239) cmd ::= PRAGMA nm dbnm LP nmnum RP */
   184,  /* (240) cmd ::= PRAGMA nm dbnm EQ minus_num */
   184,  /* (241) cmd ::= PRAGMA nm dbnm LP minus_num RP */
   204,  /* (242) plus_num ::= PLUS INTEGER|FLOAT */
   205,  /* (243) minus_num ::= MINUS INTEGER|FLOAT */
   184,  /* (244) cmd ::= createkw trigger_decl BEGIN trigger_cmd_list END */
   275,  /* (245) trigger_decl ::= temp TRIGGER ifnotexists nm dbnm trigger_time trigger_event ON fullname foreach_clause when_clause */
   277,  /* (246) trigger_time ::= BEFORE|AFTER */
   277,  /* (247) trigger_time ::= INSTEAD OF */
   277,  /* (248) trigger_time ::= */
   278,  /* (249) trigger_event ::= DELETE|INSERT */
   278,  /* (250) trigger_event ::= UPDATE */
   278,  /* (251) trigger_event ::= UPDATE OF idlist */
   280,  /* (252) when_clause ::= */
   280,  /* (253) when_clause ::= WHEN expr */
   276,  /* (254) trigger_cmd_list ::= trigger_cmd_list trigger_cmd SEMI */
   276,  /* (255) trigger_cmd_list ::= trigger_cmd SEMI */
   282,  /* (256) trnm ::= nm DOT nm */
   283,  /* (257) tridxby ::= INDEXED BY nm */
   283,  /* (258) tridxby ::= NOT INDEXED */
   281,  /* (259) trigger_cmd ::= UPDATE orconf trnm tridxby SET setlist where_opt scanpt */
   281,  /* (260) trigger_cmd ::= scanpt insert_cmd INTO trnm idlist_opt select upsert scanpt */
   281,  /* (261) trigger_cmd ::= DELETE FROM trnm tridxby where_opt scanpt */
   281,  /* (262) trigger_cmd ::= scanpt select scanpt */
   210,  /* (263) expr ::= RAISE LP IGNORE RP */
   210,  /* (264) expr ::= RAISE LP raisetype COMMA nm RP */
   228,  /* (265) raisetype ::= ROLLBACK */
   228,  /* (266) raisetype ::= ABORT */
   228,  /* (267) raisetype ::= FAIL */
   184,  /* (268) cmd ::= DROP TRIGGER ifexists fullname */
   184,  /* (269) cmd ::= ATTACH database_kw_opt expr AS expr key_opt */
   184,  /* (270) cmd ::= DETACH database_kw_opt expr */
   285,  /* (271) key_opt ::= */
   285,  /* (272) key_opt ::= KEY expr */
   184,  /* (273) cmd ::= REINDEX */
   184,  /* (274) cmd ::= REINDEX nm dbnm */
   184,  /* (275) cmd ::= ANALYZE */
   184,  /* (276) cmd ::= ANALYZE nm dbnm */
   184,  /* (277) cmd ::= ALTER TABLE fullname RENAME TO nm */
   184,  /* (278) cmd ::= ALTER TABLE add_column_fullname ADD kwcolumn_opt columnname carglist */
   286,  /* (279) add_column_fullname ::= fullname */
   184,  /* (280) cmd ::= ALTER TABLE fullname RENAME kwcolumn_opt nm TO nm */
   184,  /* (281) cmd ::= create_vtab */
   184,  /* (282) cmd ::= create_vtab LP vtabarglist RP */
   288,  /* (283) create_vtab ::= createkw VIRTUAL TABLE ifnotexists nm dbnm USING nm */
   290,  /* (284) vtabarg ::= */
   291,  /* (285) vtabargtoken ::= ANY */
   291,  /* (286) vtabargtoken ::= lp anylist RP */
   292,  /* (287) lp ::= LP */
   258,  /* (288) with ::= WITH wqlist */
   258,  /* (289) with ::= WITH RECURSIVE wqlist */
   233,  /* (290) wqlist ::= nm eidlist_opt AS LP select RP */
   233,  /* (291) wqlist ::= wqlist COMMA nm eidlist_opt AS LP select RP */
   294,  /* (292) windowdefn_list ::= windowdefn */
   294,  /* (293) windowdefn_list ::= windowdefn_list COMMA windowdefn */
   295,  /* (294) windowdefn ::= nm AS LP window RP */
   296,  /* (295) window ::= PARTITION BY nexprlist orderby_opt frame_opt */
   296,  /* (296) window ::= nm PARTITION BY nexprlist orderby_opt frame_opt */
   296,  /* (297) window ::= ORDER BY sortlist frame_opt */
   296,  /* (298) window ::= nm ORDER BY sortlist frame_opt */
   296,  /* (299) window ::= frame_opt */
   296,  /* (300) window ::= nm frame_opt */
   297,  /* (301) frame_opt ::= */
   297,  /* (302) frame_opt ::= range_or_rows frame_bound_s frame_exclude_opt */
   297,  /* (303) frame_opt ::= range_or_rows BETWEEN frame_bound_s AND frame_bound_e frame_exclude_opt */
   301,  /* (304) range_or_rows ::= RANGE|ROWS|GROUPS */
   303,  /* (305) frame_bound_s ::= frame_bound */
   303,  /* (306) frame_bound_s ::= UNBOUNDED PRECEDING */
   304,  /* (307) frame_bound_e ::= frame_bound */
   304,  /* (308) frame_bound_e ::= UNBOUNDED FOLLOWING */
   302,  /* (309) frame_bound ::= expr PRECEDING|FOLLOWING */
   302,  /* (310) frame_bound ::= CURRENT ROW */
   305,  /* (311) frame_exclude_opt ::= */
   305,  /* (312) frame_exclude_opt ::= EXCLUDE frame_exclude */
   306,  /* (313) frame_exclude ::= NO OTHERS */
   306,  /* (314) frame_exclude ::= CURRENT ROW */
   306,  /* (315) frame_exclude ::= GROUP|TIES */
   243,  /* (316) window_clause ::= WINDOW windowdefn_list */
   263,  /* (317) filter_over ::= filter_clause over_clause */
   263,  /* (318) filter_over ::= over_clause */
   263,  /* (319) filter_over ::= filter_clause */
   300,  /* (320) over_clause ::= OVER LP window RP */
   300,  /* (321) over_clause ::= OVER nm */
   299,  /* (322) filter_clause ::= FILTER LP WHERE expr RP */
   179,  /* (323) input ::= cmdlist */
   180,  /* (324) cmdlist ::= cmdlist ecmd */
   180,  /* (325) cmdlist ::= ecmd */
   181,  /* (326) ecmd ::= SEMI */
   181,  /* (327) ecmd ::= cmdx SEMI */
   181,  /* (328) ecmd ::= explain cmdx */
   186,  /* (329) trans_opt ::= */
   186,  /* (330) trans_opt ::= TRANSACTION */
   186,  /* (331) trans_opt ::= TRANSACTION nm */
   188,  /* (332) savepoint_opt ::= SAVEPOINT */
   188,  /* (333) savepoint_opt ::= */
   184,  /* (334) cmd ::= create_table create_table_args */
   195,  /* (335) columnlist ::= columnlist COMMA columnname carglist */
   195,  /* (336) columnlist ::= columnname carglist */
   187,  /* (337) nm ::= ID|INDEXED */
   187,  /* (338) nm ::= STRING */
   187,  /* (339) nm ::= JOIN_KW */
   201,  /* (340) typetoken ::= typename */
   202,  /* (341) typename ::= ID|STRING */
   203,  /* (342) signed ::= plus_num */
   203,  /* (343) signed ::= minus_num */
   200,  /* (344) carglist ::= carglist ccons */
   200,  /* (345) carglist ::= */
   208,  /* (346) ccons ::= NULL onconf */
   196,  /* (347) conslist_opt ::= COMMA conslist */
   220,  /* (348) conslist ::= conslist tconscomma tcons */
   220,  /* (349) conslist ::= tcons */
   221,  /* (350) tconscomma ::= */
   225,  /* (351) defer_subclause_opt ::= defer_subclause */
   227,  /* (352) resolvetype ::= raisetype */
   231,  /* (353) selectnowith ::= oneselect */
   232,  /* (354) oneselect ::= values */
   246,  /* (355) sclp ::= selcollist COMMA */
   247,  /* (356) as ::= ID|STRING */
   210,  /* (357) expr ::= term */
   264,  /* (358) likeop ::= LIKE_KW|MATCH */
   254,  /* (359) exprlist ::= nexprlist */
   274,  /* (360) nmnum ::= plus_num */
   274,  /* (361) nmnum ::= nm */
   274,  /* (362) nmnum ::= ON */
   274,  /* (363) nmnum ::= DELETE */
   274,  /* (364) nmnum ::= DEFAULT */
   204,  /* (365) plus_num ::= INTEGER|FLOAT */
   279,  /* (366) foreach_clause ::= */
   279,  /* (367) foreach_clause ::= FOR EACH ROW */
   282,  /* (368) trnm ::= nm */
   283,  /* (369) tridxby ::= */
   284,  /* (370) database_kw_opt ::= DATABASE */
   284,  /* (371) database_kw_opt ::= */
   287,  /* (372) kwcolumn_opt ::= */
   287,  /* (373) kwcolumn_opt ::= COLUMNKW */
   289,  /* (374) vtabarglist ::= vtabarg */
   289,  /* (375) vtabarglist ::= vtabarglist COMMA vtabarg */
   290,  /* (376) vtabarg ::= vtabarg vtabargtoken */
   293,  /* (377) anylist ::= */
   293,  /* (378) anylist ::= anylist LP anylist RP */
   293,  /* (379) anylist ::= anylist ANY */
   258,  /* (380) with ::= */
};

/* For rule J, yyRuleInfoNRhs[J] contains the negative of the number
** of symbols on the right-hand side of that rule. */
static const signed char yyRuleInfoNRhs[] = {
   -1,  /* (0) explain ::= EXPLAIN */
   -3,  /* (1) explain ::= EXPLAIN QUERY PLAN */
   -1,  /* (2) cmdx ::= cmd */
   -3,  /* (3) cmd ::= BEGIN transtype trans_opt */
    0,  /* (4) transtype ::= */
   -1,  /* (5) transtype ::= DEFERRED */
   -1,  /* (6) transtype ::= IMMEDIATE */
   -1,  /* (7) transtype ::= EXCLUSIVE */
   -2,  /* (8) cmd ::= COMMIT|END trans_opt */
   -2,  /* (9) cmd ::= ROLLBACK trans_opt */
   -2,  /* (10) cmd ::= SAVEPOINT nm */
   -3,  /* (11) cmd ::= RELEASE savepoint_opt nm */
   -5,  /* (12) cmd ::= ROLLBACK trans_opt TO savepoint_opt nm */
   -6,  /* (13) create_table ::= createkw temp TABLE ifnotexists nm dbnm */
   -1,  /* (14) createkw ::= CREATE */
    0,  /* (15) ifnotexists ::= */
   -3,  /* (16) ifnotexists ::= IF NOT EXISTS */
   -1,  /* (17) temp ::= TEMP */
    0,  /* (18) temp ::= */
   -5,  /* (19) create_table_args ::= LP columnlist conslist_opt RP table_options */
   -2,  /* (20) create_table_args ::= AS select */
    0,  /* (21) table_options ::= */
   -2,  /* (22) table_options ::= WITHOUT nm */
   -2,  /* (23) columnname ::= nm typetoken */
    0,  /* (24) typetoken ::= */
   -4,  /* (25) typetoken ::= typename LP signed RP */
   -6,  /* (26) typetoken ::= typename LP signed COMMA signed RP */
   -2,  /* (27) typename ::= typename ID|STRING */
    0,  /* (28) scanpt ::= */
    0,  /* (29) scantok ::= */
   -2,  /* (30) ccons ::= CONSTRAINT nm */
   -3,  /* (31) ccons ::= DEFAULT scantok term */
   -4,  /* (32) ccons ::= DEFAULT LP expr RP */
   -4,  /* (33) ccons ::= DEFAULT PLUS scantok term */
   -4,  /* (34) ccons ::= DEFAULT MINUS scantok term */
   -3,  /* (35) ccons ::= DEFAULT scantok ID|INDEXED */
   -3,  /* (36) ccons ::= NOT NULL onconf */
   -5,  /* (37) ccons ::= PRIMARY KEY sortorder onconf autoinc */
   -2,  /* (38) ccons ::= UNIQUE onconf */
   -4,  /* (39) ccons ::= CHECK LP expr RP */
   -4,  /* (40) ccons ::= REFERENCES nm eidlist_opt refargs */
   -1,  /* (41) ccons ::= defer_subclause */
   -2,  /* (42) ccons ::= COLLATE ID|STRING */
    0,  /* (43) autoinc ::= */
   -1,  /* (44) autoinc ::= AUTOINCR */
    0,  /* (45) refargs ::= */
   -2,  /* (46) refargs ::= refargs refarg */
   -2,  /* (47) refarg ::= MATCH nm */
   -3,  /* (48) refarg ::= ON INSERT refact */
   -3,  /* (49) refarg ::= ON DELETE refact */
   -3,  /* (50) refarg ::= ON UPDATE refact */
   -2,  /* (51) refact ::= SET NULL */
   -2,  /* (52) refact ::= SET DEFAULT */
   -1,  /* (53) refact ::= CASCADE */
   -1,  /* (54) refact ::= RESTRICT */
   -2,  /* (55) refact ::= NO ACTION */
   -3,  /* (56) defer_subclause ::= NOT DEFERRABLE init_deferred_pred_opt */
   -2,  /* (57) defer_subclause ::= DEFERRABLE init_deferred_pred_opt */
    0,  /* (58) init_deferred_pred_opt ::= */
   -2,  /* (59) init_deferred_pred_opt ::= INITIALLY DEFERRED */
   -2,  /* (60) init_deferred_pred_opt ::= INITIALLY IMMEDIATE */
    0,  /* (61) conslist_opt ::= */
   -1,  /* (62) tconscomma ::= COMMA */
   -2,  /* (63) tcons ::= CONSTRAINT nm */
   -7,  /* (64) tcons ::= PRIMARY KEY LP sortlist autoinc RP onconf */
   -5,  /* (65) tcons ::= UNIQUE LP sortlist RP onconf */
   -5,  /* (66) tcons ::= CHECK LP expr RP onconf */
  -10,  /* (67) tcons ::= FOREIGN KEY LP eidlist RP REFERENCES nm eidlist_opt refargs defer_subclause_opt */
    0,  /* (68) defer_subclause_opt ::= */
    0,  /* (69) onconf ::= */
   -3,  /* (70) onconf ::= ON CONFLICT resolvetype */
    0,  /* (71) orconf ::= */
   -2,  /* (72) orconf ::= OR resolvetype */
   -1,  /* (73) resolvetype ::= IGNORE */
   -1,  /* (74) resolvetype ::= REPLACE */
   -4,  /* (75) cmd ::= DROP TABLE ifexists fullname */
   -2,  /* (76) ifexists ::= IF EXISTS */
    0,  /* (77) ifexists ::= */
   -9,  /* (78) cmd ::= createkw temp VIEW ifnotexists nm dbnm eidlist_opt AS select */
   -4,  /* (79) cmd ::= DROP VIEW ifexists fullname */
   -1,  /* (80) cmd ::= select */
   -3,  /* (81) select ::= WITH wqlist selectnowith */
   -4,  /* (82) select ::= WITH RECURSIVE wqlist selectnowith */
   -1,  /* (83) select ::= selectnowith */
   -3,  /* (84) selectnowith ::= selectnowith multiselect_op oneselect */
   -1,  /* (85) multiselect_op ::= UNION */
   -2,  /* (86) multiselect_op ::= UNION ALL */
   -1,  /* (87) multiselect_op ::= EXCEPT|INTERSECT */
   -9,  /* (88) oneselect ::= SELECT distinct selcollist from where_opt groupby_opt having_opt orderby_opt limit_opt */
  -10,  /* (89) oneselect ::= SELECT distinct selcollist from where_opt groupby_opt having_opt window_clause orderby_opt limit_opt */
   -4,  /* (90) values ::= VALUES LP nexprlist RP */
   -5,  /* (91) values ::= values COMMA LP nexprlist RP */
   -1,  /* (92) distinct ::= DISTINCT */
   -1,  /* (93) distinct ::= ALL */
    0,  /* (94) distinct ::= */
    0,  /* (95) sclp ::= */
   -5,  /* (96) selcollist ::= sclp scanpt expr scanpt as */
   -3,  /* (97) selcollist ::= sclp scanpt STAR */
   -5,  /* (98) selcollist ::= sclp scanpt nm DOT STAR */
   -2,  /* (99) as ::= AS nm */
    0,  /* (100) as ::= */
    0,  /* (101) from ::= */
   -2,  /* (102) from ::= FROM seltablist */
   -2,  /* (103) stl_prefix ::= seltablist joinop */
    0,  /* (104) stl_prefix ::= */
   -7,  /* (105) seltablist ::= stl_prefix nm dbnm as indexed_opt on_opt using_opt */
   -9,  /* (106) seltablist ::= stl_prefix nm dbnm LP exprlist RP as on_opt using_opt */
   -7,  /* (107) seltablist ::= stl_prefix LP select RP as on_opt using_opt */
   -7,  /* (108) seltablist ::= stl_prefix LP seltablist RP as on_opt using_opt */
    0,  /* (109) dbnm ::= */
   -2,  /* (110) dbnm ::= DOT nm */
   -1,  /* (111) fullname ::= nm */
   -3,  /* (112) fullname ::= nm DOT nm */
   -1,  /* (113) xfullname ::= nm */
   -3,  /* (114) xfullname ::= nm DOT nm */
   -5,  /* (115) xfullname ::= nm DOT nm AS nm */
   -3,  /* (116) xfullname ::= nm AS nm */
   -1,  /* (117) joinop ::= COMMA|JOIN */
   -2,  /* (118) joinop ::= JOIN_KW JOIN */
   -3,  /* (119) joinop ::= JOIN_KW nm JOIN */
   -4,  /* (120) joinop ::= JOIN_KW nm nm JOIN */
   -2,  /* (121) on_opt ::= ON expr */
    0,  /* (122) on_opt ::= */
    0,  /* (123) indexed_opt ::= */
   -3,  /* (124) indexed_opt ::= INDEXED BY nm */
   -2,  /* (125) indexed_opt ::= NOT INDEXED */
   -4,  /* (126) using_opt ::= USING LP idlist RP */
    0,  /* (127) using_opt ::= */
    0,  /* (128) orderby_opt ::= */
   -3,  /* (129) orderby_opt ::= ORDER BY sortlist */
   -5,  /* (130) sortlist ::= sortlist COMMA expr sortorder nulls */
   -3,  /* (131) sortlist ::= expr sortorder nulls */
   -1,  /* (132) sortorder ::= ASC */
   -1,  /* (133) sortorder ::= DESC */
    0,  /* (134) sortorder ::= */
   -2,  /* (135) nulls ::= NULLS FIRST */
   -2,  /* (136) nulls ::= NULLS LAST */
    0,  /* (137) nulls ::= */
    0,  /* (138) groupby_opt ::= */
   -3,  /* (139) groupby_opt ::= GROUP BY nexprlist */
    0,  /* (140) having_opt ::= */
   -2,  /* (141) having_opt ::= HAVING expr */
    0,  /* (142) limit_opt ::= */
   -2,  /* (143) limit_opt ::= LIMIT expr */
   -4,  /* (144) limit_opt ::= LIMIT expr OFFSET expr */
   -4,  /* (145) limit_opt ::= LIMIT expr COMMA expr */
   -6,  /* (146) cmd ::= with DELETE FROM xfullname indexed_opt where_opt */
    0,  /* (147) where_opt ::= */
   -2,  /* (148) where_opt ::= WHERE expr */
   -8,  /* (149) cmd ::= with UPDATE orconf xfullname indexed_opt SET setlist where_opt */
   -5,  /* (150) setlist ::= setlist COMMA nm EQ expr */
   -7,  /* (151) setlist ::= setlist COMMA LP idlist RP EQ expr */
   -3,  /* (152) setlist ::= nm EQ expr */
   -5,  /* (153) setlist ::= LP idlist RP EQ expr */
   -7,  /* (154) cmd ::= with insert_cmd INTO xfullname idlist_opt select upsert */
   -7,  /* (155) cmd ::= with insert_cmd INTO xfullname idlist_opt DEFAULT VALUES */
    0,  /* (156) upsert ::= */
  -11,  /* (157) upsert ::= ON CONFLICT LP sortlist RP where_opt DO UPDATE SET setlist where_opt */
   -8,  /* (158) upsert ::= ON CONFLICT LP sortlist RP where_opt DO NOTHING */
   -4,  /* (159) upsert ::= ON CONFLICT DO NOTHING */
   -2,  /* (160) insert_cmd ::= INSERT orconf */
   -1,  /* (161) insert_cmd ::= REPLACE */
    0,  /* (162) idlist_opt ::= */
   -3,  /* (163) idlist_opt ::= LP idlist RP */
   -3,  /* (164) idlist ::= idlist COMMA nm */
   -1,  /* (165) idlist ::= nm */
   -3,  /* (166) expr ::= LP expr RP */
   -1,  /* (167) expr ::= ID|INDEXED */
   -1,  /* (168) expr ::= JOIN_KW */
   -3,  /* (169) expr ::= nm DOT nm */
   -5,  /* (170) expr ::= nm DOT nm DOT nm */
   -1,  /* (171) term ::= NULL|FLOAT|BLOB */
   -1,  /* (172) term ::= STRING */
   -1,  /* (173) term ::= INTEGER */
   -1,  /* (174) expr ::= VARIABLE */
   -3,  /* (175) expr ::= expr COLLATE ID|STRING */
   -6,  /* (176) expr ::= CAST LP expr AS typetoken RP */
   -5,  /* (177) expr ::= ID|INDEXED LP distinct exprlist RP */
   -4,  /* (178) expr ::= ID|INDEXED LP STAR RP */
   -6,  /* (179) expr ::= ID|INDEXED LP distinct exprlist RP filter_over */
   -5,  /* (180) expr ::= ID|INDEXED LP STAR RP filter_over */
   -1,  /* (181) term ::= CTIME_KW */
   -5,  /* (182) expr ::= LP nexprlist COMMA expr RP */
   -3,  /* (183) expr ::= expr AND expr */
   -3,  /* (184) expr ::= expr OR expr */
   -3,  /* (185) expr ::= expr LT|GT|GE|LE expr */
   -3,  /* (186) expr ::= expr EQ|NE expr */
   -3,  /* (187) expr ::= expr BITAND|BITOR|LSHIFT|RSHIFT expr */
   -3,  /* (188) expr ::= expr PLUS|MINUS expr */
   -3,  /* (189) expr ::= expr STAR|SLASH|REM expr */
   -3,  /* (190) expr ::= expr CONCAT expr */
   -2,  /* (191) likeop ::= NOT LIKE_KW|MATCH */
   -3,  /* (192) expr ::= expr likeop expr */
   -5,  /* (193) expr ::= expr likeop expr ESCAPE expr */
   -2,  /* (194) expr ::= expr ISNULL|NOTNULL */
   -3,  /* (195) expr ::= expr NOT NULL */
   -3,  /* (196) expr ::= expr IS expr */
   -4,  /* (197) expr ::= expr IS NOT expr */
   -2,  /* (198) expr ::= NOT expr */
   -2,  /* (199) expr ::= BITNOT expr */
   -2,  /* (200) expr ::= PLUS|MINUS expr */
   -1,  /* (201) between_op ::= BETWEEN */
   -2,  /* (202) between_op ::= NOT BETWEEN */
   -5,  /* (203) expr ::= expr between_op expr AND expr */
   -1,  /* (204) in_op ::= IN */
   -2,  /* (205) in_op ::= NOT IN */
   -5,  /* (206) expr ::= expr in_op LP exprlist RP */
   -3,  /* (207) expr ::= LP select RP */
   -5,  /* (208) expr ::= expr in_op LP select RP */
   -5,  /* (209) expr ::= expr in_op nm dbnm paren_exprlist */
   -4,  /* (210) expr ::= EXISTS LP select RP */
   -5,  /* (211) expr ::= CASE case_operand case_exprlist case_else END */
   -5,  /* (212) case_exprlist ::= case_exprlist WHEN expr THEN expr */
   -4,  /* (213) case_exprlist ::= WHEN expr THEN expr */
   -2,  /* (214) case_else ::= ELSE expr */
    0,  /* (215) case_else ::= */
   -1,  /* (216) case_operand ::= expr */
    0,  /* (217) case_operand ::= */
    0,  /* (218) exprlist ::= */
   -3,  /* (219) nexprlist ::= nexprlist COMMA expr */
   -1,  /* (220) nexprlist ::= expr */
    0,  /* (221) paren_exprlist ::= */
   -3,  /* (222) paren_exprlist ::= LP exprlist RP */
  -12,  /* (223) cmd ::= createkw uniqueflag INDEX ifnotexists nm dbnm ON nm LP sortlist RP where_opt */
   -1,  /* (224) uniqueflag ::= UNIQUE */
    0,  /* (225) uniqueflag ::= */
    0,  /* (226) eidlist_opt ::= */
   -3,  /* (227) eidlist_opt ::= LP eidlist RP */
   -5,  /* (228) eidlist ::= eidlist COMMA nm collate sortorder */
   -3,  /* (229) eidlist ::= nm collate sortorder */
    0,  /* (230) collate ::= */
   -2,  /* (231) collate ::= COLLATE ID|STRING */
   -4,  /* (232) cmd ::= DROP INDEX ifexists fullname */
   -2,  /* (233) cmd ::= VACUUM vinto */
   -3,  /* (234) cmd ::= VACUUM nm vinto */
   -2,  /* (235) vinto ::= INTO expr */
    0,  /* (236) vinto ::= */
   -3,  /* (237) cmd ::= PRAGMA nm dbnm */
   -5,  /* (238) cmd ::= PRAGMA nm dbnm EQ nmnum */
   -6,  /* (239) cmd ::= PRAGMA nm dbnm LP nmnum RP */
   -5,  /* (240) cmd ::= PRAGMA nm dbnm EQ minus_num */
   -6,  /* (241) cmd ::= PRAGMA nm dbnm LP minus_num RP */
   -2,  /* (242) plus_num ::= PLUS INTEGER|FLOAT */
   -2,  /* (243) minus_num ::= MINUS INTEGER|FLOAT */
   -5,  /* (244) cmd ::= createkw trigger_decl BEGIN trigger_cmd_list END */
  -11,  /* (245) trigger_decl ::= temp TRIGGER ifnotexists nm dbnm trigger_time trigger_event ON fullname foreach_clause when_clause */
   -1,  /* (246) trigger_time ::= BEFORE|AFTER */
   -2,  /* (247) trigger_time ::= INSTEAD OF */
    0,  /* (248) trigger_time ::= */
   -1,  /* (249) trigger_event ::= DELETE|INSERT */
   -1,  /* (250) trigger_event ::= UPDATE */
   -3,  /* (251) trigger_event ::= UPDATE OF idlist */
    0,  /* (252) when_clause ::= */
   -2,  /* (253) when_clause ::= WHEN expr */
   -3,  /* (254) trigger_cmd_list ::= trigger_cmd_list trigger_cmd SEMI */
   -2,  /* (255) trigger_cmd_list ::= trigger_cmd SEMI */
   -3,  /* (256) trnm ::= nm DOT nm */
   -3,  /* (257) tridxby ::= INDEXED BY nm */
   -2,  /* (258) tridxby ::= NOT INDEXED */
   -8,  /* (259) trigger_cmd ::= UPDATE orconf trnm tridxby SET setlist where_opt scanpt */
   -8,  /* (260) trigger_cmd ::= scanpt insert_cmd INTO trnm idlist_opt select upsert scanpt */
   -6,  /* (261) trigger_cmd ::= DELETE FROM trnm tridxby where_opt scanpt */
   -3,  /* (262) trigger_cmd ::= scanpt select scanpt */
   -4,  /* (263) expr ::= RAISE LP IGNORE RP */
   -6,  /* (264) expr ::= RAISE LP raisetype COMMA nm RP */
   -1,  /* (265) raisetype ::= ROLLBACK */
   -1,  /* (266) raisetype ::= ABORT */
   -1,  /* (267) raisetype ::= FAIL */
   -4,  /* (268) cmd ::= DROP TRIGGER ifexists fullname */
   -6,  /* (269) cmd ::= ATTACH database_kw_opt expr AS expr key_opt */
   -3,  /* (270) cmd ::= DETACH database_kw_opt expr */
    0,  /* (271) key_opt ::= */
   -2,  /* (272) key_opt ::= KEY expr */
   -1,  /* (273) cmd ::= REINDEX */
   -3,  /* (274) cmd ::= REINDEX nm dbnm */
   -1,  /* (275) cmd ::= ANALYZE */
   -3,  /* (276) cmd ::= ANALYZE nm dbnm */
   -6,  /* (277) cmd ::= ALTER TABLE fullname RENAME TO nm */
   -7,  /* (278) cmd ::= ALTER TABLE add_column_fullname ADD kwcolumn_opt columnname carglist */
   -1,  /* (279) add_column_fullname ::= fullname */
   -8,  /* (280) cmd ::= ALTER TABLE fullname RENAME kwcolumn_opt nm TO nm */
   -1,  /* (281) cmd ::= create_vtab */
   -4,  /* (282) cmd ::= create_vtab LP vtabarglist RP */
   -8,  /* (283) create_vtab ::= createkw VIRTUAL TABLE ifnotexists nm dbnm USING nm */
    0,  /* (284) vtabarg ::= */
   -1,  /* (285) vtabargtoken ::= ANY */
   -3,  /* (286) vtabargtoken ::= lp anylist RP */
   -1,  /* (287) lp ::= LP */
   -2,  /* (288) with ::= WITH wqlist */
   -3,  /* (289) with ::= WITH RECURSIVE wqlist */
   -6,  /* (290) wqlist ::= nm eidlist_opt AS LP select RP */
   -8,  /* (291) wqlist ::= wqlist COMMA nm eidlist_opt AS LP select RP */
   -1,  /* (292) windowdefn_list ::= windowdefn */
   -3,  /* (293) windowdefn_list ::= windowdefn_list COMMA windowdefn */
   -5,  /* (294) windowdefn ::= nm AS LP window RP */
   -5,  /* (295) window ::= PARTITION BY nexprlist orderby_opt frame_opt */
   -6,  /* (296) window ::= nm PARTITION BY nexprlist orderby_opt frame_opt */
   -4,  /* (297) window ::= ORDER BY sortlist frame_opt */
   -5,  /* (298) window ::= nm ORDER BY sortlist frame_opt */
   -1,  /* (299) window ::= frame_opt */
   -2,  /* (300) window ::= nm frame_opt */
    0,  /* (301) frame_opt ::= */
   -3,  /* (302) frame_opt ::= range_or_rows frame_bound_s frame_exclude_opt */
   -6,  /* (303) frame_opt ::= range_or_rows BETWEEN frame_bound_s AND frame_bound_e frame_exclude_opt */
   -1,  /* (304) range_or_rows ::= RANGE|ROWS|GROUPS */
   -1,  /* (305) frame_bound_s ::= frame_bound */
   -2,  /* (306) frame_bound_s ::= UNBOUNDED PRECEDING */
   -1,  /* (307) frame_bound_e ::= frame_bound */
   -2,  /* (308) frame_bound_e ::= UNBOUNDED FOLLOWING */
   -2,  /* (309) frame_bound ::= expr PRECEDING|FOLLOWING */
   -2,  /* (310) frame_bound ::= CURRENT ROW */
    0,  /* (311) frame_exclude_opt ::= */
   -2,  /* (312) frame_exclude_opt ::= EXCLUDE frame_exclude */
   -2,  /* (313) frame_exclude ::= NO OTHERS */
   -2,  /* (314) frame_exclude ::= CURRENT ROW */
   -1,  /* (315) frame_exclude ::= GROUP|TIES */
   -2,  /* (316) window_clause ::= WINDOW windowdefn_list */
   -2,  /* (317) filter_over ::= filter_clause over_clause */
   -1,  /* (318) filter_over ::= over_clause */
   -1,  /* (319) filter_over ::= filter_clause */
   -4,  /* (320) over_clause ::= OVER LP window RP */
   -2,  /* (321) over_clause ::= OVER nm */
   -5,  /* (322) filter_clause ::= FILTER LP WHERE expr RP */
   -1,  /* (323) input ::= cmdlist */
   -2,  /* (324) cmdlist ::= cmdlist ecmd */
   -1,  /* (325) cmdlist ::= ecmd */
   -1,  /* (326) ecmd ::= SEMI */
   -2,  /* (327) ecmd ::= cmdx SEMI */
   -2,  /* (328) ecmd ::= explain cmdx */
    0,  /* (329) trans_opt ::= */
   -1,  /* (330) trans_opt ::= TRANSACTION */
   -2,  /* (331) trans_opt ::= TRANSACTION nm */
   -1,  /* (332) savepoint_opt ::= SAVEPOINT */
    0,  /* (333) savepoint_opt ::= */
   -2,  /* (334) cmd ::= create_table create_table_args */
   -4,  /* (335) columnlist ::= columnlist COMMA columnname carglist */
   -2,  /* (336) columnlist ::= columnname carglist */
   -1,  /* (337) nm ::= ID|INDEXED */
   -1,  /* (338) nm ::= STRING */
   -1,  /* (339) nm ::= JOIN_KW */
   -1,  /* (340) typetoken ::= typename */
   -1,  /* (341) typename ::= ID|STRING */
   -1,  /* (342) signed ::= plus_num */
   -1,  /* (343) signed ::= minus_num */
   -2,  /* (344) carglist ::= carglist ccons */
    0,  /* (345) carglist ::= */
   -2,  /* (346) ccons ::= NULL onconf */
   -2,  /* (347) conslist_opt ::= COMMA conslist */
   -3,  /* (348) conslist ::= conslist tconscomma tcons */
   -1,  /* (349) conslist ::= tcons */
    0,  /* (350) tconscomma ::= */
   -1,  /* (351) defer_subclause_opt ::= defer_subclause */
   -1,  /* (352) resolvetype ::= raisetype */
   -1,  /* (353) selectnowith ::= oneselect */
   -1,  /* (354) oneselect ::= values */
   -2,  /* (355) sclp ::= selcollist COMMA */
   -1,  /* (356) as ::= ID|STRING */
   -1,  /* (357) expr ::= term */
   -1,  /* (358) likeop ::= LIKE_KW|MATCH */
   -1,  /* (359) exprlist ::= nexprlist */
   -1,  /* (360) nmnum ::= plus_num */
   -1,  /* (361) nmnum ::= nm */
   -1,  /* (362) nmnum ::= ON */
   -1,  /* (363) nmnum ::= DELETE */
   -1,  /* (364) nmnum ::= DEFAULT */
   -1,  /* (365) plus_num ::= INTEGER|FLOAT */
    0,  /* (366) foreach_clause ::= */
   -3,  /* (367) foreach_clause ::= FOR EACH ROW */
   -1,  /* (368) trnm ::= nm */
    0,  /* (369) tridxby ::= */
   -1,  /* (370) database_kw_opt ::= DATABASE */
    0,  /* (371) database_kw_opt ::= */
    0,  /* (372) kwcolumn_opt ::= */
   -1,  /* (373) kwcolumn_opt ::= COLUMNKW */
   -1,  /* (374) vtabarglist ::= vtabarg */
   -3,  /* (375) vtabarglist ::= vtabarglist COMMA vtabarg */
   -2,  /* (376) vtabarg ::= vtabarg vtabargtoken */
    0,  /* (377) anylist ::= */
   -4,  /* (378) anylist ::= anylist LP anylist RP */
   -2,  /* (379) anylist ::= anylist ANY */
    0,  /* (380) with ::= */
};

static void yy_accept(yyParser*);  /* Forward Declaration */

/*
** Perform a reduce action and the shift that must immediately
** follow the reduce.
**
** The yyLookahead and yyLookaheadToken parameters provide reduce actions
** access to the lookahead token (if any).  The yyLookahead will be YYNOCODE
** if the lookahead token has already been consumed.  As this procedure is
** only called from one place, optimizing compilers will in-line it, which
** means that the extra parameters have no performance impact.
*/
static YYACTIONTYPE yy_reduce(
  yyParser *yypParser,         /* The parser */
  unsigned int yyruleno,       /* Number of the rule by which to reduce */
  int yyLookahead,             /* Lookahead token, or YYNOCODE if none */
  sqlite3ParserTOKENTYPE yyLookaheadToken  /* Value of the lookahead token */
  sqlite3ParserCTX_PDECL                   /* %extra_context */
){
  int yygoto;                     /* The next state */
  YYACTIONTYPE yyact;             /* The next action */
  yyStackEntry *yymsp;            /* The top of the parser's stack */
  int yysize;                     /* Amount to pop the stack */
  sqlite3ParserARG_FETCH
  (void)yyLookahead;
  (void)yyLookaheadToken;
  yymsp = yypParser->yytos;
#ifndef NDEBUG
  if( yyTraceFILE && yyruleno<(int)(sizeof(yyRuleName)/sizeof(yyRuleName[0])) ){
    yysize = yyRuleInfoNRhs[yyruleno];
    if( yysize ){
      fprintf(yyTraceFILE, "%sReduce %d [%s], go to state %d.\n",
        yyTracePrompt,
        yyruleno, yyRuleName[yyruleno], yymsp[yysize].stateno);
    }else{
      fprintf(yyTraceFILE, "%sReduce %d [%s].\n",
        yyTracePrompt, yyruleno, yyRuleName[yyruleno]);
    }
  }
#endif /* NDEBUG */

  /* Check that the stack is large enough to grow by a single entry
  ** if the RHS of the rule is empty.  This ensures that there is room
  ** enough on the stack to push the LHS value */
  if( yyRuleInfoNRhs[yyruleno]==0 ){
#ifdef YYTRACKMAXSTACKDEPTH
    if( (int)(yypParser->yytos - yypParser->yystack)>yypParser->yyhwm ){
      yypParser->yyhwm++;
      assert( yypParser->yyhwm == (int)(yypParser->yytos - yypParser->yystack));
    }
#endif
#if YYSTACKDEPTH>0 
    if( yypParser->yytos>=yypParser->yystackEnd ){
      yyStackOverflow(yypParser);
      /* The call to yyStackOverflow() above pops the stack until it is
      ** empty, causing the main parser loop to exit.  So the return value
      ** is never used and does not matter. */
      return 0;
    }
#else
    if( yypParser->yytos>=&yypParser->yystack[yypParser->yystksz-1] ){
      if( yyGrowStack(yypParser) ){
        yyStackOverflow(yypParser);
        /* The call to yyStackOverflow() above pops the stack until it is
        ** empty, causing the main parser loop to exit.  So the return value
        ** is never used and does not matter. */
        return 0;
      }
      yymsp = yypParser->yytos;
    }
#endif
  }

  switch( yyruleno ){
  /* Beginning here are the reduction cases.  A typical example
  ** follows:
  **   case 0:
  **  #line <lineno> <grammarfile>
  **     { ... }           // User supplied code
  **  #line <lineno> <thisfile>
  **     break;
  */
/********** Begin reduce actions **********************************************/
        YYMINORTYPE yylhsminor;
      case 0: /* explain ::= EXPLAIN */
#line 123 "parse.y"
{ pParse->explain = 1; }
#line 3306 "parse.c"
        break;
      case 1: /* explain ::= EXPLAIN QUERY PLAN */
#line 124 "parse.y"
{ pParse->explain = 2; }
#line 3311 "parse.c"
        break;
      case 2: /* cmdx ::= cmd */
#line 126 "parse.y"
{ sqlite3FinishCoding(pParse); }
#line 3316 "parse.c"
        break;
      case 3: /* cmd ::= BEGIN transtype trans_opt */
#line 131 "parse.y"
{sqlite3BeginTransaction(pParse, yymsp[-1].minor.yy32);}
#line 3321 "parse.c"
        break;
      case 4: /* transtype ::= */
#line 136 "parse.y"
{yymsp[1].minor.yy32 = TK_DEFERRED;}
#line 3326 "parse.c"
        break;
      case 5: /* transtype ::= DEFERRED */
      case 6: /* transtype ::= IMMEDIATE */ yytestcase(yyruleno==6);
      case 7: /* transtype ::= EXCLUSIVE */ yytestcase(yyruleno==7);
      case 304: /* range_or_rows ::= RANGE|ROWS|GROUPS */ yytestcase(yyruleno==304);
#line 137 "parse.y"
{yymsp[0].minor.yy32 = yymsp[0].major; /*A-overwrites-X*/}
#line 3334 "parse.c"
        break;
      case 8: /* cmd ::= COMMIT|END trans_opt */
      case 9: /* cmd ::= ROLLBACK trans_opt */ yytestcase(yyruleno==9);
#line 140 "parse.y"
{sqlite3EndTransaction(pParse,yymsp[-1].major);}
#line 3340 "parse.c"
        break;
      case 10: /* cmd ::= SAVEPOINT nm */
#line 145 "parse.y"
{
  sqlite3Savepoint(pParse, SAVEPOINT_BEGIN, &yymsp[0].minor.yy0);
}
#line 3347 "parse.c"
        break;
      case 11: /* cmd ::= RELEASE savepoint_opt nm */
#line 148 "parse.y"
{
  sqlite3Savepoint(pParse, SAVEPOINT_RELEASE, &yymsp[0].minor.yy0);
}
#line 3354 "parse.c"
        break;
      case 12: /* cmd ::= ROLLBACK trans_opt TO savepoint_opt nm */
#line 151 "parse.y"
{
  sqlite3Savepoint(pParse, SAVEPOINT_ROLLBACK, &yymsp[0].minor.yy0);
}
#line 3361 "parse.c"
        break;
      case 13: /* create_table ::= createkw temp TABLE ifnotexists nm dbnm */
#line 158 "parse.y"
{
   sqlite3StartTable(pParse,&yymsp[-1].minor.yy0,&yymsp[0].minor.yy0,yymsp[-4].minor.yy32,0,0,yymsp[-2].minor.yy32);
}
#line 3368 "parse.c"
        break;
      case 14: /* createkw ::= CREATE */
#line 161 "parse.y"
{disableLookaside(pParse);}
#line 3373 "parse.c"
        break;
      case 15: /* ifnotexists ::= */
      case 18: /* temp ::= */ yytestcase(yyruleno==18);
      case 21: /* table_options ::= */ yytestcase(yyruleno==21);
      case 43: /* autoinc ::= */ yytestcase(yyruleno==43);
      case 58: /* init_deferred_pred_opt ::= */ yytestcase(yyruleno==58);
      case 68: /* defer_subclause_opt ::= */ yytestcase(yyruleno==68);
      case 77: /* ifexists ::= */ yytestcase(yyruleno==77);
      case 94: /* distinct ::= */ yytestcase(yyruleno==94);
      case 230: /* collate ::= */ yytestcase(yyruleno==230);
#line 164 "parse.y"
{yymsp[1].minor.yy32 = 0;}
#line 3386 "parse.c"
        break;
      case 16: /* ifnotexists ::= IF NOT EXISTS */
#line 165 "parse.y"
{yymsp[-2].minor.yy32 = 1;}
#line 3391 "parse.c"
        break;
      case 17: /* temp ::= TEMP */
      case 44: /* autoinc ::= AUTOINCR */ yytestcase(yyruleno==44);
#line 168 "parse.y"
{yymsp[0].minor.yy32 = 1;}
#line 3397 "parse.c"
        break;
      case 19: /* create_table_args ::= LP columnlist conslist_opt RP table_options */
#line 171 "parse.y"
{
  sqlite3EndTable(pParse,&yymsp[-2].minor.yy0,&yymsp[-1].minor.yy0,yymsp[0].minor.yy32,0);
}
#line 3404 "parse.c"
        break;
      case 20: /* create_table_args ::= AS select */
#line 174 "parse.y"
{
  sqlite3EndTable(pParse,0,0,0,yymsp[0].minor.yy25);
  sqlite3SelectDelete(pParse->db, yymsp[0].minor.yy25);
}
#line 3412 "parse.c"
        break;
      case 22: /* table_options ::= WITHOUT nm */
#line 180 "parse.y"
{
  if( yymsp[0].minor.yy0.n==5 && sqlite3_strnicmp(yymsp[0].minor.yy0.z,"rowid",5)==0 ){
    yymsp[-1].minor.yy32 = TF_WithoutRowid | TF_NoVisibleRowid;
  }else{
    yymsp[-1].minor.yy32 = 0;
    sqlite3ErrorMsg(pParse, "unknown table option: %.*s", yymsp[0].minor.yy0.n, yymsp[0].minor.yy0.z);
  }
}
#line 3424 "parse.c"
        break;
      case 23: /* columnname ::= nm typetoken */
#line 190 "parse.y"
{sqlite3AddColumn(pParse,&yymsp[-1].minor.yy0,&yymsp[0].minor.yy0);}
#line 3429 "parse.c"
        break;
      case 24: /* typetoken ::= */
      case 61: /* conslist_opt ::= */ yytestcase(yyruleno==61);
      case 100: /* as ::= */ yytestcase(yyruleno==100);
#line 273 "parse.y"
{yymsp[1].minor.yy0.n = 0; yymsp[1].minor.yy0.z = 0;}
#line 3436 "parse.c"
        break;
      case 25: /* typetoken ::= typename LP signed RP */
#line 275 "parse.y"
{
  yymsp[-3].minor.yy0.n = (int)(&yymsp[0].minor.yy0.z[yymsp[0].minor.yy0.n] - yymsp[-3].minor.yy0.z);
}
#line 3443 "parse.c"
        break;
      case 26: /* typetoken ::= typename LP signed COMMA signed RP */
#line 278 "parse.y"
{
  yymsp[-5].minor.yy0.n = (int)(&yymsp[0].minor.yy0.z[yymsp[0].minor.yy0.n] - yymsp[-5].minor.yy0.z);
}
#line 3450 "parse.c"
        break;
      case 27: /* typename ::= typename ID|STRING */
#line 283 "parse.y"
{yymsp[-1].minor.yy0.n=yymsp[0].minor.yy0.n+(int)(yymsp[0].minor.yy0.z-yymsp[-1].minor.yy0.z);}
#line 3455 "parse.c"
        break;
      case 28: /* scanpt ::= */
#line 301 "parse.y"
{
  assert( yyLookahead!=YYNOCODE );
  yymsp[1].minor.yy8 = yyLookaheadToken.z;
}
#line 3463 "parse.c"
        break;
      case 29: /* scantok ::= */
#line 305 "parse.y"
{
  assert( yyLookahead!=YYNOCODE );
  yymsp[1].minor.yy0 = yyLookaheadToken;
}
#line 3471 "parse.c"
        break;
      case 30: /* ccons ::= CONSTRAINT nm */
      case 63: /* tcons ::= CONSTRAINT nm */ yytestcase(yyruleno==63);
#line 315 "parse.y"
{pParse->constraintName = yymsp[0].minor.yy0;}
#line 3477 "parse.c"
        break;
      case 31: /* ccons ::= DEFAULT scantok term */
#line 317 "parse.y"
{sqlite3AddDefaultValue(pParse,yymsp[0].minor.yy46,yymsp[-1].minor.yy0.z,&yymsp[-1].minor.yy0.z[yymsp[-1].minor.yy0.n]);}
#line 3482 "parse.c"
        break;
      case 32: /* ccons ::= DEFAULT LP expr RP */
#line 319 "parse.y"
{sqlite3AddDefaultValue(pParse,yymsp[-1].minor.yy46,yymsp[-2].minor.yy0.z+1,yymsp[0].minor.yy0.z);}
#line 3487 "parse.c"
        break;
      case 33: /* ccons ::= DEFAULT PLUS scantok term */
#line 321 "parse.y"
{sqlite3AddDefaultValue(pParse,yymsp[0].minor.yy46,yymsp[-2].minor.yy0.z,&yymsp[-1].minor.yy0.z[yymsp[-1].minor.yy0.n]);}
#line 3492 "parse.c"
        break;
      case 34: /* ccons ::= DEFAULT MINUS scantok term */
#line 322 "parse.y"
{
  Expr *p = sqlite3PExpr(pParse, TK_UMINUS, yymsp[0].minor.yy46, 0);
  sqlite3AddDefaultValue(pParse,p,yymsp[-2].minor.yy0.z,&yymsp[-1].minor.yy0.z[yymsp[-1].minor.yy0.n]);
}
#line 3500 "parse.c"
        break;
      case 35: /* ccons ::= DEFAULT scantok ID|INDEXED */
#line 326 "parse.y"
{
  Expr *p = tokenExpr(pParse, TK_STRING, yymsp[0].minor.yy0);
  if( p ){
    sqlite3ExprIdToTrueFalse(p);
    testcase( p->op==TK_TRUEFALSE && sqlite3ExprTruthValue(p) );
  }
    sqlite3AddDefaultValue(pParse,p,yymsp[0].minor.yy0.z,yymsp[0].minor.yy0.z+yymsp[0].minor.yy0.n);
}
#line 3512 "parse.c"
        break;
      case 36: /* ccons ::= NOT NULL onconf */
#line 339 "parse.y"
{sqlite3AddNotNull(pParse, yymsp[0].minor.yy32);}
#line 3517 "parse.c"
        break;
      case 37: /* ccons ::= PRIMARY KEY sortorder onconf autoinc */
#line 341 "parse.y"
{sqlite3AddPrimaryKey(pParse,0,yymsp[-1].minor.yy32,yymsp[0].minor.yy32,yymsp[-2].minor.yy32);}
#line 3522 "parse.c"
        break;
      case 38: /* ccons ::= UNIQUE onconf */
#line 342 "parse.y"
{sqlite3CreateIndex(pParse,0,0,0,0,yymsp[0].minor.yy32,0,0,0,0,
                                   SQLITE_IDXTYPE_UNIQUE);}
#line 3528 "parse.c"
        break;
      case 39: /* ccons ::= CHECK LP expr RP */
#line 344 "parse.y"
{sqlite3AddCheckConstraint(pParse,yymsp[-1].minor.yy46);}
#line 3533 "parse.c"
        break;
      case 40: /* ccons ::= REFERENCES nm eidlist_opt refargs */
#line 346 "parse.y"
{sqlite3CreateForeignKey(pParse,0,&yymsp[-2].minor.yy0,yymsp[-1].minor.yy138,yymsp[0].minor.yy32);}
#line 3538 "parse.c"
        break;
      case 41: /* ccons ::= defer_subclause */
#line 347 "parse.y"
{sqlite3DeferForeignKey(pParse,yymsp[0].minor.yy32);}
#line 3543 "parse.c"
        break;
      case 42: /* ccons ::= COLLATE ID|STRING */
#line 348 "parse.y"
{sqlite3AddCollateType(pParse, &yymsp[0].minor.yy0);}
#line 3548 "parse.c"
        break;
      case 45: /* refargs ::= */
#line 361 "parse.y"
{ yymsp[1].minor.yy32 = OE_None*0x0101; /* EV: R-19803-45884 */}
#line 3553 "parse.c"
        break;
      case 46: /* refargs ::= refargs refarg */
#line 362 "parse.y"
{ yymsp[-1].minor.yy32 = (yymsp[-1].minor.yy32 & ~yymsp[0].minor.yy495.mask) | yymsp[0].minor.yy495.value; }
#line 3558 "parse.c"
        break;
      case 47: /* refarg ::= MATCH nm */
#line 364 "parse.y"
{ yymsp[-1].minor.yy495.value = 0;     yymsp[-1].minor.yy495.mask = 0x000000; }
#line 3563 "parse.c"
        break;
      case 48: /* refarg ::= ON INSERT refact */
#line 365 "parse.y"
{ yymsp[-2].minor.yy495.value = 0;     yymsp[-2].minor.yy495.mask = 0x000000; }
#line 3568 "parse.c"
        break;
      case 49: /* refarg ::= ON DELETE refact */
#line 366 "parse.y"
{ yymsp[-2].minor.yy495.value = yymsp[0].minor.yy32;     yymsp[-2].minor.yy495.mask = 0x0000ff; }
#line 3573 "parse.c"
        break;
      case 50: /* refarg ::= ON UPDATE refact */
#line 367 "parse.y"
{ yymsp[-2].minor.yy495.value = yymsp[0].minor.yy32<<8;  yymsp[-2].minor.yy495.mask = 0x00ff00; }
#line 3578 "parse.c"
        break;
      case 51: /* refact ::= SET NULL */
#line 369 "parse.y"
{ yymsp[-1].minor.yy32 = OE_SetNull;  /* EV: R-33326-45252 */}
#line 3583 "parse.c"
        break;
      case 52: /* refact ::= SET DEFAULT */
#line 370 "parse.y"
{ yymsp[-1].minor.yy32 = OE_SetDflt;  /* EV: R-33326-45252 */}
#line 3588 "parse.c"
        break;
      case 53: /* refact ::= CASCADE */
#line 371 "parse.y"
{ yymsp[0].minor.yy32 = OE_Cascade;  /* EV: R-33326-45252 */}
#line 3593 "parse.c"
        break;
      case 54: /* refact ::= RESTRICT */
#line 372 "parse.y"
{ yymsp[0].minor.yy32 = OE_Restrict; /* EV: R-33326-45252 */}
#line 3598 "parse.c"
        break;
      case 55: /* refact ::= NO ACTION */
#line 373 "parse.y"
{ yymsp[-1].minor.yy32 = OE_None;     /* EV: R-33326-45252 */}
#line 3603 "parse.c"
        break;
      case 56: /* defer_subclause ::= NOT DEFERRABLE init_deferred_pred_opt */
#line 375 "parse.y"
{yymsp[-2].minor.yy32 = 0;}
#line 3608 "parse.c"
        break;
      case 57: /* defer_subclause ::= DEFERRABLE init_deferred_pred_opt */
      case 72: /* orconf ::= OR resolvetype */ yytestcase(yyruleno==72);
      case 160: /* insert_cmd ::= INSERT orconf */ yytestcase(yyruleno==160);
#line 376 "parse.y"
{yymsp[-1].minor.yy32 = yymsp[0].minor.yy32;}
#line 3615 "parse.c"
        break;
      case 59: /* init_deferred_pred_opt ::= INITIALLY DEFERRED */
      case 76: /* ifexists ::= IF EXISTS */ yytestcase(yyruleno==76);
      case 202: /* between_op ::= NOT BETWEEN */ yytestcase(yyruleno==202);
      case 205: /* in_op ::= NOT IN */ yytestcase(yyruleno==205);
      case 231: /* collate ::= COLLATE ID|STRING */ yytestcase(yyruleno==231);
#line 379 "parse.y"
{yymsp[-1].minor.yy32 = 1;}
#line 3624 "parse.c"
        break;
      case 60: /* init_deferred_pred_opt ::= INITIALLY IMMEDIATE */
#line 380 "parse.y"
{yymsp[-1].minor.yy32 = 0;}
#line 3629 "parse.c"
        break;
      case 62: /* tconscomma ::= COMMA */
#line 386 "parse.y"
{pParse->constraintName.n = 0;}
#line 3634 "parse.c"
        break;
      case 64: /* tcons ::= PRIMARY KEY LP sortlist autoinc RP onconf */
#line 390 "parse.y"
{sqlite3AddPrimaryKey(pParse,yymsp[-3].minor.yy138,yymsp[0].minor.yy32,yymsp[-2].minor.yy32,0);}
#line 3639 "parse.c"
        break;
      case 65: /* tcons ::= UNIQUE LP sortlist RP onconf */
#line 392 "parse.y"
{sqlite3CreateIndex(pParse,0,0,0,yymsp[-2].minor.yy138,yymsp[0].minor.yy32,0,0,0,0,
                                       SQLITE_IDXTYPE_UNIQUE);}
#line 3645 "parse.c"
        break;
      case 66: /* tcons ::= CHECK LP expr RP onconf */
#line 395 "parse.y"
{sqlite3AddCheckConstraint(pParse,yymsp[-2].minor.yy46);}
#line 3650 "parse.c"
        break;
      case 67: /* tcons ::= FOREIGN KEY LP eidlist RP REFERENCES nm eidlist_opt refargs defer_subclause_opt */
#line 397 "parse.y"
{
    sqlite3CreateForeignKey(pParse, yymsp[-6].minor.yy138, &yymsp[-3].minor.yy0, yymsp[-2].minor.yy138, yymsp[-1].minor.yy32);
    sqlite3DeferForeignKey(pParse, yymsp[0].minor.yy32);
}
#line 3658 "parse.c"
        break;
      case 69: /* onconf ::= */
      case 71: /* orconf ::= */ yytestcase(yyruleno==71);
#line 411 "parse.y"
{yymsp[1].minor.yy32 = OE_Default;}
#line 3664 "parse.c"
        break;
      case 70: /* onconf ::= ON CONFLICT resolvetype */
#line 412 "parse.y"
{yymsp[-2].minor.yy32 = yymsp[0].minor.yy32;}
#line 3669 "parse.c"
        break;
      case 73: /* resolvetype ::= IGNORE */
#line 416 "parse.y"
{yymsp[0].minor.yy32 = OE_Ignore;}
#line 3674 "parse.c"
        break;
      case 74: /* resolvetype ::= REPLACE */
      case 161: /* insert_cmd ::= REPLACE */ yytestcase(yyruleno==161);
#line 417 "parse.y"
{yymsp[0].minor.yy32 = OE_Replace;}
#line 3680 "parse.c"
        break;
      case 75: /* cmd ::= DROP TABLE ifexists fullname */
#line 421 "parse.y"
{
  sqlite3DropTable(pParse, yymsp[0].minor.yy609, 0, yymsp[-1].minor.yy32);
}
#line 3687 "parse.c"
        break;
      case 78: /* cmd ::= createkw temp VIEW ifnotexists nm dbnm eidlist_opt AS select */
#line 432 "parse.y"
{
  sqlite3CreateView(pParse, &yymsp[-8].minor.yy0, &yymsp[-4].minor.yy0, &yymsp[-3].minor.yy0, yymsp[-2].minor.yy138, yymsp[0].minor.yy25, yymsp[-7].minor.yy32, yymsp[-5].minor.yy32);
}
#line 3694 "parse.c"
        break;
      case 79: /* cmd ::= DROP VIEW ifexists fullname */
#line 435 "parse.y"
{
  sqlite3DropTable(pParse, yymsp[0].minor.yy609, 1, yymsp[-1].minor.yy32);
}
#line 3701 "parse.c"
        break;
      case 80: /* cmd ::= select */
#line 442 "parse.y"
{
  SelectDest dest = {SRT_Output, 0, 0, 0, 0, 0};
  sqlite3Select(pParse, yymsp[0].minor.yy25, &dest);
  sqlite3SelectDelete(pParse->db, yymsp[0].minor.yy25);
}
#line 3710 "parse.c"
        break;
      case 81: /* select ::= WITH wqlist selectnowith */
#line 481 "parse.y"
{
  Select *p = yymsp[0].minor.yy25;
  if( p ){
    p->pWith = yymsp[-1].minor.yy297;
    parserDoubleLinkSelect(pParse, p);
  }else{
    sqlite3WithDelete(pParse->db, yymsp[-1].minor.yy297);
  }
  yymsp[-2].minor.yy25 = p;
}
#line 3724 "parse.c"
        break;
      case 82: /* select ::= WITH RECURSIVE wqlist selectnowith */
#line 491 "parse.y"
{
  Select *p = yymsp[0].minor.yy25;
  if( p ){
    p->pWith = yymsp[-1].minor.yy297;
    parserDoubleLinkSelect(pParse, p);
  }else{
    sqlite3WithDelete(pParse->db, yymsp[-1].minor.yy297);
  }
  yymsp[-3].minor.yy25 = p;
}
#line 3738 "parse.c"
        break;
      case 83: /* select ::= selectnowith */
#line 502 "parse.y"
{
  Select *p = yymsp[0].minor.yy25;
  if( p ){
    parserDoubleLinkSelect(pParse, p);
  }
  yymsp[0].minor.yy25 = p; /*A-overwrites-X*/
}
#line 3749 "parse.c"
        break;
      case 84: /* selectnowith ::= selectnowith multiselect_op oneselect */
#line 512 "parse.y"
{
  Select *pRhs = yymsp[0].minor.yy25;
  Select *pLhs = yymsp[-2].minor.yy25;
  if( pRhs && pRhs->pPrior ){
    SrcList *pFrom;
    Token x;
    x.n = 0;
    parserDoubleLinkSelect(pParse, pRhs);
    pFrom = sqlite3SrcListAppendFromTerm(pParse,0,0,0,&x,pRhs,0,0);
    pRhs = sqlite3SelectNew(pParse,0,pFrom,0,0,0,0,0,0);
  }
  if( pRhs ){
    pRhs->op = (u8)yymsp[-1].minor.yy32;
    pRhs->pPrior = pLhs;
    if( ALWAYS(pLhs) ) pLhs->selFlags &= ~SF_MultiValue;
    pRhs->selFlags &= ~SF_MultiValue;
    if( yymsp[-1].minor.yy32!=TK_ALL ) pParse->hasCompound = 1;
  }else{
    sqlite3SelectDelete(pParse->db, pLhs);
  }
  yymsp[-2].minor.yy25 = pRhs;
}
#line 3775 "parse.c"
        break;
      case 85: /* multiselect_op ::= UNION */
      case 87: /* multiselect_op ::= EXCEPT|INTERSECT */ yytestcase(yyruleno==87);
#line 535 "parse.y"
{yymsp[0].minor.yy32 = yymsp[0].major; /*A-overwrites-OP*/}
#line 3781 "parse.c"
        break;
      case 86: /* multiselect_op ::= UNION ALL */
#line 536 "parse.y"
{yymsp[-1].minor.yy32 = TK_ALL;}
#line 3786 "parse.c"
        break;
      case 88: /* oneselect ::= SELECT distinct selcollist from where_opt groupby_opt having_opt orderby_opt limit_opt */
#line 542 "parse.y"
{
  yymsp[-8].minor.yy25 = sqlite3SelectNew(pParse,yymsp[-6].minor.yy138,yymsp[-5].minor.yy609,yymsp[-4].minor.yy46,yymsp[-3].minor.yy138,yymsp[-2].minor.yy46,yymsp[-1].minor.yy138,yymsp[-7].minor.yy32,yymsp[0].minor.yy46);
}
#line 3793 "parse.c"
        break;
      case 89: /* oneselect ::= SELECT distinct selcollist from where_opt groupby_opt having_opt window_clause orderby_opt limit_opt */
#line 548 "parse.y"
{
  yymsp[-9].minor.yy25 = sqlite3SelectNew(pParse,yymsp[-7].minor.yy138,yymsp[-6].minor.yy609,yymsp[-5].minor.yy46,yymsp[-4].minor.yy138,yymsp[-3].minor.yy46,yymsp[-1].minor.yy138,yymsp[-8].minor.yy32,yymsp[0].minor.yy46);
  if( yymsp[-9].minor.yy25 ){
    yymsp[-9].minor.yy25->pWinDefn = yymsp[-2].minor.yy455;
  }else{
    sqlite3WindowListDelete(pParse->db, yymsp[-2].minor.yy455);
  }
}
#line 3805 "parse.c"
        break;
      case 90: /* values ::= VALUES LP nexprlist RP */
#line 563 "parse.y"
{
  yymsp[-3].minor.yy25 = sqlite3SelectNew(pParse,yymsp[-1].minor.yy138,0,0,0,0,0,SF_Values,0);
}
#line 3812 "parse.c"
        break;
      case 91: /* values ::= values COMMA LP nexprlist RP */
#line 566 "parse.y"
{
  Select *pRight, *pLeft = yymsp[-4].minor.yy25;
  pRight = sqlite3SelectNew(pParse,yymsp[-1].minor.yy138,0,0,0,0,0,SF_Values|SF_MultiValue,0);
  if( ALWAYS(pLeft) ) pLeft->selFlags &= ~SF_MultiValue;
  if( pRight ){
    pRight->op = TK_ALL;
    pRight->pPrior = pLeft;
    yymsp[-4].minor.yy25 = pRight;
  }else{
    yymsp[-4].minor.yy25 = pLeft;
  }
}
#line 3828 "parse.c"
        break;
      case 92: /* distinct ::= DISTINCT */
#line 583 "parse.y"
{yymsp[0].minor.yy32 = SF_Distinct;}
#line 3833 "parse.c"
        break;
      case 93: /* distinct ::= ALL */
#line 584 "parse.y"
{yymsp[0].minor.yy32 = SF_All;}
#line 3838 "parse.c"
        break;
      case 95: /* sclp ::= */
      case 128: /* orderby_opt ::= */ yytestcase(yyruleno==128);
      case 138: /* groupby_opt ::= */ yytestcase(yyruleno==138);
      case 218: /* exprlist ::= */ yytestcase(yyruleno==218);
      case 221: /* paren_exprlist ::= */ yytestcase(yyruleno==221);
      case 226: /* eidlist_opt ::= */ yytestcase(yyruleno==226);
#line 597 "parse.y"
{yymsp[1].minor.yy138 = 0;}
#line 3848 "parse.c"
        break;
      case 96: /* selcollist ::= sclp scanpt expr scanpt as */
#line 598 "parse.y"
{
   yymsp[-4].minor.yy138 = sqlite3ExprListAppend(pParse, yymsp[-4].minor.yy138, yymsp[-2].minor.yy46);
   if( yymsp[0].minor.yy0.n>0 ) sqlite3ExprListSetName(pParse, yymsp[-4].minor.yy138, &yymsp[0].minor.yy0, 1);
   sqlite3ExprListSetSpan(pParse,yymsp[-4].minor.yy138,yymsp[-3].minor.yy8,yymsp[-1].minor.yy8);
}
#line 3857 "parse.c"
        break;
      case 97: /* selcollist ::= sclp scanpt STAR */
#line 603 "parse.y"
{
  Expr *p = sqlite3Expr(pParse->db, TK_ASTERISK, 0);
  yymsp[-2].minor.yy138 = sqlite3ExprListAppend(pParse, yymsp[-2].minor.yy138, p);
}
#line 3865 "parse.c"
        break;
      case 98: /* selcollist ::= sclp scanpt nm DOT STAR */
#line 607 "parse.y"
{
  Expr *pRight = sqlite3PExpr(pParse, TK_ASTERISK, 0, 0);
  Expr *pLeft = sqlite3ExprAlloc(pParse->db, TK_ID, &yymsp[-2].minor.yy0, 1);
  Expr *pDot = sqlite3PExpr(pParse, TK_DOT, pLeft, pRight);
  yymsp[-4].minor.yy138 = sqlite3ExprListAppend(pParse,yymsp[-4].minor.yy138, pDot);
}
#line 3875 "parse.c"
        break;
      case 99: /* as ::= AS nm */
      case 110: /* dbnm ::= DOT nm */ yytestcase(yyruleno==110);
      case 242: /* plus_num ::= PLUS INTEGER|FLOAT */ yytestcase(yyruleno==242);
      case 243: /* minus_num ::= MINUS INTEGER|FLOAT */ yytestcase(yyruleno==243);
#line 618 "parse.y"
{yymsp[-1].minor.yy0 = yymsp[0].minor.yy0;}
#line 3883 "parse.c"
        break;
      case 101: /* from ::= */
#line 632 "parse.y"
{yymsp[1].minor.yy609 = sqlite3DbMallocZero(pParse->db, sizeof(*yymsp[1].minor.yy609));}
#line 3888 "parse.c"
        break;
      case 102: /* from ::= FROM seltablist */
#line 633 "parse.y"
{
  yymsp[-1].minor.yy609 = yymsp[0].minor.yy609;
  sqlite3SrcListShiftJoinType(yymsp[-1].minor.yy609);
}
#line 3896 "parse.c"
        break;
      case 103: /* stl_prefix ::= seltablist joinop */
#line 641 "parse.y"
{
   if( ALWAYS(yymsp[-1].minor.yy609 && yymsp[-1].minor.yy609->nSrc>0) ) yymsp[-1].minor.yy609->a[yymsp[-1].minor.yy609->nSrc-1].fg.jointype = (u8)yymsp[0].minor.yy32;
}
#line 3903 "parse.c"
        break;
      case 104: /* stl_prefix ::= */
#line 644 "parse.y"
{yymsp[1].minor.yy609 = 0;}
#line 3908 "parse.c"
        break;
      case 105: /* seltablist ::= stl_prefix nm dbnm as indexed_opt on_opt using_opt */
#line 646 "parse.y"
{
  yymsp[-6].minor.yy609 = sqlite3SrcListAppendFromTerm(pParse,yymsp[-6].minor.yy609,&yymsp[-5].minor.yy0,&yymsp[-4].minor.yy0,&yymsp[-3].minor.yy0,0,yymsp[-1].minor.yy46,yymsp[0].minor.yy406);
  sqlite3SrcListIndexedBy(pParse, yymsp[-6].minor.yy609, &yymsp[-2].minor.yy0);
}
#line 3916 "parse.c"
        break;
      case 106: /* seltablist ::= stl_prefix nm dbnm LP exprlist RP as on_opt using_opt */
#line 651 "parse.y"
{
  yymsp[-8].minor.yy609 = sqlite3SrcListAppendFromTerm(pParse,yymsp[-8].minor.yy609,&yymsp[-7].minor.yy0,&yymsp[-6].minor.yy0,&yymsp[-2].minor.yy0,0,yymsp[-1].minor.yy46,yymsp[0].minor.yy406);
  sqlite3SrcListFuncArgs(pParse, yymsp[-8].minor.yy609, yymsp[-4].minor.yy138);
}
#line 3924 "parse.c"
        break;
      case 107: /* seltablist ::= stl_prefix LP select RP as on_opt using_opt */
#line 657 "parse.y"
{
    yymsp[-6].minor.yy609 = sqlite3SrcListAppendFromTerm(pParse,yymsp[-6].minor.yy609,0,0,&yymsp[-2].minor.yy0,yymsp[-4].minor.yy25,yymsp[-1].minor.yy46,yymsp[0].minor.yy406);
  }
#line 3931 "parse.c"
        break;
      case 108: /* seltablist ::= stl_prefix LP seltablist RP as on_opt using_opt */
#line 661 "parse.y"
{
    if( yymsp[-6].minor.yy609==0 && yymsp[-2].minor.yy0.n==0 && yymsp[-1].minor.yy46==0 && yymsp[0].minor.yy406==0 ){
      yymsp[-6].minor.yy609 = yymsp[-4].minor.yy609;
    }else if( yymsp[-4].minor.yy609->nSrc==1 ){
      yymsp[-6].minor.yy609 = sqlite3SrcListAppendFromTerm(pParse,yymsp[-6].minor.yy609,0,0,&yymsp[-2].minor.yy0,0,yymsp[-1].minor.yy46,yymsp[0].minor.yy406);
      if( yymsp[-6].minor.yy609 ){
        struct SrcList_item *pNew = &yymsp[-6].minor.yy609->a[yymsp[-6].minor.yy609->nSrc-1];
        struct SrcList_item *pOld = yymsp[-4].minor.yy609->a;
        pNew->zName = pOld->zName;
        pNew->zDatabase = pOld->zDatabase;
        pNew->pSelect = pOld->pSelect;
        if( pOld->fg.isTabFunc ){
          pNew->u1.pFuncArg = pOld->u1.pFuncArg;
          pOld->u1.pFuncArg = 0;
          pOld->fg.isTabFunc = 0;
          pNew->fg.isTabFunc = 1;
        }
        pOld->zName = pOld->zDatabase = 0;
        pOld->pSelect = 0;
      }
      sqlite3SrcListDelete(pParse->db, yymsp[-4].minor.yy609);
    }else{
      Select *pSubquery;
      sqlite3SrcListShiftJoinType(yymsp[-4].minor.yy609);
      pSubquery = sqlite3SelectNew(pParse,0,yymsp[-4].minor.yy609,0,0,0,0,SF_NestedFrom,0);
      yymsp[-6].minor.yy609 = sqlite3SrcListAppendFromTerm(pParse,yymsp[-6].minor.yy609,0,0,&yymsp[-2].minor.yy0,pSubquery,yymsp[-1].minor.yy46,yymsp[0].minor.yy406);
    }
  }
#line 3963 "parse.c"
        break;
      case 109: /* dbnm ::= */
      case 123: /* indexed_opt ::= */ yytestcase(yyruleno==123);
#line 692 "parse.y"
{yymsp[1].minor.yy0.z=0; yymsp[1].minor.yy0.n=0;}
#line 3969 "parse.c"
        break;
      case 111: /* fullname ::= nm */
#line 697 "parse.y"
{
  yylhsminor.yy609 = sqlite3SrcListAppend(pParse,0,&yymsp[0].minor.yy0,0);
  if( IN_RENAME_OBJECT && yylhsminor.yy609 ) sqlite3RenameTokenMap(pParse, yylhsminor.yy609->a[0].zName, &yymsp[0].minor.yy0);
}
#line 3977 "parse.c"
  yymsp[0].minor.yy609 = yylhsminor.yy609;
        break;
      case 112: /* fullname ::= nm DOT nm */
#line 701 "parse.y"
{
  yylhsminor.yy609 = sqlite3SrcListAppend(pParse,0,&yymsp[-2].minor.yy0,&yymsp[0].minor.yy0);
  if( IN_RENAME_OBJECT && yylhsminor.yy609 ) sqlite3RenameTokenMap(pParse, yylhsminor.yy609->a[0].zName, &yymsp[0].minor.yy0);
}
#line 3986 "parse.c"
  yymsp[-2].minor.yy609 = yylhsminor.yy609;
        break;
      case 113: /* xfullname ::= nm */
#line 709 "parse.y"
{yymsp[0].minor.yy609 = sqlite3SrcListAppend(pParse,0,&yymsp[0].minor.yy0,0); /*A-overwrites-X*/}
#line 3992 "parse.c"
        break;
      case 114: /* xfullname ::= nm DOT nm */
#line 711 "parse.y"
{yymsp[-2].minor.yy609 = sqlite3SrcListAppend(pParse,0,&yymsp[-2].minor.yy0,&yymsp[0].minor.yy0); /*A-overwrites-X*/}
#line 3997 "parse.c"
        break;
      case 115: /* xfullname ::= nm DOT nm AS nm */
#line 712 "parse.y"
{
   yymsp[-4].minor.yy609 = sqlite3SrcListAppend(pParse,0,&yymsp[-4].minor.yy0,&yymsp[-2].minor.yy0); /*A-overwrites-X*/
   if( yymsp[-4].minor.yy609 ) yymsp[-4].minor.yy609->a[0].zAlias = sqlite3NameFromToken(pParse->db, &yymsp[0].minor.yy0);
}
#line 4005 "parse.c"
        break;
      case 116: /* xfullname ::= nm AS nm */
#line 716 "parse.y"
{  
   yymsp[-2].minor.yy609 = sqlite3SrcListAppend(pParse,0,&yymsp[-2].minor.yy0,0); /*A-overwrites-X*/
   if( yymsp[-2].minor.yy609 ) yymsp[-2].minor.yy609->a[0].zAlias = sqlite3NameFromToken(pParse->db, &yymsp[0].minor.yy0);
}
#line 4013 "parse.c"
        break;
      case 117: /* joinop ::= COMMA|JOIN */
#line 722 "parse.y"
{ yymsp[0].minor.yy32 = JT_INNER; }
#line 4018 "parse.c"
        break;
      case 118: /* joinop ::= JOIN_KW JOIN */
#line 724 "parse.y"
{yymsp[-1].minor.yy32 = sqlite3JoinType(pParse,&yymsp[-1].minor.yy0,0,0);  /*X-overwrites-A*/}
#line 4023 "parse.c"
        break;
      case 119: /* joinop ::= JOIN_KW nm JOIN */
#line 726 "parse.y"
{yymsp[-2].minor.yy32 = sqlite3JoinType(pParse,&yymsp[-2].minor.yy0,&yymsp[-1].minor.yy0,0); /*X-overwrites-A*/}
#line 4028 "parse.c"
        break;
      case 120: /* joinop ::= JOIN_KW nm nm JOIN */
#line 728 "parse.y"
{yymsp[-3].minor.yy32 = sqlite3JoinType(pParse,&yymsp[-3].minor.yy0,&yymsp[-2].minor.yy0,&yymsp[-1].minor.yy0);/*X-overwrites-A*/}
#line 4033 "parse.c"
        break;
      case 121: /* on_opt ::= ON expr */
      case 141: /* having_opt ::= HAVING expr */ yytestcase(yyruleno==141);
      case 148: /* where_opt ::= WHERE expr */ yytestcase(yyruleno==148);
      case 214: /* case_else ::= ELSE expr */ yytestcase(yyruleno==214);
      case 235: /* vinto ::= INTO expr */ yytestcase(yyruleno==235);
#line 749 "parse.y"
{yymsp[-1].minor.yy46 = yymsp[0].minor.yy46;}
#line 4042 "parse.c"
        break;
      case 122: /* on_opt ::= */
      case 140: /* having_opt ::= */ yytestcase(yyruleno==140);
      case 142: /* limit_opt ::= */ yytestcase(yyruleno==142);
      case 147: /* where_opt ::= */ yytestcase(yyruleno==147);
      case 215: /* case_else ::= */ yytestcase(yyruleno==215);
      case 217: /* case_operand ::= */ yytestcase(yyruleno==217);
      case 236: /* vinto ::= */ yytestcase(yyruleno==236);
#line 750 "parse.y"
{yymsp[1].minor.yy46 = 0;}
#line 4053 "parse.c"
        break;
      case 124: /* indexed_opt ::= INDEXED BY nm */
#line 764 "parse.y"
{yymsp[-2].minor.yy0 = yymsp[0].minor.yy0;}
#line 4058 "parse.c"
        break;
      case 125: /* indexed_opt ::= NOT INDEXED */
#line 765 "parse.y"
{yymsp[-1].minor.yy0.z=0; yymsp[-1].minor.yy0.n=1;}
#line 4063 "parse.c"
        break;
      case 126: /* using_opt ::= USING LP idlist RP */
#line 769 "parse.y"
{yymsp[-3].minor.yy406 = yymsp[-1].minor.yy406;}
#line 4068 "parse.c"
        break;
      case 127: /* using_opt ::= */
      case 162: /* idlist_opt ::= */ yytestcase(yyruleno==162);
#line 770 "parse.y"
{yymsp[1].minor.yy406 = 0;}
#line 4074 "parse.c"
        break;
      case 129: /* orderby_opt ::= ORDER BY sortlist */
      case 139: /* groupby_opt ::= GROUP BY nexprlist */ yytestcase(yyruleno==139);
#line 784 "parse.y"
{yymsp[-2].minor.yy138 = yymsp[0].minor.yy138;}
#line 4080 "parse.c"
        break;
      case 130: /* sortlist ::= sortlist COMMA expr sortorder nulls */
#line 785 "parse.y"
{
  yymsp[-4].minor.yy138 = sqlite3ExprListAppend(pParse,yymsp[-4].minor.yy138,yymsp[-2].minor.yy46);
  sqlite3ExprListSetSortOrder(yymsp[-4].minor.yy138,yymsp[-1].minor.yy32,yymsp[0].minor.yy32);
}
#line 4088 "parse.c"
        break;
      case 131: /* sortlist ::= expr sortorder nulls */
#line 789 "parse.y"
{
  yymsp[-2].minor.yy138 = sqlite3ExprListAppend(pParse,0,yymsp[-2].minor.yy46); /*A-overwrites-Y*/
  sqlite3ExprListSetSortOrder(yymsp[-2].minor.yy138,yymsp[-1].minor.yy32,yymsp[0].minor.yy32);
}
#line 4096 "parse.c"
        break;
      case 132: /* sortorder ::= ASC */
#line 796 "parse.y"
{yymsp[0].minor.yy32 = SQLITE_SO_ASC;}
#line 4101 "parse.c"
        break;
      case 133: /* sortorder ::= DESC */
#line 797 "parse.y"
{yymsp[0].minor.yy32 = SQLITE_SO_DESC;}
#line 4106 "parse.c"
        break;
      case 134: /* sortorder ::= */
      case 137: /* nulls ::= */ yytestcase(yyruleno==137);
#line 798 "parse.y"
{yymsp[1].minor.yy32 = SQLITE_SO_UNDEFINED;}
#line 4112 "parse.c"
        break;
      case 135: /* nulls ::= NULLS FIRST */
#line 801 "parse.y"
{yymsp[-1].minor.yy32 = SQLITE_SO_ASC;}
#line 4117 "parse.c"
        break;
      case 136: /* nulls ::= NULLS LAST */
#line 802 "parse.y"
{yymsp[-1].minor.yy32 = SQLITE_SO_DESC;}
#line 4122 "parse.c"
        break;
      case 143: /* limit_opt ::= LIMIT expr */
#line 827 "parse.y"
{yymsp[-1].minor.yy46 = sqlite3PExpr(pParse,TK_LIMIT,yymsp[0].minor.yy46,0);}
#line 4127 "parse.c"
        break;
      case 144: /* limit_opt ::= LIMIT expr OFFSET expr */
#line 829 "parse.y"
{yymsp[-3].minor.yy46 = sqlite3PExpr(pParse,TK_LIMIT,yymsp[-2].minor.yy46,yymsp[0].minor.yy46);}
#line 4132 "parse.c"
        break;
      case 145: /* limit_opt ::= LIMIT expr COMMA expr */
#line 831 "parse.y"
{yymsp[-3].minor.yy46 = sqlite3PExpr(pParse,TK_LIMIT,yymsp[0].minor.yy46,yymsp[-2].minor.yy46);}
#line 4137 "parse.c"
        break;
      case 146: /* cmd ::= with DELETE FROM xfullname indexed_opt where_opt */
#line 847 "parse.y"
{
  sqlite3SrcListIndexedBy(pParse, yymsp[-2].minor.yy609, &yymsp[-1].minor.yy0);
  sqlite3DeleteFrom(pParse,yymsp[-2].minor.yy609,yymsp[0].minor.yy46,0,0);
}
#line 4145 "parse.c"
        break;
      case 149: /* cmd ::= with UPDATE orconf xfullname indexed_opt SET setlist where_opt */
#line 871 "parse.y"
{
  sqlite3SrcListIndexedBy(pParse, yymsp[-4].minor.yy609, &yymsp[-3].minor.yy0);
  sqlite3ExprListCheckLength(pParse,yymsp[-1].minor.yy138,"set list"); 
  sqlite3Update(pParse,yymsp[-4].minor.yy609,yymsp[-1].minor.yy138,yymsp[0].minor.yy46,yymsp[-5].minor.yy32,0,0,0);
}
#line 4154 "parse.c"
        break;
      case 150: /* setlist ::= setlist COMMA nm EQ expr */
#line 881 "parse.y"
{
  yymsp[-4].minor.yy138 = sqlite3ExprListAppend(pParse, yymsp[-4].minor.yy138, yymsp[0].minor.yy46);
  sqlite3ExprListSetName(pParse, yymsp[-4].minor.yy138, &yymsp[-2].minor.yy0, 1);
}
#line 4162 "parse.c"
        break;
      case 151: /* setlist ::= setlist COMMA LP idlist RP EQ expr */
#line 885 "parse.y"
{
  yymsp[-6].minor.yy138 = sqlite3ExprListAppendVector(pParse, yymsp[-6].minor.yy138, yymsp[-3].minor.yy406, yymsp[0].minor.yy46);
}
#line 4169 "parse.c"
        break;
      case 152: /* setlist ::= nm EQ expr */
#line 888 "parse.y"
{
  yylhsminor.yy138 = sqlite3ExprListAppend(pParse, 0, yymsp[0].minor.yy46);
  sqlite3ExprListSetName(pParse, yylhsminor.yy138, &yymsp[-2].minor.yy0, 1);
}
#line 4177 "parse.c"
  yymsp[-2].minor.yy138 = yylhsminor.yy138;
        break;
      case 153: /* setlist ::= LP idlist RP EQ expr */
#line 892 "parse.y"
{
  yymsp[-4].minor.yy138 = sqlite3ExprListAppendVector(pParse, 0, yymsp[-3].minor.yy406, yymsp[0].minor.yy46);
}
#line 4185 "parse.c"
        break;
      case 154: /* cmd ::= with insert_cmd INTO xfullname idlist_opt select upsert */
#line 899 "parse.y"
{
  sqlite3Insert(pParse, yymsp[-3].minor.yy609, yymsp[-1].minor.yy25, yymsp[-2].minor.yy406, yymsp[-5].minor.yy32, yymsp[0].minor.yy288);
}
#line 4192 "parse.c"
        break;
      case 155: /* cmd ::= with insert_cmd INTO xfullname idlist_opt DEFAULT VALUES */
#line 903 "parse.y"
{
  sqlite3Insert(pParse, yymsp[-3].minor.yy609, 0, yymsp[-2].minor.yy406, yymsp[-5].minor.yy32, 0);
}
#line 4199 "parse.c"
        break;
      case 156: /* upsert ::= */
#line 914 "parse.y"
{ yymsp[1].minor.yy288 = 0; }
#line 4204 "parse.c"
        break;
      case 157: /* upsert ::= ON CONFLICT LP sortlist RP where_opt DO UPDATE SET setlist where_opt */
#line 917 "parse.y"
{ yymsp[-10].minor.yy288 = sqlite3UpsertNew(pParse->db,yymsp[-7].minor.yy138,yymsp[-5].minor.yy46,yymsp[-1].minor.yy138,yymsp[0].minor.yy46);}
#line 4209 "parse.c"
        break;
      case 158: /* upsert ::= ON CONFLICT LP sortlist RP where_opt DO NOTHING */
#line 919 "parse.y"
{ yymsp[-7].minor.yy288 = sqlite3UpsertNew(pParse->db,yymsp[-4].minor.yy138,yymsp[-2].minor.yy46,0,0); }
#line 4214 "parse.c"
        break;
      case 159: /* upsert ::= ON CONFLICT DO NOTHING */
#line 921 "parse.y"
{ yymsp[-3].minor.yy288 = sqlite3UpsertNew(pParse->db,0,0,0,0); }
#line 4219 "parse.c"
        break;
      case 163: /* idlist_opt ::= LP idlist RP */
#line 933 "parse.y"
{yymsp[-2].minor.yy406 = yymsp[-1].minor.yy406;}
#line 4224 "parse.c"
        break;
      case 164: /* idlist ::= idlist COMMA nm */
#line 935 "parse.y"
{yymsp[-2].minor.yy406 = sqlite3IdListAppend(pParse,yymsp[-2].minor.yy406,&yymsp[0].minor.yy0);}
#line 4229 "parse.c"
        break;
      case 165: /* idlist ::= nm */
#line 937 "parse.y"
{yymsp[0].minor.yy406 = sqlite3IdListAppend(pParse,0,&yymsp[0].minor.yy0); /*A-overwrites-Y*/}
#line 4234 "parse.c"
        break;
      case 166: /* expr ::= LP expr RP */
#line 987 "parse.y"
{yymsp[-2].minor.yy46 = yymsp[-1].minor.yy46;}
#line 4239 "parse.c"
        break;
      case 167: /* expr ::= ID|INDEXED */
      case 168: /* expr ::= JOIN_KW */ yytestcase(yyruleno==168);
#line 988 "parse.y"
{yymsp[0].minor.yy46=tokenExpr(pParse,TK_ID,yymsp[0].minor.yy0); /*A-overwrites-X*/}
#line 4245 "parse.c"
        break;
      case 169: /* expr ::= nm DOT nm */
#line 990 "parse.y"
{
  Expr *temp1 = sqlite3ExprAlloc(pParse->db, TK_ID, &yymsp[-2].minor.yy0, 1);
  Expr *temp2 = sqlite3ExprAlloc(pParse->db, TK_ID, &yymsp[0].minor.yy0, 1);
  if( IN_RENAME_OBJECT ){
    sqlite3RenameTokenMap(pParse, (void*)temp2, &yymsp[0].minor.yy0);
    sqlite3RenameTokenMap(pParse, (void*)temp1, &yymsp[-2].minor.yy0);
  }
  yylhsminor.yy46 = sqlite3PExpr(pParse, TK_DOT, temp1, temp2);
}
#line 4258 "parse.c"
  yymsp[-2].minor.yy46 = yylhsminor.yy46;
        break;
      case 170: /* expr ::= nm DOT nm DOT nm */
#line 999 "parse.y"
{
  Expr *temp1 = sqlite3ExprAlloc(pParse->db, TK_ID, &yymsp[-4].minor.yy0, 1);
  Expr *temp2 = sqlite3ExprAlloc(pParse->db, TK_ID, &yymsp[-2].minor.yy0, 1);
  Expr *temp3 = sqlite3ExprAlloc(pParse->db, TK_ID, &yymsp[0].minor.yy0, 1);
  Expr *temp4 = sqlite3PExpr(pParse, TK_DOT, temp2, temp3);
  if( IN_RENAME_OBJECT ){
    sqlite3RenameTokenMap(pParse, (void*)temp3, &yymsp[0].minor.yy0);
    sqlite3RenameTokenMap(pParse, (void*)temp2, &yymsp[-2].minor.yy0);
  }
  yylhsminor.yy46 = sqlite3PExpr(pParse, TK_DOT, temp1, temp4);
}
#line 4274 "parse.c"
  yymsp[-4].minor.yy46 = yylhsminor.yy46;
        break;
      case 171: /* term ::= NULL|FLOAT|BLOB */
      case 172: /* term ::= STRING */ yytestcase(yyruleno==172);
#line 1010 "parse.y"
{yymsp[0].minor.yy46=tokenExpr(pParse,yymsp[0].major,yymsp[0].minor.yy0); /*A-overwrites-X*/}
#line 4281 "parse.c"
        break;
      case 173: /* term ::= INTEGER */
#line 1012 "parse.y"
{
  yylhsminor.yy46 = sqlite3ExprAlloc(pParse->db, TK_INTEGER, &yymsp[0].minor.yy0, 1);
}
#line 4288 "parse.c"
  yymsp[0].minor.yy46 = yylhsminor.yy46;
        break;
      case 174: /* expr ::= VARIABLE */
#line 1015 "parse.y"
{
  if( !(yymsp[0].minor.yy0.z[0]=='#' && sqlite3Isdigit(yymsp[0].minor.yy0.z[1])) ){
    u32 n = yymsp[0].minor.yy0.n;
    yymsp[0].minor.yy46 = tokenExpr(pParse, TK_VARIABLE, yymsp[0].minor.yy0);
    sqlite3ExprAssignVarNumber(pParse, yymsp[0].minor.yy46, n);
  }else{
    /* When doing a nested parse, one can include terms in an expression
    ** that look like this:   #1 #2 ...  These terms refer to registers
    ** in the virtual machine.  #N is the N-th register. */
    Token t = yymsp[0].minor.yy0; /*A-overwrites-X*/
    assert( t.n>=2 );
    if( pParse->nested==0 ){
      sqlite3ErrorMsg(pParse, "near \"%T\": syntax error", &t);
      yymsp[0].minor.yy46 = 0;
    }else{
      yymsp[0].minor.yy46 = sqlite3PExpr(pParse, TK_REGISTER, 0, 0);
      if( yymsp[0].minor.yy46 ) sqlite3GetInt32(&t.z[1], &yymsp[0].minor.yy46->iTable);
    }
  }
}
#line 4313 "parse.c"
        break;
      case 175: /* expr ::= expr COLLATE ID|STRING */
#line 1035 "parse.y"
{
  yymsp[-2].minor.yy46 = sqlite3ExprAddCollateToken(pParse, yymsp[-2].minor.yy46, &yymsp[0].minor.yy0, 1);
}
#line 4320 "parse.c"
        break;
      case 176: /* expr ::= CAST LP expr AS typetoken RP */
#line 1039 "parse.y"
{
  yymsp[-5].minor.yy46 = sqlite3ExprAlloc(pParse->db, TK_CAST, &yymsp[-1].minor.yy0, 1);
  sqlite3ExprAttachSubtrees(pParse->db, yymsp[-5].minor.yy46, yymsp[-3].minor.yy46, 0);
}
#line 4328 "parse.c"
        break;
      case 177: /* expr ::= ID|INDEXED LP distinct exprlist RP */
#line 1046 "parse.y"
{
  yylhsminor.yy46 = sqlite3ExprFunction(pParse, yymsp[-1].minor.yy138, &yymsp[-4].minor.yy0, yymsp[-2].minor.yy32);
}
#line 4335 "parse.c"
  yymsp[-4].minor.yy46 = yylhsminor.yy46;
        break;
      case 178: /* expr ::= ID|INDEXED LP STAR RP */
#line 1049 "parse.y"
{
  yylhsminor.yy46 = sqlite3ExprFunction(pParse, 0, &yymsp[-3].minor.yy0, 0);
}
#line 4343 "parse.c"
  yymsp[-3].minor.yy46 = yylhsminor.yy46;
        break;
      case 179: /* expr ::= ID|INDEXED LP distinct exprlist RP filter_over */
#line 1054 "parse.y"
{
  yylhsminor.yy46 = sqlite3ExprFunction(pParse, yymsp[-2].minor.yy138, &yymsp[-5].minor.yy0, yymsp[-3].minor.yy32);
  sqlite3WindowAttach(pParse, yylhsminor.yy46, yymsp[0].minor.yy455);
}
#line 4352 "parse.c"
  yymsp[-5].minor.yy46 = yylhsminor.yy46;
        break;
      case 180: /* expr ::= ID|INDEXED LP STAR RP filter_over */
#line 1058 "parse.y"
{
  yylhsminor.yy46 = sqlite3ExprFunction(pParse, 0, &yymsp[-4].minor.yy0, 0);
  sqlite3WindowAttach(pParse, yylhsminor.yy46, yymsp[0].minor.yy455);
}
#line 4361 "parse.c"
  yymsp[-4].minor.yy46 = yylhsminor.yy46;
        break;
      case 181: /* term ::= CTIME_KW */
#line 1064 "parse.y"
{
  yylhsminor.yy46 = sqlite3ExprFunction(pParse, 0, &yymsp[0].minor.yy0, 0);
}
#line 4369 "parse.c"
  yymsp[0].minor.yy46 = yylhsminor.yy46;
        break;
      case 182: /* expr ::= LP nexprlist COMMA expr RP */
#line 1068 "parse.y"
{
  ExprList *pList = sqlite3ExprListAppend(pParse, yymsp[-3].minor.yy138, yymsp[-1].minor.yy46);
  yymsp[-4].minor.yy46 = sqlite3PExpr(pParse, TK_VECTOR, 0, 0);
  if( yymsp[-4].minor.yy46 ){
    yymsp[-4].minor.yy46->x.pList = pList;
  }else{
    sqlite3ExprListDelete(pParse->db, pList);
  }
}
#line 4383 "parse.c"
        break;
      case 183: /* expr ::= expr AND expr */
#line 1078 "parse.y"
{yymsp[-2].minor.yy46=sqlite3ExprAnd(pParse,yymsp[-2].minor.yy46,yymsp[0].minor.yy46);}
#line 4388 "parse.c"
        break;
      case 184: /* expr ::= expr OR expr */
      case 185: /* expr ::= expr LT|GT|GE|LE expr */ yytestcase(yyruleno==185);
      case 186: /* expr ::= expr EQ|NE expr */ yytestcase(yyruleno==186);
      case 187: /* expr ::= expr BITAND|BITOR|LSHIFT|RSHIFT expr */ yytestcase(yyruleno==187);
      case 188: /* expr ::= expr PLUS|MINUS expr */ yytestcase(yyruleno==188);
      case 189: /* expr ::= expr STAR|SLASH|REM expr */ yytestcase(yyruleno==189);
      case 190: /* expr ::= expr CONCAT expr */ yytestcase(yyruleno==190);
#line 1079 "parse.y"
{yymsp[-2].minor.yy46=sqlite3PExpr(pParse,yymsp[-1].major,yymsp[-2].minor.yy46,yymsp[0].minor.yy46);}
#line 4399 "parse.c"
        break;
      case 191: /* likeop ::= NOT LIKE_KW|MATCH */
#line 1092 "parse.y"
{yymsp[-1].minor.yy0=yymsp[0].minor.yy0; yymsp[-1].minor.yy0.n|=0x80000000; /*yymsp[-1].minor.yy0-overwrite-yymsp[0].minor.yy0*/}
#line 4404 "parse.c"
        break;
      case 192: /* expr ::= expr likeop expr */
#line 1093 "parse.y"
{
  ExprList *pList;
  int bNot = yymsp[-1].minor.yy0.n & 0x80000000;
  yymsp[-1].minor.yy0.n &= 0x7fffffff;
  pList = sqlite3ExprListAppend(pParse,0, yymsp[0].minor.yy46);
  pList = sqlite3ExprListAppend(pParse,pList, yymsp[-2].minor.yy46);
  yymsp[-2].minor.yy46 = sqlite3ExprFunction(pParse, pList, &yymsp[-1].minor.yy0, 0);
  if( bNot ) yymsp[-2].minor.yy46 = sqlite3PExpr(pParse, TK_NOT, yymsp[-2].minor.yy46, 0);
  if( yymsp[-2].minor.yy46 ) yymsp[-2].minor.yy46->flags |= EP_InfixFunc;
}
#line 4418 "parse.c"
        break;
      case 193: /* expr ::= expr likeop expr ESCAPE expr */
#line 1103 "parse.y"
{
  ExprList *pList;
  int bNot = yymsp[-3].minor.yy0.n & 0x80000000;
  yymsp[-3].minor.yy0.n &= 0x7fffffff;
  pList = sqlite3ExprListAppend(pParse,0, yymsp[-2].minor.yy46);
  pList = sqlite3ExprListAppend(pParse,pList, yymsp[-4].minor.yy46);
  pList = sqlite3ExprListAppend(pParse,pList, yymsp[0].minor.yy46);
  yymsp[-4].minor.yy46 = sqlite3ExprFunction(pParse, pList, &yymsp[-3].minor.yy0, 0);
  if( bNot ) yymsp[-4].minor.yy46 = sqlite3PExpr(pParse, TK_NOT, yymsp[-4].minor.yy46, 0);
  if( yymsp[-4].minor.yy46 ) yymsp[-4].minor.yy46->flags |= EP_InfixFunc;
}
#line 4433 "parse.c"
        break;
      case 194: /* expr ::= expr ISNULL|NOTNULL */
#line 1115 "parse.y"
{yymsp[-1].minor.yy46 = sqlite3PExpr(pParse,yymsp[0].major,yymsp[-1].minor.yy46,0);}
#line 4438 "parse.c"
        break;
      case 195: /* expr ::= expr NOT NULL */
#line 1116 "parse.y"
{yymsp[-2].minor.yy46 = sqlite3PExpr(pParse,TK_NOTNULL,yymsp[-2].minor.yy46,0);}
#line 4443 "parse.c"
        break;
      case 196: /* expr ::= expr IS expr */
#line 1137 "parse.y"
{
  yymsp[-2].minor.yy46 = sqlite3PExpr(pParse,TK_IS,yymsp[-2].minor.yy46,yymsp[0].minor.yy46);
  binaryToUnaryIfNull(pParse, yymsp[0].minor.yy46, yymsp[-2].minor.yy46, TK_ISNULL);
}
#line 4451 "parse.c"
        break;
      case 197: /* expr ::= expr IS NOT expr */
#line 1141 "parse.y"
{
  yymsp[-3].minor.yy46 = sqlite3PExpr(pParse,TK_ISNOT,yymsp[-3].minor.yy46,yymsp[0].minor.yy46);
  binaryToUnaryIfNull(pParse, yymsp[0].minor.yy46, yymsp[-3].minor.yy46, TK_NOTNULL);
}
#line 4459 "parse.c"
        break;
      case 198: /* expr ::= NOT expr */
      case 199: /* expr ::= BITNOT expr */ yytestcase(yyruleno==199);
#line 1147 "parse.y"
{yymsp[-1].minor.yy46 = sqlite3PExpr(pParse, yymsp[-1].major, yymsp[0].minor.yy46, 0);/*A-overwrites-B*/}
#line 4465 "parse.c"
        break;
      case 200: /* expr ::= PLUS|MINUS expr */
#line 1150 "parse.y"
{
  yymsp[-1].minor.yy46 = sqlite3PExpr(pParse, yymsp[-1].major==TK_PLUS ? TK_UPLUS : TK_UMINUS, yymsp[0].minor.yy46, 0);
  /*A-overwrites-B*/
}
#line 4473 "parse.c"
        break;
      case 201: /* between_op ::= BETWEEN */
      case 204: /* in_op ::= IN */ yytestcase(yyruleno==204);
#line 1156 "parse.y"
{yymsp[0].minor.yy32 = 0;}
#line 4479 "parse.c"
        break;
      case 203: /* expr ::= expr between_op expr AND expr */
#line 1158 "parse.y"
{
  ExprList *pList = sqlite3ExprListAppend(pParse,0, yymsp[-2].minor.yy46);
  pList = sqlite3ExprListAppend(pParse,pList, yymsp[0].minor.yy46);
  yymsp[-4].minor.yy46 = sqlite3PExpr(pParse, TK_BETWEEN, yymsp[-4].minor.yy46, 0);
  if( yymsp[-4].minor.yy46 ){
    yymsp[-4].minor.yy46->x.pList = pList;
  }else{
    sqlite3ExprListDelete(pParse->db, pList);
  } 
  if( yymsp[-3].minor.yy32 ) yymsp[-4].minor.yy46 = sqlite3PExpr(pParse, TK_NOT, yymsp[-4].minor.yy46, 0);
}
#line 4494 "parse.c"
        break;
      case 206: /* expr ::= expr in_op LP exprlist RP */
#line 1173 "parse.y"
{
    if( yymsp[-1].minor.yy138==0 ){
      /* Expressions of the form
      **
      **      expr1 IN ()
      **      expr1 NOT IN ()
      **
      ** simplify to constants 0 (false) and 1 (true), respectively,
      ** regardless of the value of expr1.
      */
      sqlite3ExprUnmapAndDelete(pParse, yymsp[-4].minor.yy46);
      yymsp[-4].minor.yy46 = sqlite3Expr(pParse->db, TK_INTEGER, yymsp[-3].minor.yy32 ? "1" : "0");
    }else{
      yymsp[-4].minor.yy46 = sqlite3PExpr(pParse, TK_IN, yymsp[-4].minor.yy46, 0);
      if( yymsp[-4].minor.yy46 ){
        yymsp[-4].minor.yy46->x.pList = yymsp[-1].minor.yy138;
        sqlite3ExprSetHeightAndFlags(pParse, yymsp[-4].minor.yy46);
      }else{
        sqlite3ExprListDelete(pParse->db, yymsp[-1].minor.yy138);
      }
      if( yymsp[-3].minor.yy32 ) yymsp[-4].minor.yy46 = sqlite3PExpr(pParse, TK_NOT, yymsp[-4].minor.yy46, 0);
    }
  }
#line 4521 "parse.c"
        break;
      case 207: /* expr ::= LP select RP */
#line 1196 "parse.y"
{
    yymsp[-2].minor.yy46 = sqlite3PExpr(pParse, TK_SELECT, 0, 0);
    sqlite3PExprAddSelect(pParse, yymsp[-2].minor.yy46, yymsp[-1].minor.yy25);
  }
#line 4529 "parse.c"
        break;
      case 208: /* expr ::= expr in_op LP select RP */
#line 1200 "parse.y"
{
    yymsp[-4].minor.yy46 = sqlite3PExpr(pParse, TK_IN, yymsp[-4].minor.yy46, 0);
    sqlite3PExprAddSelect(pParse, yymsp[-4].minor.yy46, yymsp[-1].minor.yy25);
    if( yymsp[-3].minor.yy32 ) yymsp[-4].minor.yy46 = sqlite3PExpr(pParse, TK_NOT, yymsp[-4].minor.yy46, 0);
  }
#line 4538 "parse.c"
        break;
      case 209: /* expr ::= expr in_op nm dbnm paren_exprlist */
#line 1205 "parse.y"
{
    SrcList *pSrc = sqlite3SrcListAppend(pParse, 0,&yymsp[-2].minor.yy0,&yymsp[-1].minor.yy0);
    Select *pSelect = sqlite3SelectNew(pParse, 0,pSrc,0,0,0,0,0,0);
    if( yymsp[0].minor.yy138 )  sqlite3SrcListFuncArgs(pParse, pSelect ? pSrc : 0, yymsp[0].minor.yy138);
    yymsp[-4].minor.yy46 = sqlite3PExpr(pParse, TK_IN, yymsp[-4].minor.yy46, 0);
    sqlite3PExprAddSelect(pParse, yymsp[-4].minor.yy46, pSelect);
    if( yymsp[-3].minor.yy32 ) yymsp[-4].minor.yy46 = sqlite3PExpr(pParse, TK_NOT, yymsp[-4].minor.yy46, 0);
  }
#line 4550 "parse.c"
        break;
      case 210: /* expr ::= EXISTS LP select RP */
#line 1213 "parse.y"
{
    Expr *p;
    p = yymsp[-3].minor.yy46 = sqlite3PExpr(pParse, TK_EXISTS, 0, 0);
    sqlite3PExprAddSelect(pParse, p, yymsp[-1].minor.yy25);
  }
#line 4559 "parse.c"
        break;
      case 211: /* expr ::= CASE case_operand case_exprlist case_else END */
#line 1221 "parse.y"
{
  yymsp[-4].minor.yy46 = sqlite3PExpr(pParse, TK_CASE, yymsp[-3].minor.yy46, 0);
  if( yymsp[-4].minor.yy46 ){
    yymsp[-4].minor.yy46->x.pList = yymsp[-1].minor.yy46 ? sqlite3ExprListAppend(pParse,yymsp[-2].minor.yy138,yymsp[-1].minor.yy46) : yymsp[-2].minor.yy138;
    sqlite3ExprSetHeightAndFlags(pParse, yymsp[-4].minor.yy46);
  }else{
    sqlite3ExprListDelete(pParse->db, yymsp[-2].minor.yy138);
    sqlite3ExprDelete(pParse->db, yymsp[-1].minor.yy46);
  }
}
#line 4573 "parse.c"
        break;
      case 212: /* case_exprlist ::= case_exprlist WHEN expr THEN expr */
#line 1233 "parse.y"
{
  yymsp[-4].minor.yy138 = sqlite3ExprListAppend(pParse,yymsp[-4].minor.yy138, yymsp[-2].minor.yy46);
  yymsp[-4].minor.yy138 = sqlite3ExprListAppend(pParse,yymsp[-4].minor.yy138, yymsp[0].minor.yy46);
}
#line 4581 "parse.c"
        break;
      case 213: /* case_exprlist ::= WHEN expr THEN expr */
#line 1237 "parse.y"
{
  yymsp[-3].minor.yy138 = sqlite3ExprListAppend(pParse,0, yymsp[-2].minor.yy46);
  yymsp[-3].minor.yy138 = sqlite3ExprListAppend(pParse,yymsp[-3].minor.yy138, yymsp[0].minor.yy46);
}
#line 4589 "parse.c"
        break;
      case 216: /* case_operand ::= expr */
#line 1247 "parse.y"
{yymsp[0].minor.yy46 = yymsp[0].minor.yy46; /*A-overwrites-X*/}
#line 4594 "parse.c"
        break;
      case 219: /* nexprlist ::= nexprlist COMMA expr */
#line 1258 "parse.y"
{yymsp[-2].minor.yy138 = sqlite3ExprListAppend(pParse,yymsp[-2].minor.yy138,yymsp[0].minor.yy46);}
#line 4599 "parse.c"
        break;
      case 220: /* nexprlist ::= expr */
#line 1260 "parse.y"
{yymsp[0].minor.yy138 = sqlite3ExprListAppend(pParse,0,yymsp[0].minor.yy46); /*A-overwrites-Y*/}
#line 4604 "parse.c"
        break;
      case 222: /* paren_exprlist ::= LP exprlist RP */
      case 227: /* eidlist_opt ::= LP eidlist RP */ yytestcase(yyruleno==227);
#line 1268 "parse.y"
{yymsp[-2].minor.yy138 = yymsp[-1].minor.yy138;}
#line 4610 "parse.c"
        break;
      case 223: /* cmd ::= createkw uniqueflag INDEX ifnotexists nm dbnm ON nm LP sortlist RP where_opt */
#line 1275 "parse.y"
{
  sqlite3CreateIndex(pParse, &yymsp[-7].minor.yy0, &yymsp[-6].minor.yy0, 
                     sqlite3SrcListAppend(pParse,0,&yymsp[-4].minor.yy0,0), yymsp[-2].minor.yy138, yymsp[-10].minor.yy32,
                      &yymsp[-11].minor.yy0, yymsp[0].minor.yy46, SQLITE_SO_ASC, yymsp[-8].minor.yy32, SQLITE_IDXTYPE_APPDEF);
  if( IN_RENAME_OBJECT && pParse->pNewIndex ){
    sqlite3RenameTokenMap(pParse, pParse->pNewIndex->zName, &yymsp[-4].minor.yy0);
  }
}
#line 4622 "parse.c"
        break;
      case 224: /* uniqueflag ::= UNIQUE */
      case 266: /* raisetype ::= ABORT */ yytestcase(yyruleno==266);
#line 1285 "parse.y"
{yymsp[0].minor.yy32 = OE_Abort;}
#line 4628 "parse.c"
        break;
      case 225: /* uniqueflag ::= */
#line 1286 "parse.y"
{yymsp[1].minor.yy32 = OE_None;}
#line 4633 "parse.c"
        break;
      case 228: /* eidlist ::= eidlist COMMA nm collate sortorder */
#line 1336 "parse.y"
{
  yymsp[-4].minor.yy138 = parserAddExprIdListTerm(pParse, yymsp[-4].minor.yy138, &yymsp[-2].minor.yy0, yymsp[-1].minor.yy32, yymsp[0].minor.yy32);
}
#line 4640 "parse.c"
        break;
      case 229: /* eidlist ::= nm collate sortorder */
#line 1339 "parse.y"
{
  yymsp[-2].minor.yy138 = parserAddExprIdListTerm(pParse, 0, &yymsp[-2].minor.yy0, yymsp[-1].minor.yy32, yymsp[0].minor.yy32); /*A-overwrites-Y*/
}
#line 4647 "parse.c"
        break;
      case 232: /* cmd ::= DROP INDEX ifexists fullname */
#line 1350 "parse.y"
{sqlite3DropIndex(pParse, yymsp[0].minor.yy609, yymsp[-1].minor.yy32);}
#line 4652 "parse.c"
        break;
      case 233: /* cmd ::= VACUUM vinto */
#line 1358 "parse.y"
{sqlite3Vacuum(pParse,0,yymsp[0].minor.yy46);}
#line 4657 "parse.c"
        break;
      case 234: /* cmd ::= VACUUM nm vinto */
#line 1359 "parse.y"
{sqlite3Vacuum(pParse,&yymsp[-1].minor.yy0,yymsp[0].minor.yy46);}
#line 4662 "parse.c"
        break;
      case 237: /* cmd ::= PRAGMA nm dbnm */
#line 1368 "parse.y"
{sqlite3Pragma(pParse,&yymsp[-1].minor.yy0,&yymsp[0].minor.yy0,0,0);}
#line 4667 "parse.c"
        break;
      case 238: /* cmd ::= PRAGMA nm dbnm EQ nmnum */
#line 1369 "parse.y"
{sqlite3Pragma(pParse,&yymsp[-3].minor.yy0,&yymsp[-2].minor.yy0,&yymsp[0].minor.yy0,0);}
#line 4672 "parse.c"
        break;
      case 239: /* cmd ::= PRAGMA nm dbnm LP nmnum RP */
#line 1370 "parse.y"
{sqlite3Pragma(pParse,&yymsp[-4].minor.yy0,&yymsp[-3].minor.yy0,&yymsp[-1].minor.yy0,0);}
#line 4677 "parse.c"
        break;
      case 240: /* cmd ::= PRAGMA nm dbnm EQ minus_num */
#line 1372 "parse.y"
{sqlite3Pragma(pParse,&yymsp[-3].minor.yy0,&yymsp[-2].minor.yy0,&yymsp[0].minor.yy0,1);}
#line 4682 "parse.c"
        break;
      case 241: /* cmd ::= PRAGMA nm dbnm LP minus_num RP */
#line 1374 "parse.y"
{sqlite3Pragma(pParse,&yymsp[-4].minor.yy0,&yymsp[-3].minor.yy0,&yymsp[-1].minor.yy0,1);}
#line 4687 "parse.c"
        break;
      case 244: /* cmd ::= createkw trigger_decl BEGIN trigger_cmd_list END */
#line 1390 "parse.y"
{
  Token all;
  all.z = yymsp[-3].minor.yy0.z;
  all.n = (int)(yymsp[0].minor.yy0.z - yymsp[-3].minor.yy0.z) + yymsp[0].minor.yy0.n;
  sqlite3FinishTrigger(pParse, yymsp[-1].minor.yy527, &all);
}
#line 4697 "parse.c"
        break;
      case 245: /* trigger_decl ::= temp TRIGGER ifnotexists nm dbnm trigger_time trigger_event ON fullname foreach_clause when_clause */
#line 1399 "parse.y"
{
  sqlite3BeginTrigger(pParse, &yymsp[-7].minor.yy0, &yymsp[-6].minor.yy0, yymsp[-5].minor.yy32, yymsp[-4].minor.yy572.a, yymsp[-4].minor.yy572.b, yymsp[-2].minor.yy609, yymsp[0].minor.yy46, yymsp[-10].minor.yy32, yymsp[-8].minor.yy32);
  yymsp[-10].minor.yy0 = (yymsp[-6].minor.yy0.n==0?yymsp[-7].minor.yy0:yymsp[-6].minor.yy0); /*A-overwrites-T*/
}
#line 4705 "parse.c"
        break;
      case 246: /* trigger_time ::= BEFORE|AFTER */
#line 1405 "parse.y"
{ yymsp[0].minor.yy32 = yymsp[0].major; /*A-overwrites-X*/ }
#line 4710 "parse.c"
        break;
      case 247: /* trigger_time ::= INSTEAD OF */
#line 1406 "parse.y"
{ yymsp[-1].minor.yy32 = TK_INSTEAD;}
#line 4715 "parse.c"
        break;
      case 248: /* trigger_time ::= */
#line 1407 "parse.y"
{ yymsp[1].minor.yy32 = TK_BEFORE; }
#line 4720 "parse.c"
        break;
      case 249: /* trigger_event ::= DELETE|INSERT */
      case 250: /* trigger_event ::= UPDATE */ yytestcase(yyruleno==250);
#line 1411 "parse.y"
{yymsp[0].minor.yy572.a = yymsp[0].major; /*A-overwrites-X*/ yymsp[0].minor.yy572.b = 0;}
#line 4726 "parse.c"
        break;
      case 251: /* trigger_event ::= UPDATE OF idlist */
#line 1413 "parse.y"
{yymsp[-2].minor.yy572.a = TK_UPDATE; yymsp[-2].minor.yy572.b = yymsp[0].minor.yy406;}
#line 4731 "parse.c"
        break;
      case 252: /* when_clause ::= */
      case 271: /* key_opt ::= */ yytestcase(yyruleno==271);
#line 1420 "parse.y"
{ yymsp[1].minor.yy46 = 0; }
#line 4737 "parse.c"
        break;
      case 253: /* when_clause ::= WHEN expr */
      case 272: /* key_opt ::= KEY expr */ yytestcase(yyruleno==272);
#line 1421 "parse.y"
{ yymsp[-1].minor.yy46 = yymsp[0].minor.yy46; }
#line 4743 "parse.c"
        break;
      case 254: /* trigger_cmd_list ::= trigger_cmd_list trigger_cmd SEMI */
#line 1425 "parse.y"
{
  assert( yymsp[-2].minor.yy527!=0 );
  yymsp[-2].minor.yy527->pLast->pNext = yymsp[-1].minor.yy527;
  yymsp[-2].minor.yy527->pLast = yymsp[-1].minor.yy527;
}
#line 4752 "parse.c"
        break;
      case 255: /* trigger_cmd_list ::= trigger_cmd SEMI */
#line 1430 "parse.y"
{ 
  assert( yymsp[-1].minor.yy527!=0 );
  yymsp[-1].minor.yy527->pLast = yymsp[-1].minor.yy527;
}
#line 4760 "parse.c"
        break;
      case 256: /* trnm ::= nm DOT nm */
#line 1441 "parse.y"
{
  yymsp[-2].minor.yy0 = yymsp[0].minor.yy0;
  sqlite3ErrorMsg(pParse, 
        "qualified table names are not allowed on INSERT, UPDATE, and DELETE "
        "statements within triggers");
}
#line 4770 "parse.c"
        break;
      case 257: /* tridxby ::= INDEXED BY nm */
#line 1453 "parse.y"
{
  sqlite3ErrorMsg(pParse,
        "the INDEXED BY clause is not allowed on UPDATE or DELETE statements "
        "within triggers");
}
#line 4779 "parse.c"
        break;
      case 258: /* tridxby ::= NOT INDEXED */
#line 1458 "parse.y"
{
  sqlite3ErrorMsg(pParse,
        "the NOT INDEXED clause is not allowed on UPDATE or DELETE statements "
        "within triggers");
}
#line 4788 "parse.c"
        break;
      case 259: /* trigger_cmd ::= UPDATE orconf trnm tridxby SET setlist where_opt scanpt */
#line 1471 "parse.y"
{yylhsminor.yy527 = sqlite3TriggerUpdateStep(pParse, &yymsp[-5].minor.yy0, yymsp[-2].minor.yy138, yymsp[-1].minor.yy46, yymsp[-6].minor.yy32, yymsp[-7].minor.yy0.z, yymsp[0].minor.yy8);}
#line 4793 "parse.c"
  yymsp[-7].minor.yy527 = yylhsminor.yy527;
        break;
      case 260: /* trigger_cmd ::= scanpt insert_cmd INTO trnm idlist_opt select upsert scanpt */
#line 1475 "parse.y"
{
   yylhsminor.yy527 = sqlite3TriggerInsertStep(pParse,&yymsp[-4].minor.yy0,yymsp[-3].minor.yy406,yymsp[-2].minor.yy25,yymsp[-6].minor.yy32,yymsp[-1].minor.yy288,yymsp[-7].minor.yy8,yymsp[0].minor.yy8);/*yylhsminor.yy527-overwrites-yymsp[-6].minor.yy32*/
}
#line 4801 "parse.c"
  yymsp[-7].minor.yy527 = yylhsminor.yy527;
        break;
      case 261: /* trigger_cmd ::= DELETE FROM trnm tridxby where_opt scanpt */
#line 1480 "parse.y"
{yylhsminor.yy527 = sqlite3TriggerDeleteStep(pParse, &yymsp[-3].minor.yy0, yymsp[-1].minor.yy46, yymsp[-5].minor.yy0.z, yymsp[0].minor.yy8);}
#line 4807 "parse.c"
  yymsp[-5].minor.yy527 = yylhsminor.yy527;
        break;
      case 262: /* trigger_cmd ::= scanpt select scanpt */
#line 1484 "parse.y"
{yylhsminor.yy527 = sqlite3TriggerSelectStep(pParse->db, yymsp[-1].minor.yy25, yymsp[-2].minor.yy8, yymsp[0].minor.yy8); /*yylhsminor.yy527-overwrites-yymsp[-1].minor.yy25*/}
#line 4813 "parse.c"
  yymsp[-2].minor.yy527 = yylhsminor.yy527;
        break;
      case 263: /* expr ::= RAISE LP IGNORE RP */
#line 1487 "parse.y"
{
  yymsp[-3].minor.yy46 = sqlite3PExpr(pParse, TK_RAISE, 0, 0); 
  if( yymsp[-3].minor.yy46 ){
    yymsp[-3].minor.yy46->affExpr = OE_Ignore;
  }
}
#line 4824 "parse.c"
        break;
      case 264: /* expr ::= RAISE LP raisetype COMMA nm RP */
#line 1493 "parse.y"
{
  yymsp[-5].minor.yy46 = sqlite3ExprAlloc(pParse->db, TK_RAISE, &yymsp[-1].minor.yy0, 1); 
  if( yymsp[-5].minor.yy46 ) {
    yymsp[-5].minor.yy46->affExpr = (char)yymsp[-3].minor.yy32;
  }
}
#line 4834 "parse.c"
        break;
      case 265: /* raisetype ::= ROLLBACK */
#line 1502 "parse.y"
{yymsp[0].minor.yy32 = OE_Rollback;}
#line 4839 "parse.c"
        break;
      case 267: /* raisetype ::= FAIL */
#line 1504 "parse.y"
{yymsp[0].minor.yy32 = OE_Fail;}
#line 4844 "parse.c"
        break;
      case 268: /* cmd ::= DROP TRIGGER ifexists fullname */
#line 1509 "parse.y"
{
  sqlite3DropTrigger(pParse,yymsp[0].minor.yy609,yymsp[-1].minor.yy32);
}
#line 4851 "parse.c"
        break;
      case 269: /* cmd ::= ATTACH database_kw_opt expr AS expr key_opt */
#line 1516 "parse.y"
{
  sqlite3Attach(pParse, yymsp[-3].minor.yy46, yymsp[-1].minor.yy46, yymsp[0].minor.yy46);
}
#line 4858 "parse.c"
        break;
      case 270: /* cmd ::= DETACH database_kw_opt expr */
#line 1519 "parse.y"
{
  sqlite3Detach(pParse, yymsp[0].minor.yy46);
}
#line 4865 "parse.c"
        break;
      case 273: /* cmd ::= REINDEX */
#line 1534 "parse.y"
{sqlite3Reindex(pParse, 0, 0);}
#line 4870 "parse.c"
        break;
      case 274: /* cmd ::= REINDEX nm dbnm */
#line 1535 "parse.y"
{sqlite3Reindex(pParse, &yymsp[-1].minor.yy0, &yymsp[0].minor.yy0);}
#line 4875 "parse.c"
        break;
      case 275: /* cmd ::= ANALYZE */
#line 1540 "parse.y"
{sqlite3Analyze(pParse, 0, 0);}
#line 4880 "parse.c"
        break;
      case 276: /* cmd ::= ANALYZE nm dbnm */
#line 1541 "parse.y"
{sqlite3Analyze(pParse, &yymsp[-1].minor.yy0, &yymsp[0].minor.yy0);}
#line 4885 "parse.c"
        break;
      case 277: /* cmd ::= ALTER TABLE fullname RENAME TO nm */
#line 1546 "parse.y"
{
  sqlite3AlterRenameTable(pParse,yymsp[-3].minor.yy609,&yymsp[0].minor.yy0);
}
#line 4892 "parse.c"
        break;
      case 278: /* cmd ::= ALTER TABLE add_column_fullname ADD kwcolumn_opt columnname carglist */
#line 1550 "parse.y"
{
  yymsp[-1].minor.yy0.n = (int)(pParse->sLastToken.z-yymsp[-1].minor.yy0.z) + pParse->sLastToken.n;
  sqlite3AlterFinishAddColumn(pParse, &yymsp[-1].minor.yy0);
}
#line 4900 "parse.c"
        break;
      case 279: /* add_column_fullname ::= fullname */
#line 1554 "parse.y"
{
  disableLookaside(pParse);
  sqlite3AlterBeginAddColumn(pParse, yymsp[0].minor.yy609);
}
#line 4908 "parse.c"
        break;
      case 280: /* cmd ::= ALTER TABLE fullname RENAME kwcolumn_opt nm TO nm */
#line 1558 "parse.y"
{
  sqlite3AlterRenameColumn(pParse, yymsp[-5].minor.yy609, &yymsp[-2].minor.yy0, &yymsp[0].minor.yy0);
}
#line 4915 "parse.c"
        break;
      case 281: /* cmd ::= create_vtab */
#line 1569 "parse.y"
{sqlite3VtabFinishParse(pParse,0);}
#line 4920 "parse.c"
        break;
      case 282: /* cmd ::= create_vtab LP vtabarglist RP */
#line 1570 "parse.y"
{sqlite3VtabFinishParse(pParse,&yymsp[0].minor.yy0);}
#line 4925 "parse.c"
        break;
      case 283: /* create_vtab ::= createkw VIRTUAL TABLE ifnotexists nm dbnm USING nm */
#line 1572 "parse.y"
{
    sqlite3VtabBeginParse(pParse, &yymsp[-3].minor.yy0, &yymsp[-2].minor.yy0, &yymsp[0].minor.yy0, yymsp[-4].minor.yy32);
}
#line 4932 "parse.c"
        break;
      case 284: /* vtabarg ::= */
#line 1577 "parse.y"
{sqlite3VtabArgInit(pParse);}
#line 4937 "parse.c"
        break;
      case 285: /* vtabargtoken ::= ANY */
      case 286: /* vtabargtoken ::= lp anylist RP */ yytestcase(yyruleno==286);
      case 287: /* lp ::= LP */ yytestcase(yyruleno==287);
#line 1579 "parse.y"
{sqlite3VtabArgExtend(pParse,&yymsp[0].minor.yy0);}
#line 4944 "parse.c"
        break;
      case 288: /* with ::= WITH wqlist */
      case 289: /* with ::= WITH RECURSIVE wqlist */ yytestcase(yyruleno==289);
#line 1594 "parse.y"
{ sqlite3WithPush(pParse, yymsp[0].minor.yy297, 1); }
#line 4950 "parse.c"
        break;
      case 290: /* wqlist ::= nm eidlist_opt AS LP select RP */
#line 1597 "parse.y"
{
  yymsp[-5].minor.yy297 = sqlite3WithAdd(pParse, 0, &yymsp[-5].minor.yy0, yymsp[-4].minor.yy138, yymsp[-1].minor.yy25); /*A-overwrites-X*/
}
#line 4957 "parse.c"
        break;
      case 291: /* wqlist ::= wqlist COMMA nm eidlist_opt AS LP select RP */
#line 1600 "parse.y"
{
  yymsp[-7].minor.yy297 = sqlite3WithAdd(pParse, yymsp[-7].minor.yy297, &yymsp[-5].minor.yy0, yymsp[-4].minor.yy138, yymsp[-1].minor.yy25);
}
#line 4964 "parse.c"
        break;
      case 292: /* windowdefn_list ::= windowdefn */
#line 1614 "parse.y"
{ yylhsminor.yy455 = yymsp[0].minor.yy455; }
#line 4969 "parse.c"
  yymsp[0].minor.yy455 = yylhsminor.yy455;
        break;
      case 293: /* windowdefn_list ::= windowdefn_list COMMA windowdefn */
#line 1615 "parse.y"
{
  assert( yymsp[0].minor.yy455!=0 );
  sqlite3WindowChain(pParse, yymsp[0].minor.yy455, yymsp[-2].minor.yy455);
  yymsp[0].minor.yy455->pNextWin = yymsp[-2].minor.yy455;
  yylhsminor.yy455 = yymsp[0].minor.yy455;
}
#line 4980 "parse.c"
  yymsp[-2].minor.yy455 = yylhsminor.yy455;
        break;
      case 294: /* windowdefn ::= nm AS LP window RP */
#line 1624 "parse.y"
{
  if( ALWAYS(yymsp[-1].minor.yy455) ){
    yymsp[-1].minor.yy455->zName = sqlite3DbStrNDup(pParse->db, yymsp[-4].minor.yy0.z, yymsp[-4].minor.yy0.n);
  }
  yylhsminor.yy455 = yymsp[-1].minor.yy455;
}
#line 4991 "parse.c"
  yymsp[-4].minor.yy455 = yylhsminor.yy455;
        break;
      case 295: /* window ::= PARTITION BY nexprlist orderby_opt frame_opt */
#line 1658 "parse.y"
{
  yymsp[-4].minor.yy455 = sqlite3WindowAssemble(pParse, yymsp[0].minor.yy455, yymsp[-2].minor.yy138, yymsp[-1].minor.yy138, 0);
}
#line 4999 "parse.c"
        break;
      case 296: /* window ::= nm PARTITION BY nexprlist orderby_opt frame_opt */
#line 1661 "parse.y"
{
  yylhsminor.yy455 = sqlite3WindowAssemble(pParse, yymsp[0].minor.yy455, yymsp[-2].minor.yy138, yymsp[-1].minor.yy138, &yymsp[-5].minor.yy0);
}
#line 5006 "parse.c"
  yymsp[-5].minor.yy455 = yylhsminor.yy455;
        break;
      case 297: /* window ::= ORDER BY sortlist frame_opt */
#line 1664 "parse.y"
{
  yymsp[-3].minor.yy455 = sqlite3WindowAssemble(pParse, yymsp[0].minor.yy455, 0, yymsp[-1].minor.yy138, 0);
}
#line 5014 "parse.c"
        break;
      case 298: /* window ::= nm ORDER BY sortlist frame_opt */
#line 1667 "parse.y"
{
  yylhsminor.yy455 = sqlite3WindowAssemble(pParse, yymsp[0].minor.yy455, 0, yymsp[-1].minor.yy138, &yymsp[-4].minor.yy0);
}
#line 5021 "parse.c"
  yymsp[-4].minor.yy455 = yylhsminor.yy455;
        break;
      case 299: /* window ::= frame_opt */
      case 318: /* filter_over ::= over_clause */ yytestcase(yyruleno==318);
#line 1670 "parse.y"
{
  yylhsminor.yy455 = yymsp[0].minor.yy455;
}
#line 5030 "parse.c"
  yymsp[0].minor.yy455 = yylhsminor.yy455;
        break;
      case 300: /* window ::= nm frame_opt */
#line 1673 "parse.y"
{
  yylhsminor.yy455 = sqlite3WindowAssemble(pParse, yymsp[0].minor.yy455, 0, 0, &yymsp[-1].minor.yy0);
}
#line 5038 "parse.c"
  yymsp[-1].minor.yy455 = yylhsminor.yy455;
        break;
      case 301: /* frame_opt ::= */
#line 1677 "parse.y"
{ 
  yymsp[1].minor.yy455 = sqlite3WindowAlloc(pParse, 0, TK_UNBOUNDED, 0, TK_CURRENT, 0, 0);
}
#line 5046 "parse.c"
        break;
      case 302: /* frame_opt ::= range_or_rows frame_bound_s frame_exclude_opt */
#line 1680 "parse.y"
{ 
  yylhsminor.yy455 = sqlite3WindowAlloc(pParse, yymsp[-2].minor.yy32, yymsp[-1].minor.yy57.eType, yymsp[-1].minor.yy57.pExpr, TK_CURRENT, 0, yymsp[0].minor.yy118);
}
#line 5053 "parse.c"
  yymsp[-2].minor.yy455 = yylhsminor.yy455;
        break;
      case 303: /* frame_opt ::= range_or_rows BETWEEN frame_bound_s AND frame_bound_e frame_exclude_opt */
#line 1684 "parse.y"
{ 
  yylhsminor.yy455 = sqlite3WindowAlloc(pParse, yymsp[-5].minor.yy32, yymsp[-3].minor.yy57.eType, yymsp[-3].minor.yy57.pExpr, yymsp[-1].minor.yy57.eType, yymsp[-1].minor.yy57.pExpr, yymsp[0].minor.yy118);
}
#line 5061 "parse.c"
  yymsp[-5].minor.yy455 = yylhsminor.yy455;
        break;
      case 305: /* frame_bound_s ::= frame_bound */
      case 307: /* frame_bound_e ::= frame_bound */ yytestcase(yyruleno==307);
#line 1690 "parse.y"
{yylhsminor.yy57 = yymsp[0].minor.yy57;}
#line 5068 "parse.c"
  yymsp[0].minor.yy57 = yylhsminor.yy57;
        break;
      case 306: /* frame_bound_s ::= UNBOUNDED PRECEDING */
      case 308: /* frame_bound_e ::= UNBOUNDED FOLLOWING */ yytestcase(yyruleno==308);
      case 310: /* frame_bound ::= CURRENT ROW */ yytestcase(yyruleno==310);
#line 1691 "parse.y"
{yylhsminor.yy57.eType = yymsp[-1].major; yylhsminor.yy57.pExpr = 0;}
#line 5076 "parse.c"
  yymsp[-1].minor.yy57 = yylhsminor.yy57;
        break;
      case 309: /* frame_bound ::= expr PRECEDING|FOLLOWING */
#line 1696 "parse.y"
{yylhsminor.yy57.eType = yymsp[0].major; yylhsminor.yy57.pExpr = yymsp[-1].minor.yy46;}
#line 5082 "parse.c"
  yymsp[-1].minor.yy57 = yylhsminor.yy57;
        break;
      case 311: /* frame_exclude_opt ::= */
#line 1700 "parse.y"
{yymsp[1].minor.yy118 = 0;}
#line 5088 "parse.c"
        break;
      case 312: /* frame_exclude_opt ::= EXCLUDE frame_exclude */
#line 1701 "parse.y"
{yymsp[-1].minor.yy118 = yymsp[0].minor.yy118;}
#line 5093 "parse.c"
        break;
      case 313: /* frame_exclude ::= NO OTHERS */
      case 314: /* frame_exclude ::= CURRENT ROW */ yytestcase(yyruleno==314);
#line 1704 "parse.y"
{yymsp[-1].minor.yy118 = yymsp[-1].major; /*A-overwrites-X*/}
#line 5099 "parse.c"
        break;
      case 315: /* frame_exclude ::= GROUP|TIES */
#line 1706 "parse.y"
{yymsp[0].minor.yy118 = yymsp[0].major; /*A-overwrites-X*/}
#line 5104 "parse.c"
        break;
      case 316: /* window_clause ::= WINDOW windowdefn_list */
#line 1711 "parse.y"
{ yymsp[-1].minor.yy455 = yymsp[0].minor.yy455; }
#line 5109 "parse.c"
        break;
      case 317: /* filter_over ::= filter_clause over_clause */
#line 1713 "parse.y"
{
  yymsp[0].minor.yy455->pFilter = yymsp[-1].minor.yy46;
  yylhsminor.yy455 = yymsp[0].minor.yy455;
}
#line 5117 "parse.c"
  yymsp[-1].minor.yy455 = yylhsminor.yy455;
        break;
      case 319: /* filter_over ::= filter_clause */
#line 1720 "parse.y"
{
  yylhsminor.yy455 = (Window*)sqlite3DbMallocZero(pParse->db, sizeof(Window));
  if( yylhsminor.yy455 ){
    yylhsminor.yy455->eFrmType = TK_FILTER;
    yylhsminor.yy455->pFilter = yymsp[0].minor.yy46;
  }else{
    sqlite3ExprDelete(pParse->db, yymsp[0].minor.yy46);
  }
}
#line 5131 "parse.c"
  yymsp[0].minor.yy455 = yylhsminor.yy455;
        break;
      case 320: /* over_clause ::= OVER LP window RP */
#line 1730 "parse.y"
{
  yymsp[-3].minor.yy455 = yymsp[-1].minor.yy455;
  assert( yymsp[-3].minor.yy455!=0 );
}
#line 5140 "parse.c"
        break;
      case 321: /* over_clause ::= OVER nm */
#line 1734 "parse.y"
{
  yymsp[-1].minor.yy455 = (Window*)sqlite3DbMallocZero(pParse->db, sizeof(Window));
  if( yymsp[-1].minor.yy455 ){
    yymsp[-1].minor.yy455->zName = sqlite3DbStrNDup(pParse->db, yymsp[0].minor.yy0.z, yymsp[0].minor.yy0.n);
  }
}
#line 5150 "parse.c"
        break;
      case 322: /* filter_clause ::= FILTER LP WHERE expr RP */
#line 1741 "parse.y"
{ yymsp[-4].minor.yy46 = yymsp[-1].minor.yy46; }
#line 5155 "parse.c"
        break;
      default:
      /* (323) input ::= cmdlist */ yytestcase(yyruleno==323);
      /* (324) cmdlist ::= cmdlist ecmd */ yytestcase(yyruleno==324);
      /* (325) cmdlist ::= ecmd (OPTIMIZED OUT) */ assert(yyruleno!=325);
      /* (326) ecmd ::= SEMI */ yytestcase(yyruleno==326);
      /* (327) ecmd ::= cmdx SEMI */ yytestcase(yyruleno==327);
      /* (328) ecmd ::= explain cmdx */ yytestcase(yyruleno==328);
      /* (329) trans_opt ::= */ yytestcase(yyruleno==329);
      /* (330) trans_opt ::= TRANSACTION */ yytestcase(yyruleno==330);
      /* (331) trans_opt ::= TRANSACTION nm */ yytestcase(yyruleno==331);
      /* (332) savepoint_opt ::= SAVEPOINT */ yytestcase(yyruleno==332);
      /* (333) savepoint_opt ::= */ yytestcase(yyruleno==333);
      /* (334) cmd ::= create_table create_table_args */ yytestcase(yyruleno==334);
      /* (335) columnlist ::= columnlist COMMA columnname carglist */ yytestcase(yyruleno==335);
      /* (336) columnlist ::= columnname carglist */ yytestcase(yyruleno==336);
      /* (337) nm ::= ID|INDEXED */ yytestcase(yyruleno==337);
      /* (338) nm ::= STRING */ yytestcase(yyruleno==338);
      /* (339) nm ::= JOIN_KW */ yytestcase(yyruleno==339);
      /* (340) typetoken ::= typename */ yytestcase(yyruleno==340);
      /* (341) typename ::= ID|STRING */ yytestcase(yyruleno==341);
      /* (342) signed ::= plus_num (OPTIMIZED OUT) */ assert(yyruleno!=342);
      /* (343) signed ::= minus_num (OPTIMIZED OUT) */ assert(yyruleno!=343);
      /* (344) carglist ::= carglist ccons */ yytestcase(yyruleno==344);
      /* (345) carglist ::= */ yytestcase(yyruleno==345);
      /* (346) ccons ::= NULL onconf */ yytestcase(yyruleno==346);
      /* (347) conslist_opt ::= COMMA conslist */ yytestcase(yyruleno==347);
      /* (348) conslist ::= conslist tconscomma tcons */ yytestcase(yyruleno==348);
      /* (349) conslist ::= tcons (OPTIMIZED OUT) */ assert(yyruleno!=349);
      /* (350) tconscomma ::= */ yytestcase(yyruleno==350);
      /* (351) defer_subclause_opt ::= defer_subclause (OPTIMIZED OUT) */ assert(yyruleno!=351);
      /* (352) resolvetype ::= raisetype (OPTIMIZED OUT) */ assert(yyruleno!=352);
      /* (353) selectnowith ::= oneselect (OPTIMIZED OUT) */ assert(yyruleno!=353);
      /* (354) oneselect ::= values */ yytestcase(yyruleno==354);
      /* (355) sclp ::= selcollist COMMA */ yytestcase(yyruleno==355);
      /* (356) as ::= ID|STRING */ yytestcase(yyruleno==356);
      /* (357) expr ::= term (OPTIMIZED OUT) */ assert(yyruleno!=357);
      /* (358) likeop ::= LIKE_KW|MATCH */ yytestcase(yyruleno==358);
      /* (359) exprlist ::= nexprlist */ yytestcase(yyruleno==359);
      /* (360) nmnum ::= plus_num (OPTIMIZED OUT) */ assert(yyruleno!=360);
      /* (361) nmnum ::= nm (OPTIMIZED OUT) */ assert(yyruleno!=361);
      /* (362) nmnum ::= ON */ yytestcase(yyruleno==362);
      /* (363) nmnum ::= DELETE */ yytestcase(yyruleno==363);
      /* (364) nmnum ::= DEFAULT */ yytestcase(yyruleno==364);
      /* (365) plus_num ::= INTEGER|FLOAT */ yytestcase(yyruleno==365);
      /* (366) foreach_clause ::= */ yytestcase(yyruleno==366);
      /* (367) foreach_clause ::= FOR EACH ROW */ yytestcase(yyruleno==367);
      /* (368) trnm ::= nm */ yytestcase(yyruleno==368);
      /* (369) tridxby ::= */ yytestcase(yyruleno==369);
      /* (370) database_kw_opt ::= DATABASE */ yytestcase(yyruleno==370);
      /* (371) database_kw_opt ::= */ yytestcase(yyruleno==371);
      /* (372) kwcolumn_opt ::= */ yytestcase(yyruleno==372);
      /* (373) kwcolumn_opt ::= COLUMNKW */ yytestcase(yyruleno==373);
      /* (374) vtabarglist ::= vtabarg */ yytestcase(yyruleno==374);
      /* (375) vtabarglist ::= vtabarglist COMMA vtabarg */ yytestcase(yyruleno==375);
      /* (376) vtabarg ::= vtabarg vtabargtoken */ yytestcase(yyruleno==376);
      /* (377) anylist ::= */ yytestcase(yyruleno==377);
      /* (378) anylist ::= anylist LP anylist RP */ yytestcase(yyruleno==378);
      /* (379) anylist ::= anylist ANY */ yytestcase(yyruleno==379);
      /* (380) with ::= */ yytestcase(yyruleno==380);
        break;
/********** End reduce actions ************************************************/
  };
  assert( yyruleno<sizeof(yyRuleInfoLhs)/sizeof(yyRuleInfoLhs[0]) );
  yygoto = yyRuleInfoLhs[yyruleno];
  yysize = yyRuleInfoNRhs[yyruleno];
  yyact = yy_find_reduce_action(yymsp[yysize].stateno,(YYCODETYPE)yygoto);

  /* There are no SHIFTREDUCE actions on nonterminals because the table
  ** generator has simplified them to pure REDUCE actions. */
  assert( !(yyact>YY_MAX_SHIFT && yyact<=YY_MAX_SHIFTREDUCE) );

  /* It is not possible for a REDUCE to be followed by an error */
  assert( yyact!=YY_ERROR_ACTION );

  yymsp += yysize+1;
  yypParser->yytos = yymsp;
  yymsp->stateno = (YYACTIONTYPE)yyact;
  yymsp->major = (YYCODETYPE)yygoto;
  yyTraceShift(yypParser, yyact, "... then shift");
  return yyact;
}

/*
** The following code executes when the parse fails
*/
#ifndef YYNOERRORRECOVERY
static void yy_parse_failed(
  yyParser *yypParser           /* The parser */
){
  sqlite3ParserARG_FETCH
  sqlite3ParserCTX_FETCH
#ifndef NDEBUG
  if( yyTraceFILE ){
    fprintf(yyTraceFILE,"%sFail!\n",yyTracePrompt);
  }
#endif
  while( yypParser->yytos>yypParser->yystack ) yy_pop_parser_stack(yypParser);
  /* Here code is inserted which will be executed whenever the
  ** parser fails */
/************ Begin %parse_failure code ***************************************/
/************ End %parse_failure code *****************************************/
  sqlite3ParserARG_STORE /* Suppress warning about unused %extra_argument variable */
  sqlite3ParserCTX_STORE
}
#endif /* YYNOERRORRECOVERY */

/*
** The following code executes when a syntax error first occurs.
*/
static void yy_syntax_error(
  yyParser *yypParser,           /* The parser */
  int yymajor,                   /* The major type of the error token */
  sqlite3ParserTOKENTYPE yyminor         /* The minor type of the error token */
){
  sqlite3ParserARG_FETCH
  sqlite3ParserCTX_FETCH
#define TOKEN yyminor
/************ Begin %syntax_error code ****************************************/
#line 33 "parse.y"

  UNUSED_PARAMETER(yymajor);  /* Silence some compiler warnings */
  if( TOKEN.z[0] ){
    sqlite3ErrorMsg(pParse, "near \"%T\": syntax error", &TOKEN);
  }else{
    sqlite3ErrorMsg(pParse, "incomplete input");
  }
#line 5283 "parse.c"
/************ End %syntax_error code ******************************************/
  sqlite3ParserARG_STORE /* Suppress warning about unused %extra_argument variable */
  sqlite3ParserCTX_STORE
}

/*
** The following is executed when the parser accepts
*/
static void yy_accept(
  yyParser *yypParser           /* The parser */
){
  sqlite3ParserARG_FETCH
  sqlite3ParserCTX_FETCH
#ifndef NDEBUG
  if( yyTraceFILE ){
    fprintf(yyTraceFILE,"%sAccept!\n",yyTracePrompt);
  }
#endif
#ifndef YYNOERRORRECOVERY
  yypParser->yyerrcnt = -1;
#endif
  assert( yypParser->yytos==yypParser->yystack );
  /* Here code is inserted which will be executed whenever the
  ** parser accepts */
/*********** Begin %parse_accept code *****************************************/
/*********** End %parse_accept code *******************************************/
  sqlite3ParserARG_STORE /* Suppress warning about unused %extra_argument variable */
  sqlite3ParserCTX_STORE
}

/* The main parser program.
** The first argument is a pointer to a structure obtained from
** "sqlite3ParserAlloc" which describes the current state of the parser.
** The second argument is the major token number.  The third is
** the minor token.  The fourth optional argument is whatever the
** user wants (and specified in the grammar) and is available for
** use by the action routines.
**
** Inputs:
** <ul>
** <li> A pointer to the parser (an opaque structure.)
** <li> The major token number.
** <li> The minor token number.
** <li> An option argument of a grammar-specified type.
** </ul>
**
** Outputs:
** None.
*/
void sqlite3Parser(
  void *yyp,                   /* The parser */
  int yymajor,                 /* The major token code number */
  sqlite3ParserTOKENTYPE yyminor       /* The value for the token */
  sqlite3ParserARG_PDECL               /* Optional %extra_argument parameter */
){
  YYMINORTYPE yyminorunion;
  YYACTIONTYPE yyact;   /* The parser action. */
#if !defined(YYERRORSYMBOL) && !defined(YYNOERRORRECOVERY)
  int yyendofinput;     /* True if we are at the end of input */
#endif
#ifdef YYERRORSYMBOL
  int yyerrorhit = 0;   /* True if yymajor has invoked an error */
#endif
  yyParser *yypParser = (yyParser*)yyp;  /* The parser */
  sqlite3ParserCTX_FETCH
  sqlite3ParserARG_STORE

  assert( yypParser->yytos!=0 );
#if !defined(YYERRORSYMBOL) && !defined(YYNOERRORRECOVERY)
  yyendofinput = (yymajor==0);
#endif

  yyact = yypParser->yytos->stateno;
#ifndef NDEBUG
  if( yyTraceFILE ){
    if( yyact < YY_MIN_REDUCE ){
      fprintf(yyTraceFILE,"%sInput '%s' in state %d\n",
              yyTracePrompt,yyTokenName[yymajor],yyact);
    }else{
      fprintf(yyTraceFILE,"%sInput '%s' with pending reduce %d\n",
              yyTracePrompt,yyTokenName[yymajor],yyact-YY_MIN_REDUCE);
    }
  }
#endif

  do{
    assert( yyact==yypParser->yytos->stateno );
    yyact = yy_find_shift_action((YYCODETYPE)yymajor,yyact);
    if( yyact >= YY_MIN_REDUCE ){
      yyact = yy_reduce(yypParser,yyact-YY_MIN_REDUCE,yymajor,
                        yyminor sqlite3ParserCTX_PARAM);
    }else if( yyact <= YY_MAX_SHIFTREDUCE ){
      yy_shift(yypParser,yyact,(YYCODETYPE)yymajor,yyminor);
#ifndef YYNOERRORRECOVERY
      yypParser->yyerrcnt--;
#endif
      break;
    }else if( yyact==YY_ACCEPT_ACTION ){
      yypParser->yytos--;
      yy_accept(yypParser);
      return;
    }else{
      assert( yyact == YY_ERROR_ACTION );
      yyminorunion.yy0 = yyminor;
#ifdef YYERRORSYMBOL
      int yymx;
#endif
#ifndef NDEBUG
      if( yyTraceFILE ){
        fprintf(yyTraceFILE,"%sSyntax Error!\n",yyTracePrompt);
      }
#endif
#ifdef YYERRORSYMBOL
      /* A syntax error has occurred.
      ** The response to an error depends upon whether or not the
      ** grammar defines an error token "ERROR".  
      **
      ** This is what we do if the grammar does define ERROR:
      **
      **  * Call the %syntax_error function.
      **
      **  * Begin popping the stack until we enter a state where
      **    it is legal to shift the error symbol, then shift
      **    the error symbol.
      **
      **  * Set the error count to three.
      **
      **  * Begin accepting and shifting new tokens.  No new error
      **    processing will occur until three tokens have been
      **    shifted successfully.
      **
      */
      if( yypParser->yyerrcnt<0 ){
        yy_syntax_error(yypParser,yymajor,yyminor);
      }
      yymx = yypParser->yytos->major;
      if( yymx==YYERRORSYMBOL || yyerrorhit ){
#ifndef NDEBUG
        if( yyTraceFILE ){
          fprintf(yyTraceFILE,"%sDiscard input token %s\n",
             yyTracePrompt,yyTokenName[yymajor]);
        }
#endif
        yy_destructor(yypParser, (YYCODETYPE)yymajor, &yyminorunion);
        yymajor = YYNOCODE;
      }else{
        while( yypParser->yytos >= yypParser->yystack
            && (yyact = yy_find_reduce_action(
                        yypParser->yytos->stateno,
                        YYERRORSYMBOL)) > YY_MAX_SHIFTREDUCE
        ){
          yy_pop_parser_stack(yypParser);
        }
        if( yypParser->yytos < yypParser->yystack || yymajor==0 ){
          yy_destructor(yypParser,(YYCODETYPE)yymajor,&yyminorunion);
          yy_parse_failed(yypParser);
#ifndef YYNOERRORRECOVERY
          yypParser->yyerrcnt = -1;
#endif
          yymajor = YYNOCODE;
        }else if( yymx!=YYERRORSYMBOL ){
          yy_shift(yypParser,yyact,YYERRORSYMBOL,yyminor);
        }
      }
      yypParser->yyerrcnt = 3;
      yyerrorhit = 1;
      if( yymajor==YYNOCODE ) break;
      yyact = yypParser->yytos->stateno;
#elif defined(YYNOERRORRECOVERY)
      /* If the YYNOERRORRECOVERY macro is defined, then do not attempt to
      ** do any kind of error recovery.  Instead, simply invoke the syntax
      ** error routine and continue going as if nothing had happened.
      **
      ** Applications can set this macro (for example inside %include) if
      ** they intend to abandon the parse upon the first syntax error seen.
      */
      yy_syntax_error(yypParser,yymajor, yyminor);
      yy_destructor(yypParser,(YYCODETYPE)yymajor,&yyminorunion);
      break;
#else  /* YYERRORSYMBOL is not defined */
      /* This is what we do if the grammar does not define ERROR:
      **
      **  * Report an error message, and throw away the input token.
      **
      **  * If the input token is $, then fail the parse.
      **
      ** As before, subsequent error messages are suppressed until
      ** three input tokens have been successfully shifted.
      */
      if( yypParser->yyerrcnt<=0 ){
        yy_syntax_error(yypParser,yymajor, yyminor);
      }
      yypParser->yyerrcnt = 3;
      yy_destructor(yypParser,(YYCODETYPE)yymajor,&yyminorunion);
      if( yyendofinput ){
        yy_parse_failed(yypParser);
#ifndef YYNOERRORRECOVERY
        yypParser->yyerrcnt = -1;
#endif
      }
      break;
#endif
    }
  }while( yypParser->yytos>yypParser->yystack );
#ifndef NDEBUG
  if( yyTraceFILE ){
    yyStackEntry *i;
    char cDiv = '[';
    fprintf(yyTraceFILE,"%sReturn. Stack=",yyTracePrompt);
    for(i=&yypParser->yystack[1]; i<=yypParser->yytos; i++){
      fprintf(yyTraceFILE,"%c%s", cDiv, yyTokenName[i->major]);
      cDiv = ' ';
    }
    fprintf(yyTraceFILE,"]\n");
  }
#endif
  return;
}

/*
** Return the fallback token corresponding to canonical token iToken, or
** 0 if iToken has no fallback.
*/
int sqlite3ParserFallback(int iToken){
#ifdef YYFALLBACK
  assert( iToken<(int)(sizeof(yyFallback)/sizeof(yyFallback[0])) );
  return yyFallback[iToken];
#else
  (void)iToken;
#endif
  return 0;
}
